/*
 * NucleotideSNP.cpp
 *
 *  Created on: 8 oct. 2012
 *      Author: abernard
 */

#include "AlleleFrequency.h"

// Basic C++ classes
#include <string>

using namespace std;
using namespace bpp;
using namespace ppc;

/*-------------------------------*/
/*          Constructor          */
/*-------------------------------*/

// Work with simple PolymorphismSequenceContainer
AlleleFrequency::AlleleFrequency(PolymorphismSequenceContainer* ingroupContainer, int codonSiteNumber, int positionNumber, const StandardCodonAlphabet* codonAlphabet, const NucleicAlphabet* nucleoAlphabet):
			_number_of_allele(0), _nb_typed(0), _allele_1_code(-1), _allele_2_code(-1), _allele_3_code(-1), _allele_2_rank(0), _nb_observed_heterozygosity(0), _allele_1_occurence(0),
			_allele_2_occurence(0),	_allele_3_occurence(0), _GC_absolute_frequency(-1), _GC_relative_frequency(-1.), _bi_allelic(false) {

	// Initializations
	int currentPositionIntCode(-1);

	// Loop through the list of sequence of the container and collect data on the alleles
	for (unsigned int sequenceIndex = 0; sequenceIndex < ingroupContainer->getNumberOfSequences(); sequenceIndex++) {

		// Get the base int code for the current position
		currentPositionIntCode = codonAlphabet->getNPosition(ingroupContainer->valueAt(sequenceIndex, codonSiteNumber), positionNumber);

		// If the base is A, C, G, T then the position is resolved
		if (nucleoAlphabet->isUnresolved(currentPositionIntCode)) continue;
		_nb_typed++;

		_updateAlleleData(currentPositionIntCode, sequenceIndex);
	}

	// Check if the position is biallelic
	if (_number_of_allele == 2) {
		_bi_allelic = true;
	} else {
		_bi_allelic = false;
	}
}

/******************************************************************************/

// Work with a PairedSequenceContainer (ie vector of PolymorphismSequenceContainer)
AlleleFrequency::AlleleFrequency(PairedSequenceContainer* pairContainer, int codonSiteNumber, int positionNumber, const StandardCodonAlphabet* codonAlphabet, const NucleicAlphabet* nucleoAlphabet):
			_number_of_allele(0), _nb_typed(0), _allele_1_code(-1), _allele_2_code(-1), _allele_3_code(-1), _allele_2_rank(0), _nb_observed_heterozygosity(0), _allele_1_occurence(0),
			_allele_2_occurence(0),	_allele_3_occurence(0), _GC_absolute_frequency(-1), _GC_relative_frequency(-1.), _bi_allelic(false) {

	// Browse pairs
	for (unsigned int pairNumber = 0; pairNumber < pairContainer->get_nb_pair(); pairNumber++) {

		// Get a pair
		PolymorphismSequenceContainer* currentPair = pairContainer->get_pair(pairNumber);

		int intCodeAllele1 = codonAlphabet->getNPosition(currentPair->valueAt(0, codonSiteNumber), positionNumber);
		int intCodeAllele2 = codonAlphabet->getNPosition(currentPair->valueAt(1, codonSiteNumber), positionNumber);

		if (nucleoAlphabet->isUnresolved(intCodeAllele1)) continue;
		if (nucleoAlphabet->isUnresolved(intCodeAllele2)) continue;

		if (intCodeAllele1 != intCodeAllele2) _nb_observed_heterozygosity++;

		_nb_typed++;

		_updateAlleleData(intCodeAllele1, pairNumber);
		_updateAlleleData(intCodeAllele2, pairNumber);
	}

	// Check if the position is biallelic
	if (_number_of_allele == 2) {
		_bi_allelic = true;
	} else {
		_bi_allelic = false;
	}
}

/*--------------------------------*/
/*           Destructor           */
/*--------------------------------*/

AlleleFrequency::~AlleleFrequency() {
	// TODO Auto-generated destructor stub
}

/*--------------------------------*/
/*            Accessor            */
/*--------------------------------*/

unsigned int AlleleFrequency::get_number_of_allele() { return _number_of_allele; }
unsigned int AlleleFrequency::get_nb_typed() { return _nb_typed; }

int AlleleFrequency::get_allele_1_code() { return _allele_1_code; }
int AlleleFrequency::get_allele_2_code() { return _allele_2_code; }
int AlleleFrequency::get_allele_3_code() { return _allele_3_code; }

unsigned int AlleleFrequency::get_allele_2_rank() { return _allele_2_rank; }
unsigned int AlleleFrequency::get_nb_observed_heterozygosity() { return _nb_observed_heterozygosity; }

unsigned int AlleleFrequency::get_allele_1_occurence() { return _allele_1_occurence; }
unsigned int AlleleFrequency::get_allele_2_occurence() { return _allele_2_occurence; }
unsigned int AlleleFrequency::get_allele_3_occurence() { return _allele_3_occurence; }

int AlleleFrequency::get_GC_absolute_frequency() { return _GC_absolute_frequency; }
double AlleleFrequency::get_GC_relative_frequency() { return _GC_relative_frequency; }

bool AlleleFrequency::is_bi_allelic() { return _bi_allelic; }

/*------------------------------------*/
/*           Public methods           */
/*------------------------------------*/

void AlleleFrequency::determine_GC_frequencies() {

	if (_bi_allelic) {

		if ((_allele_1_code == 1 || _allele_1_code == 2) && (_allele_2_code == 0 || _allele_2_code == 3)) {
			_GC_absolute_frequency = _allele_1_occurence;
			_GC_relative_frequency = (double)_allele_1_occurence / (double)_nb_typed;

		} else if ((_allele_1_code == 0 || _allele_1_code == 3) && (_allele_2_code == 1 || _allele_2_code == 2)) {
			_GC_absolute_frequency = _allele_2_occurence;
			_GC_relative_frequency = (double)_allele_2_occurence / (double)_nb_typed;
		}

	} else {
		ApplicationTools::displayWarning("You are trying to determine GC frequencies on a non bi-allelic position ! Frequencies will not be computed..");
	}
}

/*------------------------------------*/
/*           Public methods           */
/*------------------------------------*/

void AlleleFrequency::_updateAlleleData(int currentPositionIntCode, unsigned int index) {

	switch (_number_of_allele) {

		case 0:
			_allele_1_code = currentPositionIntCode;
			_number_of_allele++;
			_allele_1_occurence++;
			break;

		case 1:
			// We have a new allele or the first one
			if (currentPositionIntCode != _allele_1_code) {
				_allele_2_code = currentPositionIntCode;
				_allele_2_occurence++;
				_allele_2_rank = index;
				_number_of_allele++;
			} else {
				_allele_1_occurence++;
			}
			break;

		case 2:
			// We have a new allele or one of the first two
			if (currentPositionIntCode != _allele_1_code && currentPositionIntCode != _allele_2_code) {
				_allele_3_code = currentPositionIntCode;
				_allele_3_occurence++;
				_number_of_allele++;
			} else if (currentPositionIntCode == _allele_1_code) {
				_allele_1_occurence++;
			} else {
				_allele_2_occurence++;
			}
			break;

		case 3:
			// We have a new allele or one of the first three (we don't won't to keep the fourth allele code)
			if (currentPositionIntCode != _allele_1_code && currentPositionIntCode != _allele_2_code && currentPositionIntCode != _allele_3_code) {
				_number_of_allele++;
			} else if (currentPositionIntCode == _allele_1_code) {
				_allele_1_occurence++;
			} else if (currentPositionIntCode == _allele_2_code) {
				_allele_2_occurence++;
			} else {
				_allele_3_occurence++;
			}
			break;

		default:
			if (currentPositionIntCode == _allele_1_code) {
				_allele_1_occurence++;
			} else if (currentPositionIntCode == _allele_2_code) {
				_allele_2_occurence++;
			} else if (currentPositionIntCode == _allele_3_code) {
				_allele_3_occurence++;
			} else {
				// Nothing to do here
			}
			break;
	}
}
