/*
 * DistanceBetweenIndividuals.cpp
 *
 *  Created on: 28 nov. 2012
 *      Author: abernard
 */

#include "DistanceBetweenIndividuals.h"

// PopPhyl Project Developments
#include "Tools/PolymorphismSequenceContainerAddons.h"

// Namespaces
using namespace std;
using namespace bpp;
using namespace ppc;

/*--------------------------------*/
/*          Constructors          */
/*--------------------------------*/

DistanceBetweenIndividuals::DistanceBetweenIndividuals(std::string individualName1, std::string individualName2): _first_individual_name(individualName1),
		_second_individual_name(individualName2), _pair_name(individualName1 + "|" + individualName2), _distance(-1), _nb_resolved_positions(0) {
	// Nothing to do here, everything is done in the initialization list
}

/******************************************************************************/

DistanceBetweenIndividuals::DistanceBetweenIndividuals(PolymorphismSequenceContainer* individual1, PolymorphismSequenceContainer* individual2): _distance(-1),
		_nb_resolved_positions(0) {

	_first_individual_name = PolymorphismSequenceContainerAddons::getIndividualNameFromContainer(individual1, 0);
	_second_individual_name = PolymorphismSequenceContainerAddons::getIndividualNameFromContainer(individual2, 0);

	_pair_name = _first_individual_name + "|" + _second_individual_name;
}

/******************************************************************************/

DistanceBetweenIndividuals::DistanceBetweenIndividuals(PairedSequenceContainer* pairContainer, unsigned int pairIndex1, unsigned int pairIndex2): _distance(-1),
		_nb_resolved_positions(0) {

	vector<string> allNames = pairContainer->get_pair_names();

	_first_individual_name = allNames[pairIndex1];
	_second_individual_name = allNames[pairIndex2];

	_pair_name = _first_individual_name + "|" + _second_individual_name;
}


/*--------------------------------*/
/*           Destructor           */
/*--------------------------------*/

DistanceBetweenIndividuals::~DistanceBetweenIndividuals() {
	// TODO Auto-generated destructor stub
}

/*---------------------------------*/
/*            Accessors            */
/*---------------------------------*/

// GET
string DistanceBetweenIndividuals::get_first_individual_name() { return _first_individual_name; }
string DistanceBetweenIndividuals::get_second_individual_name() { return _second_individual_name; }
string DistanceBetweenIndividuals::get_pair_name() { return _pair_name; }

double DistanceBetweenIndividuals::get_distance() { return _distance; }
int DistanceBetweenIndividuals::get_nb_resolved_positions() { return _nb_resolved_positions; }

// SET
void DistanceBetweenIndividuals::set_distance(double computedDistance) { _distance = computedDistance; }
void DistanceBetweenIndividuals::set_nb_resolved_positions(int determinedNbResolvedPositions) { _nb_resolved_positions = determinedNbResolvedPositions; }

/*------------------------------------*/
/*           Public methods           */
/*------------------------------------*/

/* Return the probability of drawing distinct alleles when sampling at the
same site */
/* in two distinct bi-allelic individuals */
/* AA vs AA: 0 */
/* AA vs AG: 0.5 */
/* AA vs GG: 1 */
/* AG vs AG: 0.5 */
/* AG vs CC: 1 */
/* AG vs CG: 0.75 */
/* AG vs CT: 1 */

void DistanceBetweenIndividuals::computeDistance(PolymorphismSequenceContainer* individual1, PolymorphismSequenceContainer* individual2,
		const StandardCodonAlphabet* codonAlphabet, const NucleicAlphabet* nucleoAlphabet, bool debugMode) {

	// Initializations
	double diff(0.);

	// Check the number of site in each pair
	if (individual1->getNumberOfSites() == individual2->getNumberOfSites()) {

		for (unsigned int codonSiteNumber = 0; codonSiteNumber < individual1->getNumberOfSites(); codonSiteNumber++) {

			for (unsigned int positionNumber = 0; positionNumber < 3; positionNumber++) {

				int individual1Allele1 = codonAlphabet->getNPosition(individual1->valueAt(0, codonSiteNumber), positionNumber);
				int individual1Allele2 = codonAlphabet->getNPosition(individual1->valueAt(1, codonSiteNumber), positionNumber);

				int individual2Allele1 = codonAlphabet->getNPosition(individual2->valueAt(0, codonSiteNumber), positionNumber);
				int individual2Allele2 = codonAlphabet->getNPosition(individual2->valueAt(1, codonSiteNumber), positionNumber);

				if (debugMode) {
					cout << "(DEBUG - DistanceBetweenIndividuals::computeDistance) Alleles 1-1, 1-2, 2-1, 2-2: " << nucleoAlphabet->getName(individual1Allele1)
							<< ", " << nucleoAlphabet->getName(individual1Allele2) << ", " << nucleoAlphabet->getName(individual2Allele1) << ", " << nucleoAlphabet->getName(individual2Allele2) << endl << endl;
				}

				// Count difference if there is no Unresolved bases
				if (nucleoAlphabet->isUnresolved(individual1Allele1) || nucleoAlphabet->isUnresolved(individual1Allele2) || nucleoAlphabet->isUnresolved(individual2Allele1) || nucleoAlphabet->isUnresolved(individual2Allele2)) {
					continue;
				} else {
					_nb_resolved_positions++;
					if(individual1Allele1 != individual2Allele1) diff += 0.25;
					if(individual1Allele1 != individual2Allele2) diff += 0.25;
					if(individual1Allele2 != individual2Allele1) diff += 0.25;
					if(individual1Allele2 != individual2Allele2) diff += 0.25;
				}
			}
		}

		if (_nb_resolved_positions != 0) _distance = diff / _nb_resolved_positions;

	} else if (debugMode) {
		cout << "(DEBUG - DistanceBetweenIndividuals::computeDistance) Warning: Different number of site detected for individual " << _first_individual_name << " and individual " << _second_individual_name << "! Distance will not be computed !" << endl;
	}
}

void DistanceBetweenIndividuals::computeDistanceWithPair(PairedSequenceContainer* pairContainer, unsigned int pairIndex1, unsigned int pairIndex2,
		const StandardCodonAlphabet* codonAlphabet, const NucleicAlphabet* nucleoAlphabet, bool debugMode) {

	computeDistance(pairContainer->get_pair(pairIndex1), pairContainer->get_pair(pairIndex2), codonAlphabet, nucleoAlphabet, debugMode);
}


