/*
 * Population.cpp
 *
 *  Created on: 24 oct. 2012
 *      Author: abernard
 */

#include "Population.h"

// Basic C++ classes
#include <stdlib.h>
#include <set>

// Bio++ - Message management
#include <Bpp/App/ApplicationTools.h>
#include <Bpp/Text/TextTools.h>

// PopPhyl Project Developments
#include "Tools/PolymorphismSequenceContainerAddons.h"

using namespace std;
using namespace bpp;
using namespace ppc;

/*--------------------------------*/
/*          Constructors          */
/*--------------------------------*/

// Work with a population file
Population::Population(string populationFile): _number_of_population(0) {

	vector<string> populationFileLines = convert_file_into_vector_of_string(populationFile);

	build_population_list(populationFileLines);

	_number_of_population = _population_list.size();
}

/******************************************************************************/

// Work with a vector of PolymorphismSequenceContainer
Population::Population(vector< PolymorphismSequenceContainer* >& allDatasets, string ingroupSpeciesName): _number_of_population(1) {

	// Initializations
	set<string> ingroupIndividuals;

	// Extract all focal individual names from the container
	for (unsigned int i = 0; i < allDatasets.size(); i++) {
		for (unsigned int j = 0; j < allDatasets[i]->getNumberOfSequences(); j++) {
			vector<string> nameElements = PolymorphismSequenceContainerAddons::getNameElementsFromContainer(allDatasets[i], "|", j, 4);

			if (nameElements.size() < 3) ApplicationTools::displayError("The name of dataset " + TextTools::toString(i) + " contains less than 3 elements separated by a | (pipe) character !\n" + "Global population can't be generated !");

			if (nameElements[1] == ingroupSpeciesName) {
				ingroupIndividuals.insert(nameElements[2]);
			}
		}
	}

	// Generate the unique population
	vector<string> builtPopulation;

	if (ingroupIndividuals.size() > 0) {
		for(set<string>::iterator individualIterator = ingroupIndividuals.begin();  individualIterator != ingroupIndividuals.end(); individualIterator++) builtPopulation.push_back(*individualIterator);

		_population_list.push_back(builtPopulation);
		_population_names.push_back("All_individuals");
	} else {
		ApplicationTools::displayError("There is no focal individuals in the selected PolymorphismSequenceContainer");
	}

	_number_of_population = 1;
}


/*--------------------------------*/
/*           Destructor           */
/*--------------------------------*/

Population::~Population() {
	// TODO Auto-generated destructor stub
}


/*---------------------------------------*/
/*           Protected methods           */
/*---------------------------------------*/

vector<string> Population::convert_file_into_vector_of_string(string populationFile) {

	//Initializations
	vector<string> allStrings;
	string currentLine = "";

	// Open population file
	ifstream popFileStream (populationFile.c_str(), ios::in);
	if (!popFileStream.is_open()) {
		ApplicationTools::displayError("Cannot open/read file: " + populationFile);
		exit(EXIT_FAILURE);
	}

	// Convert
	while (getline(popFileStream, currentLine)) {
		allStrings.push_back(currentLine);
	}

	// Close population file
	popFileStream.close();

	return allStrings;
}

/******************************************************************************/

void Population::build_population_list(std::vector<std::string> AllPopdata) {

	for (unsigned int i = 0; i < AllPopdata.size(); i++) {

		// Jump over comment lines and empty lines
		if (AllPopdata[i][0] == '#' || AllPopdata[i].empty()) continue;

		// Initializations
		vector<string> newPopulation;
		string newPopulationName("");

		// Split line of interest (Example: Population_name;Individual_1;Individual_2;Individual_3)
		StringTokenizer tokens = StringTokenizer(AllPopdata[i], ";");

		// Save population name
		if (tokens.hasMoreToken()) newPopulationName = tokens.nextToken();

		// Add each individual to the new population
		while (tokens.hasMoreToken()) newPopulation.push_back(tokens.nextToken());

		// If the population is not empty we had it to the list
		if (newPopulation.size() > 0) {
			_population_list.push_back(newPopulation);
			_population_names.push_back(newPopulationName);
		}

		// Delete local data
		newPopulation.erase(newPopulation.begin(), newPopulation.end());
	}
}


/*--------------------------------*/
/*            Accessor            */
/*--------------------------------*/

// GET
vector<string> Population::get_population_names() {	return _population_names; }
vector< vector<string> > Population::get_population_list() { return _population_list; }
vector<string> Population::get_duplicates() { return _duplicates; }

unsigned int Population::get_number_of_population() { return _number_of_population; }

/*------------------------------------*/
/*           Public methods           */
/*------------------------------------*/

int Population::get_population_id(string individualName) throw (Exception) {

	// Initialization
	int validPopulationIdentifier(-1);

	// Browse individual
	for (unsigned int i = 0; i < _population_list.size(); i++) {
		for (unsigned int j = 0; j < _population_list[i].size(); j++) {
			if (_population_list[i][j] == individualName) validPopulationIdentifier = i;
		}
	}

	if (validPopulationIdentifier == -1) throw Exception("Individual " + individualName + " doesn't belong to any population !");

	return validPopulationIdentifier;
}

/******************************************************************************/

string Population::get_population_name(string individualName) {

	// Call get_population_id method
	int validPopulationIdentifier = get_population_id(individualName);

	if (validPopulationIdentifier != -1) {
		return _population_names[validPopulationIdentifier];
	} else {
		return "";
	}
}

/******************************************************************************/

vector<string> Population::get_all_individuals() {

	// Initializations
	vector<string> allIndividuals(0);

	for (unsigned int i = 0; i < _population_list.size(); i++) {
		for (unsigned int j = 0; j < _population_list[i].size(); j++) {
			allIndividuals.push_back(_population_list[i][j]);
		}
	}

	return allIndividuals;
}

/******************************************************************************/

vector<string> Population::get_all_couples() {

	// Initializations
	vector<string> allCoupleNames(0);

	if (_number_of_population > 1) {
		for (unsigned int popIndex1 = 0; popIndex1 < _number_of_population; popIndex1++) {
			for (unsigned int popIndex2 = popIndex1 + 1; popIndex2 < _number_of_population; popIndex2++) {
				allCoupleNames.push_back(_population_names[popIndex1] + "|" + _population_names[popIndex2]);
			}
		}

	}

	return allCoupleNames;
}

/******************************************************************************/

unsigned int Population::get_number_of_individuals() {

	// Initializations
	unsigned int nbIndividuals(0);

	// Count the total number of individuals in all the populations
	for (unsigned int i = 0; i < _population_list.size(); i++) {
		nbIndividuals += _population_list[i].size();
	}

	return nbIndividuals;
}

/******************************************************************************/

bool Population::check_for_duplicates() {

	// Initializations
	bool hasDuplicates(false);
	map<string, int> duplicatesIdentifier;

	// Browse populations
	for (unsigned int i = 0; i < _population_list.size(); i++) {

		// Browse individuals
		for (unsigned int j = 0; j < _population_list[i].size(); j++) {

			duplicatesIdentifier[_population_list[i][j]]++;

			if (duplicatesIdentifier[_population_list[i][j]] > 1) {
				hasDuplicates = true;
				_duplicates.push_back(_population_list[i][j]);
			}
		}
	}

	return hasDuplicates;
}

/******************************************************************************/

void Population::display_populations(bool generated) {

	if (generated) {
		cout << "The following population has been extracted from your data:" << endl << endl;
	} else {
		cout << "The selected population file contains " << _number_of_population << " population(s):" << endl << endl;
	}

	for (unsigned int i = 0; i < _population_list.size(); i++) {
		cout << " -> Population " << i+1 << ":";
		for (unsigned int j = 0; j < _population_list[i].size(); j++) {
			cout << " " << _population_list[i][j];

		}
		cout << endl;
	}
	cout << endl;
}

/******************************************************************************/

void Population::display_duplicates() {

	ApplicationTools::displayWarning("Duplicates has been detected in the selected population file !");

	cout << endl << "Here is the list of individual that exists in at least 2 populations:";

	for (unsigned int i = 0; i < _duplicates.size(); i++) {
		cout << " " << _duplicates[i];
	}

	cout << endl << endl;
}
