/*
 * PopulationDiversification.cpp
 *
 *  Created on: 14 nov. 2012
 *      Author: abernard
 */

#include "PopulationDiversification.h"

// Bio++ classes
//#include <Bpp/PopGen/PolymorphismSequenceContainerTools.h>

// PopPhyl Project Developments
#include "Tools/PolymorphismSequenceContainerAddons.h"

// Namespaces
using namespace std;
using namespace bpp;
using namespace ppc;

/*-------------------------------*/
/*          Constructor          */
/*-------------------------------*/

PopulationDiversification::PopulationDiversification(PolymorphismSequenceContainer* selectedContainer, Population* populationData, const StandardCodonAlphabet* codonAlphabet,
		const NucleicAlphabet* nucleoAlphabet): _differenceWithin(populationData->get_number_of_population(), 0), _totalWithin(populationData->get_number_of_population(), 0),
		_internal_pop_pointer(populationData) {

	// Initialize vectors
	int nbPopulation = _internal_pop_pointer->get_number_of_population();

	vector<int> init_int_matrix (nbPopulation, 0);
	vector<double> init_double_matrix (nbPopulation, 0.);
	vector<double> init_Fst (nbPopulation, -999.);

	for (int i = 0; i < nbPopulation; i++) {
		_differenceBetween.push_back(init_int_matrix);
		_totalBetween.push_back(init_int_matrix);
		_globalHeteroWithin.push_back(init_double_matrix);
		_globalHeteroBetween.push_back(init_double_matrix);
		_Fst.push_back(init_Fst);
	}

	// Collect within/between difference data
	_collect_differences(selectedContainer, codonAlphabet, nucleoAlphabet);
}


/*--------------------------------*/
/*           Destructor           */
/*--------------------------------*/

PopulationDiversification::~PopulationDiversification() {

	_internal_pop_pointer = 0;
}


/*---------------------------------*/
/*            Accessors            */
/*---------------------------------*/

// GET ALL
vector<int> PopulationDiversification::get_all_differenceWithin() { return _differenceWithin; }
vector<int> PopulationDiversification::get_all_totalWithin() {return _totalWithin; }

vector< vector<int> > PopulationDiversification::get_all_differenceBetween() { return _differenceBetween; }
vector< vector<int> > PopulationDiversification::get_all_totalBetween() { return _totalBetween; }

map<int, double> PopulationDiversification::get_all_heterozygosityWithinPops() { return _heterozygosityWithinPops; }

vector< vector<double> > PopulationDiversification::get_all_globalHeteroWithin() { return _globalHeteroWithin; }
vector< vector<double> > PopulationDiversification::get_all_globalHeteroBetween() { return _globalHeteroBetween; }
vector< vector<double> > PopulationDiversification::get_all_Fst() { return _Fst; }


// GET SPECIFIC
int PopulationDiversification::get_specific_differenceWithin(int popIndex) { return _differenceWithin[popIndex]; }
int PopulationDiversification::get_specific_totalWithin(int popIndex) { return _totalWithin[popIndex]; }

int PopulationDiversification::get_specific_differenceBetween(int popIndex1, int popIndex2) { return _differenceBetween[popIndex1][popIndex2]; }
int PopulationDiversification::get_specific_totalBetween(int popIndex1, int popIndex2) { return _totalBetween[popIndex1][popIndex2]; }

double PopulationDiversification::get_specific_heterozygosityWithinPop(int popIndex) { return _heterozygosityWithinPops[popIndex]; }

double PopulationDiversification::get_specific_globalHeteroWithin(int popIndex1, int popIndex2) { return _globalHeteroWithin[popIndex1][popIndex2]; }
double PopulationDiversification::get_specific_globalHeteroBetween(int popIndex1, int popIndex2) { return _globalHeteroBetween[popIndex1][popIndex2]; }
double PopulationDiversification::get_specific_Fst(int popIndex1, int popIndex2) { return _Fst[popIndex1][popIndex2]; }


/*---------------------------------------*/
/*           Protected methods           */
/*---------------------------------------*/

void PopulationDiversification::_collect_differences(PolymorphismSequenceContainer* selectedContainer, const bpp::StandardCodonAlphabet* codonAlphabet, const bpp::NucleicAlphabet* nucleoAlphabet) {

	// Count within/between differences
	for (unsigned int seqIndex1 = 0; seqIndex1 < selectedContainer->getNumberOfSequences(); seqIndex1++) {

		string individual1 = PolymorphismSequenceContainerAddons::getIndividualNameFromContainer(selectedContainer, seqIndex1);
		int population1 = _internal_pop_pointer->get_population_id(individual1);

		for (unsigned int seqIndex2 = 0; seqIndex2 < selectedContainer->getNumberOfSequences(); seqIndex2++) {

			string individual2 = PolymorphismSequenceContainerAddons::getIndividualNameFromContainer(selectedContainer, seqIndex2);
			if (individual1 == individual2) continue;
			int population2 = _internal_pop_pointer->get_population_id(individual2);

			// Browse codon sites
			for (unsigned int codonSiteNumber = 0; codonSiteNumber < selectedContainer->getNumberOfSites(); codonSiteNumber++) {
//			for (unsigned int codonSiteNumber = 0; codonSiteNumber < PolymorphismSequenceContainerTools::getNumberOfCompleteSites(*selectedContainer, false); codonSiteNumber++) {

				// Look at each position of the codon site
				for (unsigned int positionNumber = 0; positionNumber < 3; positionNumber++) {

					int numberOfDifference(0);

					int position1 = codonAlphabet->getNPosition(selectedContainer->valueAt(seqIndex1, codonSiteNumber), positionNumber);
					int position2 = codonAlphabet->getNPosition(selectedContainer->valueAt(seqIndex2, codonSiteNumber), positionNumber);

					if (nucleoAlphabet->isUnresolved(position1)) continue;
					if (nucleoAlphabet->isUnresolved(position2)) continue;

					if (position1 != position2) numberOfDifference = 1;

					if (population1 == population2) {
						_differenceWithin[population1] += numberOfDifference;
						_totalWithin[population1]++;
					} else {
						_differenceBetween[population1][population2] += numberOfDifference;
						_totalBetween[population1][population2]++;
					}
				}
			}
		}
	}
}


/*------------------------------------*/
/*           Public methods           */
/*------------------------------------*/

void PopulationDiversification::determine_Fst() {

	// Browse populations
	for (unsigned int popIndex1 = 0; popIndex1 < _internal_pop_pointer->get_number_of_population(); popIndex1++) {

		if (_totalWithin[popIndex1] == 0) continue;

		_heterozygosityWithinPops[popIndex1] = (double)_differenceWithin[popIndex1] / (double)_totalWithin[popIndex1];

		for (unsigned int popIndex2 = popIndex1 + 1; popIndex2 < _internal_pop_pointer->get_number_of_population(); popIndex2++) {

			if (_totalWithin[popIndex2] == 0 || _totalBetween[popIndex1][popIndex2] == 0) continue;

			_heterozygosityWithinPops[popIndex2] = (double)_differenceWithin[popIndex2] / (double)_totalWithin[popIndex2];

			_globalHeteroBetween[popIndex1][popIndex2] = (double)_differenceBetween[popIndex1][popIndex2] / (double)_totalBetween[popIndex1][popIndex2];
			_globalHeteroBetween[popIndex2][popIndex1] = _globalHeteroBetween[popIndex1][popIndex2];

			_globalHeteroWithin[popIndex1][popIndex2] = (_heterozygosityWithinPops[popIndex1] + _heterozygosityWithinPops[popIndex2]) / 2.;
			_globalHeteroWithin[popIndex2][popIndex1] = _globalHeteroWithin[popIndex1][popIndex2];

			if (_globalHeteroBetween[popIndex1][popIndex2] != 0.) {
				_Fst[popIndex1][popIndex2] = (_globalHeteroBetween[popIndex1][popIndex2] - _globalHeteroWithin[popIndex1][popIndex2]) / _globalHeteroBetween[popIndex1][popIndex2];
				_Fst[popIndex2][popIndex1] = _Fst[popIndex1][popIndex2];
			}
		}
	}
}

/******************************************************************************/

void PopulationDiversification::determine_heterozygosities_only() {

	// Browse populations
	for (unsigned int popIndex1 = 0; popIndex1 < _internal_pop_pointer->get_number_of_population(); popIndex1++) {

		if (_totalWithin[popIndex1] == 0) continue;

		_heterozygosityWithinPops[popIndex1] = (double)_differenceWithin[popIndex1] / (double)_totalWithin[popIndex1];

		for (unsigned int popIndex2 = popIndex1 + 1; popIndex2 < _internal_pop_pointer->get_number_of_population(); popIndex2++) {

			if (_totalWithin[popIndex2] == 0 || _totalBetween[popIndex1][popIndex2] == 0) continue;

			_heterozygosityWithinPops[popIndex2] = (double)_differenceWithin[popIndex2] / (double)_totalWithin[popIndex2];

			_globalHeteroBetween[popIndex1][popIndex2] = (double)_differenceBetween[popIndex1][popIndex2] / (double)_totalBetween[popIndex1][popIndex2];
			_globalHeteroBetween[popIndex2][popIndex1] = _globalHeteroBetween[popIndex1][popIndex2];

			_globalHeteroWithin[popIndex1][popIndex2] = (_heterozygosityWithinPops[popIndex1] + _heterozygosityWithinPops[popIndex2]) / 2.;
			_globalHeteroWithin[popIndex2][popIndex1] = _globalHeteroWithin[popIndex1][popIndex2];
		}
	}

}
