/*
 * SFS.cpp
 *
 *  Created on: 22 oct. 2012
 *      Author: abernard
 */

#include "SFS.h"

// Basic C++ classes
#include <stdlib.h>
#include <math.h>
#include <iostream>

// Bio++ classes
#include <Bpp/App/ApplicationTools.h>

using namespace std;
using namespace bpp;
using namespace ppc;

/*-------------------------------*/
/*          Constructor          */
/*-------------------------------*/

SFS::SFS(unsigned int maxVectorSize, unsigned int sampleSize, vector< vector<double> >& originalMatrix, bool requireFolding): _sumlnk(maxVectorSize+1, 0.), _sample_size(sampleSize), _needFolding(requireFolding) {

	// Precompute some data for hyper geometric drawing
	compute_sumlnk(maxVectorSize);

	// Hyper geometric drawing
	if (_needFolding) {
		// Fold
		vector < vector<double> > foldedMatrix = fold_square_matrix(originalMatrix, maxVectorSize);

		// Draw
		folded_hypergeometric_drawing(foldedMatrix, maxVectorSize);
	} else {
		unfolded_hypergeometric_drawing(originalMatrix, maxVectorSize);
	}

}


SFS::SFS(unsigned int maxVectorSize, unsigned int sampleSize, vector<double>& hypergeometricDraw): _sample_size(sampleSize), _needFolding(false) {

	_Hypergeometric_draw = hypergeometricDraw;
}


/*--------------------------------*/
/*           Destructor           */
/*--------------------------------*/

SFS::~SFS() {
	// TODO Auto-generated destructor stub
}


/*---------------------------------------*/
/*           Protected methods           */
/*---------------------------------------*/

void SFS::compute_sumlnk (unsigned int vectorSize) {
	// Initializations
	vector<double> lnk (vectorSize, 0.);

	// Fill the sumlnk vector
	for (unsigned int i = 1; i <= vectorSize; i++) lnk[i] = log(i);
	for (unsigned int j = 1; j <= vectorSize; j++) _sumlnk[j] = _sumlnk[j - 1] + lnk[j];
}

/******************************************************************************/

/* lnCnk */
/* Returns the log of Cnk (binomial combinatory term) */
double SFS::lnCnk(int n, int k) {

	if (n <= 0 || k < 0) {
		ApplicationTools::displayWarning("lnCnk method - invalid parameters (n <= 0 or k < 0)");
		return -1.;
	}

	if (k > n) {
		ApplicationTools::displayWarning("lnCnk method - invalid parameters (k > n)");
		return -1.;
	}

	if (k == 0) return 0;

	return _sumlnk[n] - _sumlnk[k] - _sumlnk[n - k];
}

/******************************************************************************/

/* hypergeometric_prob */
/* Returns the probability of randomly drawing exactly k white balls */
/* when sampling n balls in an urn containing pa white balls out of tot. */
/* sumln must be such that sumln[i] = sum of ln(x) for x in [1..i] inclusive. */
/* (with sumln allocated and defined for i up to tot, inclusive) */
double SFS::hypergeometric_prob(int k, int n, int pa, int tot) {

	double lnp, t1, t2, t3;

	if (n < 0) {
		ApplicationTools::displayError("Hypergeometric_prob method - n < 0");
		exit(EXIT_FAILURE);
	}

	if (n > tot) {
		ApplicationTools::displayError("Hypergeometric_prob method - n > tot");
		exit(EXIT_FAILURE);
	}

	if (pa < 0) {
		ApplicationTools::displayError("Hypergeometric_prob method - pa < 0");
		exit(EXIT_FAILURE);
	}

	if (pa > tot) {
		ApplicationTools::displayError("Hypergeometric_prob method - pa > tot");
		exit(EXIT_FAILURE);
	}

	if (tot < 0) {
		ApplicationTools::displayError("Hypergeometric_prob method - tot < 0");
		exit(EXIT_FAILURE);
	}

	if (k < 0 || k > n || k > pa) return 0.;
	if (n-k > tot-pa) return 0.;

	t1 = lnCnk(pa, k);
	t2 = lnCnk(tot - pa, n-k);
	t3 = lnCnk(tot, n);

	lnp = t1 + t2-t3;

	return exp(lnp);
}


/*--------------------------------*/
/*            Accessor            */
/*--------------------------------*/

// SET
void SFS::change_sample_size(unsigned int newSampleSize) {
	_sample_size = newSampleSize;
}


// GET
unsigned int SFS::get_sample_size() {
	return _sample_size;
}

vector <double> SFS::get_Hypergeometric_draw() {
	return _Hypergeometric_draw;
}


/*------------------------------------*/
/*           Public methods           */
/*------------------------------------*/

/*       Matrix folding method       */

vector < vector<double> > SFS::fold_square_matrix(vector< vector<double> >& squareMatrix, unsigned int matrixSideSize) {
	// Initializations
	vector < vector<double> > foldedMatrix (matrixSideSize + 1, vector<double> (matrixSideSize + 1, 0));

	// Fold
	for (unsigned int j = 0; j <= matrixSideSize; j++) {

		for (unsigned int i = 0; i <= j / 2; i++) {
			foldedMatrix[j][i] = squareMatrix[j][i] + squareMatrix[j][j-i];
		}

		if (j%2 == 0) {
			foldedMatrix[j][j / 2] -= squareMatrix[j][j / 2];
		}
	}

	return foldedMatrix;
}


/*       Hypergeometric drawing methods       */

void SFS::folded_hypergeometric_drawing(vector < vector <double> >& foldedMatrix, unsigned int matrixSideSize) {

	// Hyper geometric drawing
	for (unsigned int ii = _sample_size + 1; ii <= matrixSideSize; ii++) {
		for (unsigned int k = 1; k <= _sample_size / 2; k++) {
			for (unsigned int i = k; i <= ii / 2; i++) {
				foldedMatrix[_sample_size][k] += foldedMatrix[ii][i] * (hypergeometric_prob(k, _sample_size, i, ii) + hypergeometric_prob(_sample_size - k, _sample_size, i, ii));
			}
		}
		if (_sample_size%2 == 0) {
			for (unsigned int i = _sample_size / 2; i <= ii / 2; i++) {
				foldedMatrix[_sample_size][_sample_size / 2] -= foldedMatrix[ii][i] * hypergeometric_prob(_sample_size / 2, _sample_size, i, ii);
			}
		}
	}

	// Extract the line of interest of each matrix and store them as simple vectors
	_Hypergeometric_draw = foldedMatrix[_sample_size];
}

/******************************************************************************/

void SFS::unfolded_hypergeometric_drawing(vector < vector <double> >& unfoldedMatrix, unsigned int matrixSideSize) {

	// Hyper geometric drawing
	for (unsigned int ii = _sample_size + 1; ii <= matrixSideSize; ii++) {
		for (unsigned int k = 1; k < _sample_size; k++) {
			for (unsigned int i = k; i < ii; i++) {
				unfoldedMatrix[_sample_size][k] += unfoldedMatrix[ii][i] * hypergeometric_prob(k, _sample_size, i, ii);
			}
		}
	}

	// Extract the line of interest and store it in a simple vector
	_Hypergeometric_draw = unfoldedMatrix[_sample_size];
}
