/*
 * SelectionIndicators.cpp
 *
 *  Created on: 22 oct. 2012
 *      Author: abernard
 */

#include "SelectionIndicators.h"

// Bio++ classes
#include <Bpp/Seq/Container/SiteContainerTools.h>
#include <Bpp/PopGen/SequenceStatistics.h>
#include <Bpp/PopGen/PolymorphismSequenceContainerTools.h>

// Namespaces
using namespace std;
using namespace bpp;
using namespace ppc;

/*--------------------------------*/
/*          Constructors          */
/*--------------------------------*/

// Constructor that take two full PolymorphismSequenceContainer (one for the ingroup and one for the outgroup) as input
SelectionIndicators::SelectionIndicators(PolymorphismSequenceContainer* ingroupContainer, PolymorphismSequenceContainer* outgroupContainer, const StandardCodonAlphabet* codonAlphabet,
		string ingroupConsensusName, string outgroupConsensusName): _nN(0.), _nS(0.), _fixN(0.), _fixS(0.), _fixSnS_ratio(0.),
		_number_of_sequences(0), _number_of_sites(0), _number_of_complete_sites(0), _consensus(0) {

	const Sequence* inseq = SiteContainerTools::getConsensus(*ingroupContainer, ingroupConsensusName, true ,false);
	const Sequence* outseq = SiteContainerTools::getConsensus(*outgroupContainer, outgroupConsensusName, true ,false);

	// Create a new global container with consensus sequences
	_consensus = new PolymorphismSequenceContainer(codonAlphabet);
	_consensus->addSequence(*inseq);
	_consensus->addSequence(*outseq);
}

/******************************************************************************/

// Constructor that take two consensus sequence (one for the ingroup and one for the outgroup) as input
SelectionIndicators::SelectionIndicators(const Sequence* ingroupConsensusSequence, const Sequence* outgroupConsensusSequence, const StandardCodonAlphabet* codonAlphabet):
		_nN(0.), _nS(0.), _fixN(0.), _fixS(0.), _fixSnS_ratio(0.), _number_of_sequences(0), _number_of_sites(0),  _number_of_complete_sites(0), _consensus(0) {

	// Create a new global container with consensus sequences
	_consensus = new PolymorphismSequenceContainer(codonAlphabet);
	_consensus->addSequence(*ingroupConsensusSequence);
	_consensus->addSequence(*outgroupConsensusSequence);
}

/******************************************************************************/

// Constructor that take only one PolymorphismSequenceContainer (ingroup) as input
SelectionIndicators::SelectionIndicators(PolymorphismSequenceContainer* ingroupContainer, const StandardCodonAlphabet* codonAlphabet,
		string ingroupConsensusName): _nN(0.), _nS(0.), _fixN(0.), _fixS(0.), _fixSnS_ratio(0.), _number_of_sequences(0), _number_of_sites(0),
		_number_of_complete_sites(0), _consensus(0) {

	const Sequence* inseq = SiteContainerTools::getConsensus(*ingroupContainer, ingroupConsensusName, true ,false);

	// Create a new global container with consensus sequences
	_consensus = new PolymorphismSequenceContainer(codonAlphabet);
	_consensus->addSequence(*inseq);
}


/*--------------------------------*/
/*           Destructor           */
/*--------------------------------*/

SelectionIndicators::~SelectionIndicators() {
	// Delete the PolymorphismSequenceContainer
	delete _consensus;
	_consensus = 0;
}


/*---------------------------------*/
/*            Accessors            */
/*---------------------------------*/

// GET
double SelectionIndicators::get_nN() { return _nN; }
double SelectionIndicators::get_nS() { return _nS; }
double SelectionIndicators::get_fixN() { return _fixN; }
double SelectionIndicators::get_fixS() { return _fixS; }
double SelectionIndicators::get_fixSnS_ratio() { return _fixSnS_ratio; }

unsigned int SelectionIndicators::get_number_of_sequences() { return _number_of_sequences; }
unsigned int SelectionIndicators::get_number_of_sites() { return _number_of_sites; }
unsigned int SelectionIndicators::get_number_of_complete_sites() { return _number_of_complete_sites; }

/*------------------------------------*/
/*           Public methods           */
/*------------------------------------*/

/*       Computing methods       */

void SelectionIndicators::determine_all_attributes(StandardGeneticCode* stdGeneticCode, double kappa) {

	determine_nN(stdGeneticCode, kappa);
	determine_nS(stdGeneticCode, kappa);

	determine_fixN(stdGeneticCode);
	determine_fixS(stdGeneticCode);
	determine_fixSnS_ratio();

	countSequencesAndSites();
}

/******************************************************************************/

void SelectionIndicators::determine_nN(StandardGeneticCode* stdGeneticCode, double kappa) {

	_nN = SequenceStatistics::meanNonSynonymousSitesNumber(*_consensus, *stdGeneticCode, kappa);
}

/******************************************************************************/

void SelectionIndicators::determine_nS(StandardGeneticCode* stdGeneticCode, double kappa) {

	_nS = SequenceStatistics::meanSynonymousSitesNumber(*_consensus, *stdGeneticCode, kappa);
}


void SelectionIndicators::determine_fixN(StandardGeneticCode* stdGeneticCode) {

	if (_consensus->getNumberOfSequences() > 1) {
		_fixN = SequenceStatistics::nonSynonymousSubstitutionsNumber(*_consensus, *stdGeneticCode);
	}
}

/******************************************************************************/

void SelectionIndicators::determine_fixS(StandardGeneticCode* stdGeneticCode) {

	if (_consensus->getNumberOfSequences() > 1) {
		_fixS = SequenceStatistics::synonymousSubstitutionsNumber(*_consensus, *stdGeneticCode);
	}
}

/******************************************************************************/

void SelectionIndicators::determine_fixSnS_ratio() {

	if (_fixS > 0. && _nS > 0.) {
		_fixSnS_ratio = _fixS / _nS;
	}
}

/******************************************************************************/

void SelectionIndicators::countSequencesAndSites() {

	_number_of_sequences = _consensus->getNumberOfSequences();
	_number_of_sites = _consensus->getNumberOfSites();
	_number_of_complete_sites = PolymorphismSequenceContainerTools::getNumberOfCompleteSites(*_consensus, false);
}
