/*
 * AlignmentFileManagement.cpp
 *
 *  Created on: 19 oct. 2012
 *      Author: abernard
 */

#include "AlignmentFileManagement.h"

// Bio++ classes
#include <Bpp/Text/StringTokenizer.h>
#include <Bpp/Seq/Io/Fasta.h>
#include <Bpp/Seq/Alphabet/AlphabetTools.h>

// PopPhyl Project Developments
#include "PolymorphismSequenceContainerAddons.h"
#include "../../Seq/Alphabet/CustomNucleicAlphabet.h"
#include "../../Seq/Alphabet/FrameshiftCodonAlphabet.h"

using namespace std;
using namespace bpp;
using namespace ppc;

/*--------------------------------*/
/*           Destructor           */
/*--------------------------------*/

AlignmentFileManagement::~AlignmentFileManagement() {
	// TODO Auto-generated destructor stub
}


/*---------------------------------*/
/*         Static methods          */
/*---------------------------------*/

/*------------------------------------------------------------------------------------------------------------*/
/*    Methods to read an alignment file and convert it into a vector of bpp::PolymorphismSequenceContainer    */
/*------------------------------------------------------------------------------------------------------------*/

std::string AlignmentFileManagement::_extractContigNameFromSequenceName (string full_sequence_name) {

	// Split the full_sequence_name to collect the ContigName (PopPhyl) or Sequence name
	StringTokenizer ID_tokens = StringTokenizer(full_sequence_name, "|");
	return ID_tokens.getToken(0);
}

/******************************************************************************/

// Convert Alignment file into a Vector of PSC
vector<PolymorphismSequenceContainer*> AlignmentFileManagement::readAlignmentFile (string seq_file_name, bool removeFrameshift, const NucleicAlphabet* nucleoAlphabet454, const StandardCodonAlphabet* codonAlphabet, bool debugMode) {

	// Initializations
	vector<PolymorphismSequenceContainer*> multidata;
	AlignedSequenceContainer* temporaryContainer = 0;
	bool keepDataset(true);
	string datasetName("");
	unsigned int numberOfSequence(0), nbDataset(0), nbDatasetWithFrameshift(0);

	// Read the input file to build an Ordered Sequence Container (Nucleic alphabet)
	Fasta * seqReader = new Fasta();
	OrderedSequenceContainer* allNucleoSequences = dynamic_cast<OrderedSequenceContainer*>(seqReader->read(seq_file_name, nucleoAlphabet454));

	numberOfSequence = allNucleoSequences->getNumberOfSequences();
	cout << " -> Initial number of sequences: " << numberOfSequence << endl;

	// Browse the list of sequence to detect framehift-containing contig
	for (unsigned int i = 0; i < numberOfSequence; i++) {
		// Get the contig name
		string currentContigName = _extractContigNameFromSequenceName(allNucleoSequences->getName(i));

		// Add the temporary container (converted to codon alphabet) to the list of container if needed
		if (currentContigName != datasetName) {
			nbDataset++;

			// Convert the temporary AlignedSequenceContainer (nucleotide alphabet) into a valid PolymorphismSequenceContainer (codon alphabet)
			if (temporaryContainer) {
				OrderedSequenceContainer* destinationContainer = new AlignedSequenceContainer(codonAlphabet);
				SequenceContainerTools::convertAlphabet(*temporaryContainer, *destinationContainer);
				PolymorphismSequenceContainer* validDataset = new PolymorphismSequenceContainer(*destinationContainer);
				multidata.push_back(validDataset);
			}

			// Update or reinitialize variables
			datasetName = currentContigName;
			keepDataset = true;
			temporaryContainer = new AlignedSequenceContainer(nucleoAlphabet454);
		}

		// Look for frameshift if needed (if we haven't changed of dataset and if the previous sequences doesn't have any frameshift)
		if (keepDataset) {
			bool hasFrameshift = 0;

			// Look at each position of the sequence to check frameshift presence if requested
			if (removeFrameshift) {
				for(unsigned int j = 0; j < allNucleoSequences->getSequence(i).size(); j++) {
					if (allNucleoSequences->getSequence(i)[j] == 14) {
						hasFrameshift = 1;
						if (debugMode) cout << "Contig " << currentContigName << " has a frameshift (Position: " << j << ")" << endl;
						break;
					}
				}
			}

			// Add the current sequence to the container (no frameshift) OR delete the container (frameshift detected)
			if (hasFrameshift) {
				keepDataset = false;
				if (debugMode) cout << "   --> Frameshift detected ! The current alignment will be discarded.." << endl;
				nbDatasetWithFrameshift++;

				// Delete the object and move the pointer to a safe location
				delete temporaryContainer;
				temporaryContainer = 0;
			} else {
				temporaryContainer->addSequence(allNucleoSequences->getSequence(i));
			}
		}
	}

	// Deal with the last dataset
	if (temporaryContainer) {
		OrderedSequenceContainer* destinationContainer = new AlignedSequenceContainer(codonAlphabet);
		SequenceContainerTools::convertAlphabet(*temporaryContainer, *destinationContainer);
		PolymorphismSequenceContainer* validDataset = new PolymorphismSequenceContainer(*destinationContainer);
		multidata.push_back(validDataset);
	}

	// Some informative data
	cout << " -> Initial number of datasets/contigs: " << nbDataset << endl << endl;
	if (removeFrameshift) cout << " -> Number of conserved contigs: " << multidata.size() << endl << "   --> Number of eliminated frameshift-containing contigs: " << nbDatasetWithFrameshift << endl << endl;

	return multidata;
}

//static void convertAlphabet(const SequenceContainer& seqCont, SequenceContainer& outputCont) throw (Exception) {
//std::vector<std::string> seqNames = seqCont.getSequencesNames();
//bool checkNames = outputCont.getNumberOfSequences() > 0;
//for (size_t i = 0; i < seqNames.size(); i++)
//{
//BasicSequence seq(seqNames[i], seqCont.toString(seqNames[i]), outputCont.getAlphabet());
//outputCont.addSequence(seq, checkNames);
//}
//}

/*------------------------------------------------------------------*/
/*      Tools for vector of bpp::PolymorphismSequenceContainer      */
/*------------------------------------------------------------------*/

unsigned int AlignmentFileManagement::getMaxNumberOfFocalIndividual (vector<PolymorphismSequenceContainer*> const & allData, string ingroupSpeciesName) {

	// Initializations
	unsigned int NbFocal(0), maxNbFocal(0), minNbFocal(1000000);

	// Loop through the contigs and get the maximum numùber of focal individuals
	for (unsigned int i = 0; i < allData.size(); i++) {
		NbFocal = 0;

		vector<string> names = allData[i]->getSequencesNames();

		for (unsigned int j = 0; j < names.size(); j++) {
			if (names[j].find(ingroupSpeciesName) != string::npos) NbFocal++;
		}

		// Update Min and Max if needed
		if (NbFocal > maxNbFocal) maxNbFocal = NbFocal;
		if (NbFocal < minNbFocal) minNbFocal = NbFocal;
	}

	cout << " -> Minimum/Maximum number of focal sequences by dataset/contig:" << endl;
	cout << "   --> Min: " << minNbFocal << endl;
	cout << "   --> Max: " << maxNbFocal << endl << endl;

	return maxNbFocal;
}

/******************************************************************************/

/* Important note: there is a bug in Bio++ bpp::PolymorphismSequenceContainer::removeSequence method:
   The vector that store the group identifier of each sequence is not updated when a sequence is deleted
   This bug has been submitted on Bio++ Mantis bug tracker in november 2012
   Direct link: http://biopp.univ-montp2.fr/bug/view.php?id=67

   Currently, to avoid crashes, groupId must be refreshed after the execution of each method that suppress sequences !
*/
vector<PolymorphismSequenceContainer*> AlignmentFileManagement::cleanMultiContainer (vector<PolymorphismSequenceContainer*> const & brutMultiContainer,
		const StandardCodonAlphabet* codonAlphabet,	string ingroupSpeciesName, unsigned int minNumberOfCodon, unsigned int toleranceZone,
		bool allowInternalBadCodons, unsigned int minGapNSite, double maxGapNSeq, bool debugMode, string outgroupSpeciesName) {

	vector<PolymorphismSequenceContainer*> cleanedMultiContainer;

	for (unsigned int i = 0; i < brutMultiContainer.size(); i++) {
		// Initializations
		PolymorphismSequenceContainer* dataset = brutMultiContainer[i];

		// Tag each sequence as ingroup or outgroup member and give them a groupID (Groups: ingroup(0), outgroup(1), other(2))
		PolymorphismSequenceContainerAddons::tagAllSequences(dataset, ingroupSpeciesName, debugMode, outgroupSpeciesName);

		// Recover the number of sequence for each group
		map<unsigned int, unsigned int> initialGroupCounters = PolymorphismSequenceContainerAddons::getNumberOfMembers(dataset, debugMode);

		// Extract the current contig name from the first full sequence identifier
		string currentContigName = PolymorphismSequenceContainerAddons::getContigNameFromContainer(dataset, 0);

		cout << " -> Dataset " << i + 1 << " (" << currentContigName << "):" << endl;

		if (initialGroupCounters.size() > 0) {

			// Display dataset informations
			cout << "   --> Original number of sequences : " << dataset->getNumberOfSequences() << " sequence(s) (Ingroup: " << initialGroupCounters[0]  << ", Outgroup: " << initialGroupCounters[1] << ", Other: " << initialGroupCounters[2] << ")" << endl;
			cout << "   --> Original number of sites: " << dataset->getNumberOfSites()  << " site(s)" << endl;

			// Remove all sequences that belong to the Other group if needed
			if (initialGroupCounters[2] > 0) {
				PolymorphismSequenceContainerAddons::removeGroupFromContainer(dataset, 2, debugMode);

				// Hotfix for removeSequence method bug
				PolymorphismSequenceContainerAddons::tagAllSequences(dataset, ingroupSpeciesName, debugMode, outgroupSpeciesName);
			}

			// Deal with internal and "external" stop codons
			bool rejectDataset = PolymorphismSequenceContainerAddons::dealWithBadCodons(dataset, codonAlphabet, toleranceZone, allowInternalBadCodons, debugMode);

			if (rejectDataset) {
				cout << "   --> Warning: At least one internal Stop/Frameshift codon has been detected ! Current dataset won't be conserved !" << endl << endl;
				continue;
			}

			// Remove sites or sequences with too much gap or "N" base
			PolymorphismSequenceContainerAddons::removeHighlyGappedSites(dataset, codonAlphabet, initialGroupCounters[0], minGapNSite, debugMode);
			PolymorphismSequenceContainerAddons::removeHighlyGappedSequences(dataset, codonAlphabet, maxGapNSeq, debugMode);

			// Hotfix for removeSequence method bug
			PolymorphismSequenceContainerAddons::tagAllSequences(dataset, ingroupSpeciesName, debugMode, outgroupSpeciesName);

			if (dataset->getNumberOfSequences() <= 1 || dataset->getNumberOfSites() < minNumberOfCodon) {
				cout << "   --> Warning: Not enough data available after cleaning (stop, gap and N) ! Current dataset won't be conserved !" << endl << endl;
			} else {
				cout << "   --> Number of conserved sequences and sites after cleaning (stop, gap and N): " << dataset->getNumberOfSequences() << " sequence(s), " << dataset->getNumberOfSites()  << " site(s)" << endl;

				// Get the number of sequence of each group after cleaning
				map<unsigned int, unsigned int> groupCountersAfterCleaning = PolymorphismSequenceContainerAddons::getNumberOfMembers(dataset, debugMode);

				if (groupCountersAfterCleaning[0] == 0) {
					cout << "   --> Warning: There is no more ingroup's sequence ! Current dataset won't be conserved !" << endl << endl;
					continue;
				} else if (groupCountersAfterCleaning[1] == 0 && outgroupSpeciesName != "none") {
					cout << "   --> Warning: There is no more sequence of the specified outgroup species (" << outgroupSpeciesName << ") ! Current dataset won't be conserved !" << endl << endl;
					continue;
				} else {
					cout << endl;
				}

				cleanedMultiContainer.push_back(dataset);
			}

		} else {
			// dataset is not conserved if there is no sequence for the ingroup species
			cout << "   --> Warning: Current dataset won't be conserved because there is no ingroup's sequence !" << endl << endl;
			continue;
		}
	}

	return cleanedMultiContainer;
}
