/*
 * PolymorphismSequenceContainerAddons.cpp
 *
 *  Created on: 12 oct. 2012
 *      Author: abernard
 */

#include "PolymorphismSequenceContainerAddons.h"

// Basic C++ classes
#include <set>

// PopPhyl Project Developments
#include "../../Seq/Alphabet/CustomNucleicAlphabet.h"
#include "../../Seq/Alphabet/FrameshiftCodonAlphabet.h"

// Bio++ classes
#include <Bpp/App/ApplicationTools.h>

using namespace std;
using namespace bpp;
using namespace ppc;

/*--------------------------------*/
/*           Destructor           */
/*--------------------------------*/

PolymorphismSequenceContainerAddons::~PolymorphismSequenceContainerAddons() {
	// TODO Auto-generated destructor stub
}


/*------------------------------------------*/
/*         Public & Static methods          */
/*------------------------------------------*/

bpp::Site PolymorphismSequenceContainerAddons::generateValidUnknownSite(PolymorphismSequenceContainer* currentContainer, const StandardCodonAlphabet* codonAlphabet) {

	// Get the unknown character code for the selected codon alphabet
	int unknownCharacterCode = codonAlphabet->getUnknownCharacterCode();

	// Prepare the content of the site
	vector<int> default_codon;
	for (unsigned int i = 0; i<currentContainer->getNumberOfSequences(); i++) {
		default_codon.push_back(unknownCharacterCode);
	}

	// Build a new Site object with the specified alphabet. The content of the site is initialized from a vector of integers.
	bpp::Site unknownSite(default_codon, codonAlphabet);

	return unknownSite;
}

/******************************************************************************/

void PolymorphismSequenceContainerAddons::tagAllSequences (PolymorphismSequenceContainer* currentContainer, string ingroupSpeciesName, bool debugMode, string outgroupSpeciesName) {

	// Get the name of all sequences
	vector<string> allSequenceNames = currentContainer->getSequencesNames();

	// Check if the name of each sequence contains the name of the focal species and tag them depending on the result
	for (unsigned int i = 0; i < currentContainer->getNumberOfSequences(); i++) {

		// Get current species name
		string currentSpeciesName = getSpeciesNameFromContainer(currentContainer, i);

		// Tag and set group of each sequence
		// Groups: ingroup(0), outgroup(1), other(2)
		if (currentSpeciesName == ingroupSpeciesName) { // ingroup
			currentContainer->setAsIngroupMember(i);
			currentContainer->setGroupId(i, 0);
			if (debugMode) cout << "(DEBUG - tagAllSequences) Sequence: " << allSequenceNames[i] << " is part of the ingroup" << endl;

		} else { // outgroup
			currentContainer->setAsOutgroupMember(currentContainer->getName(i));

			if (currentSpeciesName == outgroupSpeciesName) {
				currentContainer->setGroupId(i, 1);
				if (debugMode) cout << "(DEBUG - tagAllSequences) Sequence: " << allSequenceNames[i] << " is part of the outgroup (main outgroup species)" << endl;
			} else {
				currentContainer->setGroupId(i, 2);
				if (debugMode) cout << "(DEBUG - tagAllSequences) Sequence: " << allSequenceNames[i] << " is part of the outgroup (other species)" << endl;
			}
		}
	}
}

/******************************************************************************/

map<unsigned int, unsigned int> PolymorphismSequenceContainerAddons::getNumberOfMembers (PolymorphismSequenceContainer* currentContainer, bool debugMode) {

	// Control
	if (currentContainer->getNumberOfGroups() < 1) ApplicationTools::displayError("There is no group associated to your container !");

	// Initializations
	map<unsigned int, unsigned int> counters;

	// Initialize keys
	set< unsigned int > groupIds = currentContainer->getAllGroupsIds();
	for(set<unsigned int>::iterator setIterator = groupIds.begin();  setIterator != groupIds.end(); setIterator++) counters[*setIterator] = 0;

	// Count the number of sequence in each group
	for (unsigned int i = 0; i < currentContainer->getNumberOfSequences(); i++) {

		unsigned int currentGroupID = currentContainer->getGroupId(i);
		counters[currentGroupID]++;

		if (debugMode) cout << "(DEBUG - getNumberOfMembers) Sequence " << currentContainer->getName(i) << " is a member of group " << currentGroupID << endl;
	}

	// Debug loop
	if (debugMode) {
		cout << "(DEBUG - getNumberOfMembers) Number of groups: " << currentContainer->getNumberOfGroups() << endl;

		for(map<unsigned int, unsigned int>::iterator it=counters.begin(); it!=counters.end(); ++it) {
			cout << "(DEBUG - getNumberOfMembers) Group " << it->first << ": " << it->second << " member(s)" << endl;
		}
	}

	return counters;
}

/******************************************************************************/

string PolymorphismSequenceContainerAddons::getContigNameFromContainer (PolymorphismSequenceContainer* currentContainer, unsigned int indexToUse) {

	if (indexToUse > currentContainer->getNumberOfSequences() - 1) {
		ApplicationTools::displayError("Selected index does not exists in the container !!");
		exit(1);

	} else if (currentContainer->getNumberOfSequences() > 0) {

		// Get the name of the sequence from the container
		string SequenceName = currentContainer->getName(indexToUse);

		// Split the fullname and extract the contig name
		StringTokenizer tokenizer(SequenceName, "|");

		// The contig name must be the first element
		return tokenizer.nextToken();

	} else {
		ApplicationTools::displayError("Selected container is empty ! Contig name can't be extracted !");
		exit(1);
	}
}

/******************************************************************************/

string PolymorphismSequenceContainerAddons::getSpeciesNameFromContainer (PolymorphismSequenceContainer* currentContainer, unsigned int indexToUse) {

	if (indexToUse > currentContainer->getNumberOfSequences() - 1) {
		ApplicationTools::displayError("Selected index does not exists in the container !!");
		exit(1);

	} else if (currentContainer->getNumberOfSequences() > 0) {

		// Get the name of the sequence from the container
		string SequenceName = currentContainer->getName(indexToUse);

		// Split the fullname and extract the species name
		StringTokenizer tokenizer(SequenceName, "|");

		// The species name must be the second element if PopPhyl sequence name format is respected
		if (tokenizer.numberOfRemainingTokens() > 1) {
			return tokenizer.getToken(1);
		} else {
			return tokenizer.nextToken();
		}

	} else {
		ApplicationTools::displayError("Selected container is empty ! Species name can't be extracted !");
		exit(1);
	}
}

/******************************************************************************/

string PolymorphismSequenceContainerAddons::getIndividualNameFromContainer (PolymorphismSequenceContainer* currentContainer, unsigned int indexToUse) {

	if (indexToUse > currentContainer->getNumberOfSequences() - 1) {
		ApplicationTools::displayError("Selected index does not exists in the container !!");
		exit(1);

	} else if (currentContainer->getNumberOfSequences() > 0) {

		// Get the name of the sequence from the container
		string SequenceName = currentContainer->getName(indexToUse);

		// Split the fullname and extract the individual name
		StringTokenizer tokenizer(SequenceName, "|");

		// The individual name must be the third element if PopPhyl sequence name format is respected
		if (tokenizer.numberOfRemainingTokens() > 1) {
			return tokenizer.getToken(2);
		} else {
			return tokenizer.nextToken();
		}

	} else {
		ApplicationTools::displayWarning("Selected container is empty ! Individual name can't be extracted !");
		exit(1);
	}
}

/******************************************************************************/

vector<string> PolymorphismSequenceContainerAddons::getNameElementsFromContainer (PolymorphismSequenceContainer* currentContainer, string separator, unsigned int indexToUse, unsigned int maxElementToRetrieve) {

	if (indexToUse > currentContainer->getNumberOfSequences() - 1) {
		ApplicationTools::displayError("Selected index does not exists in the container !!");
		exit(1);

	} else if (currentContainer->getNumberOfSequences() > 0) {

		// Initializations
		vector<string> nameElements;
		unsigned int nbElement(0);

		// Get the name of the sequence from the container
		string SequenceName = currentContainer->getName(indexToUse);

		// Split the fullname
		StringTokenizer tokenizer(SequenceName, separator);

		while (tokenizer.hasMoreToken()) {
			if (nbElement == maxElementToRetrieve) break;
			nameElements.push_back(tokenizer.nextToken());
			nbElement++;
		}

		return nameElements;

	} else {
		ApplicationTools::displayWarning("Selected container is empty ! Individual name can't be extracted !");
		exit(1);
	}
}

/******************************************************************************/

void PolymorphismSequenceContainerAddons::removeLeftBadSites (PolymorphismSequenceContainer* currentContainer, int nbSiteToRemoveOnTheLeft, bool debugMode) {

	if (nbSiteToRemoveOnTheLeft > -1) {
		if (debugMode) cout << "(DEBUG - removeLeftBadSites) Stop/Frameshift codon(s) have been detected in the tolerance zone (left side): ignoring first " << nbSiteToRemoveOnTheLeft << " codons" << endl;
		for (int k = 0; k < nbSiteToRemoveOnTheLeft; k++) {
			currentContainer->removeSite(0);
		}
	}
}

/******************************************************************************/

void PolymorphismSequenceContainerAddons::removeRightBadSites (PolymorphismSequenceContainer* currentContainer, int nbSiteToRemoveOnTheRight, bool debugMode) {

	if (nbSiteToRemoveOnTheRight > -1) {
		if (debugMode) cout << "(DEBUG - removeRightBadSites) Stop/Frameshift codon(s) have been detected in the tolerance zone (right side): ignoring last " << nbSiteToRemoveOnTheRight << " codons" << endl;
		unsigned int new_nb_site = currentContainer->getNumberOfSites();
		unsigned int remove_start = (new_nb_site -1) - nbSiteToRemoveOnTheRight; // We want to remove the last "nbSiteToRemoveOnTheRight" sites and the list of sites begin at 0 !

		for (unsigned int m = new_nb_site-1; m > remove_start; m--) {
			currentContainer->removeSite(m);
		}
	}
}

/******************************************************************************/

bool PolymorphismSequenceContainerAddons::dealWithBadCodons (PolymorphismSequenceContainer* currentContainer, const StandardCodonAlphabet* codonAlphabet, unsigned int toleranceZone, bool allowInternalBadCodons, bool debugMode) {

	// Initializations
	unsigned int originalNumberOfSequence = currentContainer->getNumberOfSequences();
	unsigned int originalNumberOfSite = currentContainer->getNumberOfSites();

	int ignore_left(-1), ignore_right(-1);
	int bad_codon_to_left(0), bad_codon_to_right(0);

	int left_limit = toleranceZone;
	int right_limit = originalNumberOfSite - left_limit;

	bool reject_dataset = false;

	// Creation of an "unknown" site
	bpp::Site unknownSite = generateValidUnknownSite(currentContainer, codonAlphabet);

	// For each sequence
	for (unsigned int i = 0; i < originalNumberOfSequence; i++) {
		// For each codon
		for (unsigned int j = 0; j < originalNumberOfSite; j++) {
			// Collect the int code and the string sequence for the current codon
			int codonIntCode = currentContainer->valueAt(i,j);
			string codonChar = codonAlphabet->intToChar(codonIntCode);

			// Current codon is a stop or a framsehift codon ?
			if (codonAlphabet->isStop(codonIntCode) || codonAlphabet->getName(codonChar) == "FrameShift") {
				bad_codon_to_left = j + 1; // j+1 because the for loop start at 0
				bad_codon_to_right = originalNumberOfSite - (j + 1); // j+1 because the for loop start at 0

				if (bad_codon_to_left <= left_limit) {
					if (debugMode) cout << "(DEBUG - dealWithBadCodons) " << codonAlphabet->getName(codonChar) << " codon (Int Code: " << codonIntCode << ") detected in the tolerance zone (left side) ! (Position: " << bad_codon_to_left << " - Sequence: " << i+1 << ")" << endl;
					if (bad_codon_to_left > ignore_left) {
						ignore_left = bad_codon_to_left;
					}
				} else if (bad_codon_to_left > right_limit) { // Why >: if the sequence contains 220 sites and the tolerance zone (tz) is set to 20 sites then position 200 MUST NOT be in the tz (tz span from position 220 to 201 included = 20 positions)
					if (debugMode) cout << "(DEBUG - dealWithBadCodons) " << codonAlphabet->getName(codonChar) << " codon (Int Code: " << codonIntCode << ") detected in the tolerance zone (right side) ! (Position: " << bad_codon_to_left  << " - Sequence: " << i+1 << ")" << endl;
					if (bad_codon_to_right+1 > ignore_right){
						ignore_right = bad_codon_to_right+1; //Why +1: if sequence contains 317 sites and there is a stop codon at position 314 --> We want to remove 4 sites (314,315,316,317)
					}
					break;

				} else {
					if (debugMode) cout << "(DEBUG - dealWithBadCodons) " << codonAlphabet->getName(codonChar) << " codon (Int Code: " << codonIntCode << ") detected in the core part of the sequence ! (Position: " << bad_codon_to_left << " - Sequence: " << i+1 << ")" << endl;

					if (allowInternalBadCodons) {
						// When internal stop/frameshift codons are tolerated we replace the site by an unknown site (NNN)
						currentContainer->setSite(j, unknownSite, false);
					} else {
						// When internal stop/frameshift codons are forbidden the current dataset is rejected
						reject_dataset = true;
					}
				}
			}
		}
	}

	// If internal stop codons are allowed and have been replaced we deal with stop codons located in the tolerance zone
	if (!reject_dataset) {
		removeLeftBadSites(currentContainer, ignore_left, debugMode);
		removeRightBadSites(currentContainer, ignore_right, debugMode);

//		if (DEBUG) cout << "*  --> Number of conserved sequences and sites after stop removal: " << currentContainer->getNumberOfSequences() << " sequence(s), " << currentContainer->getNumberOfSites()  << " site(s)" << endl;
	}

	return reject_dataset;
}

/******************************************************************************/

// Sites: gaps+N - Remove sites with too much gap or N base
void PolymorphismSequenceContainerAddons::removeHighlyGappedSites (PolymorphismSequenceContainer* currentContainer, const StandardCodonAlphabet* codonAlphabet, unsigned int numberOfIngroupSequence, unsigned int minGapNSite, bool debugMode) {

	// Note: we only look at ingroup's sequences to decide to remove a site or not
	for (int j = currentContainer->getNumberOfSites() - 1; j >= 0; j--) {
		int nbGapN(0);

		for (unsigned int i = 0; i < currentContainer->getNumberOfSequences(); i++) {
			if(currentContainer->isIngroupMember(currentContainer->getName(i))) {
				if (codonAlphabet->isGap(currentContainer->valueAt(i,j)) || codonAlphabet->isUnresolved(currentContainer->valueAt(i,j))) {
					nbGapN++;
				}
			}
		}

		if ((numberOfIngroupSequence - nbGapN) < minGapNSite) { // minGapNSite is an integer
			if (debugMode) cout << "(DEBUG - removeHighlyGappedSites) Site number " << j << " will be removed (" << nbGapN << " sequence(s) with too much gaps or N bases)" << endl;
			currentContainer->removeSite(j);
		}
	}
}

/******************************************************************************/

// Sequences: gaps+N - Remove sequences with a too high percentage of gap or N base
void PolymorphismSequenceContainerAddons::removeHighlyGappedSequences (PolymorphismSequenceContainer* currentContainer, const StandardCodonAlphabet* codonAlphabet, double maxGapNSeq, bool debugMode) {

	for (int i = currentContainer->getNumberOfSequences() - 1; i >= 0; i--) {
		unsigned int nbgapN = 0.;

		for (unsigned int j = 0; j < currentContainer->getNumberOfSites(); j++) {
			if (codonAlphabet->isGap(currentContainer->valueAt(i,j)) || codonAlphabet->isUnresolved(currentContainer->valueAt(i,j))) {
				nbgapN++;
			}
		}

		double gapN_percentage = (double)nbgapN / (double)currentContainer->getNumberOfSites();
		if (gapN_percentage > maxGapNSeq) { // maxGapNSeq is a float and a percentage
			if (debugMode) cout << "(DEBUG - removeHighlyGappedSequences) Sequence " << currentContainer->getName(i) << " will be removed (" << gapN_percentage*100 << " % of gaps or N bases)" << endl;
			currentContainer->removeSequence(i);
		}
	}
}

/******************************************************************************/

// Remove all sequences with a given group identifier from the input container
void PolymorphismSequenceContainerAddons::removeGroupFromContainer (PolymorphismSequenceContainer* currentContainer, unsigned int groupId, bool debugMode) {

	for (int i = currentContainer->getNumberOfSequences() - 1; i >= 0; i--) {

		if (currentContainer->getGroupId(i) == groupId) {
			if (debugMode) cout << "(DEBUG - removeGroupFromContainer) Sequence " << currentContainer->getName(i) << " will be removed (Member of group " << groupId << ")" << endl;
			currentContainer->removeSequence(i);
		}
	}
}

