/*
 * MappingContainer.h
 *
 *  Created on: 15 déc. 2010
 *      Author: V. Cahais
 */

#ifndef MAPPINGCONTAINER_H_
#define MAPPINGCONTAINER_H_

// Basic C++ classes
#include <iostream>
#include <vector>
#include <set>
#include <map>

// Bio++ classes
#include <Bpp/Seq/Sequence.h>
#include <Bpp/Seq/SequenceTools.h>
#include <Bpp/Seq/Alphabet/DNA.h>
#include <Bpp/Text/StringTokenizer.h>

namespace ppc {

	class MappingContainer {

		public:

			class Vertex {

				private:
					unsigned int order_;
					char nucleotide_;
					std::vector<unsigned int> sequences_;
					std::vector<MappingContainer::Vertex*> childrens_;
					MappingContainer::Vertex* references_;
					MappingContainer::Vertex* consensus_;

					bool isInsertion_;
					bool isReference_;

				public:
					Vertex(){isInsertion_=false;isReference_=false;consensus_=0;references_=0;order_=0;}
					virtual ~Vertex();

					void setOrder(unsigned int order){order_ = order;}
					unsigned int getOrder(){return order_;}

					void setNucleotide(char c){nucleotide_ = c;}
					char getNucleotide(){return nucleotide_;}

					void setToInsertion(){isInsertion_ = true;}
					void setToNotAnInsertion(){isInsertion_ = false;}
					bool isInsertion(){return isInsertion_;}

					void setToReference(){isReference_ = true;}
					bool isReference(){return isReference_;}

					void addSequence(int ind){sequences_.push_back(ind);}
					unsigned int getSequenceIndex(unsigned int i){return sequences_[i];}
					unsigned int getNumberOfSequences(){return sequences_.size();}
					bool isBaseOfSequence(unsigned int index);

					void addChild(MappingContainer::Vertex* v){childrens_.push_back(v);}
					void removeAllChildren(){childrens_.clear();}
					unsigned int getNumberOfChildren(){return childrens_.size();}
					MappingContainer::Vertex* getChildByIndex(unsigned int ind){return childrens_[ind];}
					MappingContainer::Vertex* getChildByNucleotide(char c);

					void setReference(MappingContainer::Vertex* v){references_ = v;}
					MappingContainer::Vertex* getReference(){return references_;}

					void setConsensus(MappingContainer::Vertex* v){consensus_ = v;}
					MappingContainer::Vertex* getConsensus(){return consensus_;}
			};

		private:

			struct sequence	{
				unsigned int index;
				std::string name;
				MappingContainer::Vertex* begin;
			};

			MappingContainer::Vertex* first_;
			std::map<std::string, sequence> sequences_;
			std::map<unsigned int, sequence*> idIndex;
			bpp::Alphabet * alpha;
			std::vector<MappingContainer::Vertex*> all_;
			std::map<unsigned int, MappingContainer::Vertex*> refpos_; // positions on the reference sequence

		public:

			bool printReference_;

			MappingContainer(bool printReference);
			virtual ~MappingContainer();
			void deleteNode(MappingContainer::Vertex * v);
			std::string refseq_;
			std::string refname_;
			unsigned int refseq_size_;

			MappingContainer::Vertex* getFirst() {return first_;}
			MappingContainer::Vertex* getNodeByRefPos(unsigned int pos, unsigned int length);

			//Sequences_
			unsigned int getNumberOfSequences(){return sequences_.size();}
			unsigned int getSequenceIndex(std::string name){return sequences_[name].index;}
			std::string getSequenceName(unsigned int index){return idIndex[index]->name;}
			MappingContainer::Vertex* getSequenceBegin(std::string name){return sequences_[name].begin;}
			MappingContainer::Vertex* getSequenceBegin(unsigned int index){return idIndex[index]->begin;}
			unsigned int registerSequence(std::string name,MappingContainer::Vertex * v);

			MappingContainer::Vertex* getVertexByPosition(unsigned int pos);

			void setRefSeq(std::string & name, std::string & seq);
			MappingContainer::Vertex* addMatch(MappingContainer::Vertex * curr, char nuc, unsigned int index);
			MappingContainer::Vertex* addInsertion(MappingContainer::Vertex * curr, char nuc, unsigned int index);

			//void setOrder(MappingContainer::Vertex* curr);
			//void setLocalOrder(MappingContainer::Vertex* curr, MappingContainer::Vertex* ref);
			//void buildReference(MappingContainer::Vertex* curr);

			std::string getSequence(std::string name);
			void toMultiFasta(std::ostream & out);
			void getNextNucleotide(MappingContainer::Vertex* curr, unsigned int index, std::string & seq);

			void toAln(std::ostream & out, std::ostream & fasta);
			std::string writeMonoMorphSite(MappingContainer::Vertex* curr, std::vector<std::string> & indiv, std::vector<std::string> & indivBySeqID);
			std::string writePolyMorphSite(MappingContainer::Vertex* curr, std::vector<std::string> & indiv, std::vector<std::string> & indivBySeqID, char nuc);
			std::string writeSpecialSite(MappingContainer::Vertex* curr, std::vector<std::string> & indiv, std::vector<std::string> & indivBySeqID);

			void indivBySequenceID(std::vector<std::string> & indiv, std::vector<std::string> & indivBySeqID);
			char getMajority(MappingContainer::Vertex* curr);

			void printTree(MappingContainer::Vertex* curr);
			void  printIdIndex();
	};

} //end of namespace PopPhyl Commons (ppc).

#endif /* MAPPINGCONTAINER_H_ */
