/*
 * ParametersManagement.cpp
 *
 *  Created on: 26 sept. 2012
 *      Author: abernard
 */

#include "./ParametersManagement.h"

// Basic C++ classes
#include <stdlib.h>
#include <iostream>
#include <fstream>

// Bio++ - Parameters Management
#include <Bpp/App/ApplicationTools.h>
#include <Bpp/Utils/AttributesTools.h>

using namespace std;
using namespace bpp;

/*-------------------------------*/
/*          Constructor          */
/*-------------------------------*/

ParametersManagement::ParametersManagement(const double versionNumber, const string versionDate, int nbArguments, char **allArguments) {

	// Set version related attributes
	_versionNumber = versionNumber;
	_versionDate = versionDate;

	// Set default values
	Default_execution_mode = "local";

	Default_min_gapN_site = 10;
	Default_max_gapN_seq = 0.5;
	Default_tolerance_zone = 20;
	Default_DFEa_write_start = 5;
	Default_DFEa_min_NS_SNP = 10;

	Default_min_nb_codon = 10;
	Default_sample_size = 12;
	Default_kappa = 1.;

	Default_remove_frameshift = false;
	Default_allow_internal_bc = false;
	Default_DEBUG = false;

	//Check minimum required number of parameters
	if (nbArguments < 2) displayHelp();

	// Store all arguments of the command line into a map of string
	map<string, string> cmdParams = AttributesTools::getAttributesMap(AttributesTools::getVector(nbArguments, allArguments), "=");

	// File parameters
	_alignement_file = ApplicationTools::getAFilePath("-alignment_file", cmdParams, true, true, "", true);

	// String parameters
	_ingroup_species_name = ApplicationTools::getStringParameter("-ingroup", cmdParams, "none", "", false, false);
	_outgroup_species_name = ApplicationTools::getStringParameter("-outgroup", cmdParams, "none", "", false, false);
	_execution_mode = ApplicationTools::getStringParameter("-exec_mode", cmdParams, Default_execution_mode, "", false, false);

	// Numeric parameters
	_min_gapN_site = ApplicationTools::getIntParameter("-gapN_site", cmdParams, Default_min_gapN_site, "", false, true);
	_max_gapN_seq = ApplicationTools::getDoubleParameter("-gapN_seq", cmdParams, Default_max_gapN_seq, "", false, true);
	_tolerance_zone = ApplicationTools::getIntParameter("-tolerance_zone", cmdParams, Default_tolerance_zone, "", false, true);
	_DFEa_write_start = ApplicationTools::getIntParameter("-dfea_write_start", cmdParams, Default_DFEa_write_start, "", false, false);
	_DFEa_min_NS_SNP = ApplicationTools::getIntParameter("-dfea_min_NS_SNP", cmdParams, Default_DFEa_min_NS_SNP, "", false, false);
	_min_nb_codon = ApplicationTools::getIntParameter("-min_nb_codon", cmdParams, Default_min_nb_codon, "", false, true);
	_sample_size = ApplicationTools::getIntParameter("-sample_size", cmdParams, Default_sample_size, "", false, true);
	_kappa = ApplicationTools::getIntParameter("-kappa", cmdParams, Default_kappa, "", false, false);

	// Boolean parameters (Recall: 1 is true, 0 is false)
	_remove_frameshift = ApplicationTools::getBooleanParameter("-remove_frameshift", cmdParams, Default_remove_frameshift, "", false, true);
	_allow_internal_bc = ApplicationTools::getBooleanParameter("-allow_internal_bc", cmdParams, Default_allow_internal_bc, "", false, true);

	_DEBUG = ApplicationTools::getBooleanParameter("-debug_mode", cmdParams, Default_DEBUG, "", false, false);
}

/*--------------------------------*/
/*           Destructor           */
/*--------------------------------*/

ParametersManagement::~ParametersManagement() {
	// TODO Auto-generated destructor stub
}

/*--------------------------------*/
/*            Accessor            */
/*--------------------------------*/

// GET
string ParametersManagement::get_alignement_file() { return _alignement_file; }
string ParametersManagement::get_ingroup_species_name() { return _ingroup_species_name; }
string ParametersManagement::get_outgroup_species_name() { return _outgroup_species_name; }
string ParametersManagement::get_execution_mode() { return _execution_mode; }

unsigned int ParametersManagement::get_min_gapN_site() { return _min_gapN_site; }
double ParametersManagement::get_max_gapN_seq() { return _max_gapN_seq; }
unsigned int ParametersManagement::get_tolerance_zone() { return _tolerance_zone; }
unsigned int ParametersManagement::get_DFEa_write_start() { return _DFEa_write_start; }
unsigned int ParametersManagement::get_DFEa_min_NS_SNP() { return _DFEa_min_NS_SNP; }

unsigned int ParametersManagement::get_min_nb_codon() { return _min_nb_codon; }
unsigned int ParametersManagement::get_sample_size() { return _sample_size; }
double ParametersManagement::get_kappa() { return _kappa; }

bool ParametersManagement::get_remove_frameshift() { return _remove_frameshift; }
bool ParametersManagement::get_allow_internal_bc() { return _allow_internal_bc; }
bool ParametersManagement::get_DEBUG() { return _DEBUG; }

// SET
void ParametersManagement::set_sample_size(unsigned int validSampleSize) { _sample_size = validSampleSize; }

/*------------------------------------*/
/*           Public methods           */
/*------------------------------------*/

void ParametersManagement::checkParameters() {

	// Check if the alignment file is empty
	ifstream alignment (_alignement_file.c_str(), ios::in);

	if (alignment.is_open()) {
		alignment.seekg(0,ios::end);
		unsigned int size = alignment.tellg();

		if( size == 0) ApplicationTools::displayError("The provided alignment file is empty !");
	}

	/* TODO: utiliser FileTools::getNextLine ou regarder si on est direct sur eof ou calculer taille avec les seek */

    // Check the string parameters
	if (_ingroup_species_name == "none") {
		ApplicationTools::displayError("Mandatory parameters -ingroup is missing ! No ingroup species name specified !");
		displayHelp();
	}

	if (_outgroup_species_name == "none") {
		ApplicationTools::displayWarning("Optional parameters -outgroup is missing ! All analysis that requires more than one species won't be executed !");
	}

	if (_execution_mode != "local" && _execution_mode != "galaxy") {
		_execution_mode = "local";
	}

	// Check the numeric parameters
	if (_min_gapN_site < 0) {
		ApplicationTools::displayWarning("The value (" + TextTools::toString(_min_gapN_site) + ") of the -gapN_site parameters is not valid ! Default value (\"" + TextTools::toString(Default_min_gapN_site) + "\") will be used !");
		_min_gapN_site = Default_min_gapN_site;
	}

	if (_max_gapN_seq < 0 or _max_gapN_seq > 1) {
		ApplicationTools::displayWarning("The value (" + TextTools::toString(_max_gapN_seq) + ") of the -gapN_seq parameters is not a valid percentage in decimal form ! Default value (\"" + TextTools::toString(Default_max_gapN_seq) + "\") will be used !");
		_max_gapN_seq = Default_max_gapN_seq;
	}

	if (_tolerance_zone < 0) {
		ApplicationTools::displayWarning("The value (" + TextTools::toString(_tolerance_zone) + ") of the -stop_tolerance parameters is not valid ! Default value (\"" + TextTools::toString(Default_tolerance_zone) + "\") will be used !");
		_tolerance_zone = Default_tolerance_zone;
	}

	if (_DFEa_write_start < 0) {
		ApplicationTools::displayWarning("The value (" + TextTools::toString(_DFEa_write_start) + ") of the -stop_tolerance parameters is not valid ! Default value (\"" + TextTools::toString(Default_DFEa_write_start) + "\") will be used !");
		_DFEa_write_start = Default_DFEa_write_start;
	}

	if (_DFEa_min_NS_SNP < 0) {
		ApplicationTools::displayWarning("The value (" + TextTools::toString(_DFEa_min_NS_SNP) + ") of the -stop_tolerance parameters is not valid ! Default value (\"" + TextTools::toString(Default_DFEa_min_NS_SNP) + "\") will be used !");
		_DFEa_min_NS_SNP = Default_DFEa_min_NS_SNP;
	}

	if (_min_nb_codon < 0) {
		ApplicationTools::displayWarning("The value (" + TextTools::toString(_min_nb_codon) + ") of the -min_nb_codon parameters is not valid ! Default value (\"" + TextTools::toString(Default_min_nb_codon) + "\") will be used !");
		_min_nb_codon = Default_min_nb_codon;
	}

	if (_sample_size < 0) {
		ApplicationTools::displayWarning("The value (" + TextTools::toString(_sample_size) + ") of the -sample_size parameters is not valid ! Default value (\"" + TextTools::toString(Default_sample_size) + "\") will be used !");
		_sample_size = Default_sample_size;
	}

	if (_kappa < 1) {
		ApplicationTools::displayWarning("The value (" + TextTools::toString(_kappa) + ") of the -kappa parameters is not valid ! Default value (\"" + TextTools::toString(Default_kappa) + "\") will be used !");
		_kappa = Default_kappa;
	}

	// Check the boolean parameters
	if (_remove_frameshift != 1 && _remove_frameshift != 0) {
		ApplicationTools::displayWarning("The value (" + TextTools::toString(_remove_frameshift) + ") of the -remove_frameshift parameters is not a valid boolean value (yes/no, true/false (case insensitive)) ! Default value (\"" + TextTools::toString(Default_remove_frameshift) + "\") will be used !");
		_remove_frameshift = Default_remove_frameshift;
	}

	if (_allow_internal_bc != 1 && _allow_internal_bc != 0) {
		ApplicationTools::displayWarning("The value (" + TextTools::toString(_allow_internal_bc) + ") of the -allow_internal_stop parameters is not a valid boolean value (yes/no, true/false (case insensitive)) ! Default value (\"" + TextTools::toString(Default_allow_internal_bc) + "\") will be used !");
		_allow_internal_bc = Default_allow_internal_bc;
	}

	if (_DEBUG != 1 && _DEBUG != 0) {
		_DEBUG = Default_DEBUG;
	}
}

void ParametersManagement::displayHelp() {

	// Initializations
	string remove_frameshift_as_word("no"), allow_internal_bc_as_word("no"), debug_mode_as_word("no");
	if (Default_remove_frameshift) remove_frameshift_as_word = "yes";
	if (Default_allow_internal_bc) allow_internal_bc_as_word = "yes";
	if (Default_DEBUG) debug_mode_as_word = "yes";

	// Display
	cout << "#################################################" << endl;
	cout << "#     Site Frequency Spectra - Help section     #" << endl;
	cout << "#################################################" << endl;
	cout << "Version: " << _versionNumber << " (" << _versionDate << ")" << endl << endl;

	cout << "Usage example (minimal):" << endl << endl;

	cout << "Site_frequency_spectra -alignment_file=seqfile.pfas -ingroup=Species_name_1" << endl << endl;

	cout << "Usage example (with optional parameters):" << endl << endl;

	cout << "Site_frequency_spectra -alignment_file=seqfile.pfas -ingroup=Species_name_1 -outgroup=Species_name_2 -gapN_site=" << Default_min_gapN_site;
	cout << " -gapN_seq=" << Default_max_gapN_seq << " -min_nb_codon=" << Default_min_nb_codon << " -remove_frameshift=" << remove_frameshift_as_word;
	cout << " -tolerance_zone=" << Default_tolerance_zone << " -allow_internal_bc=" << allow_internal_bc_as_word << " -sample_size=" << Default_sample_size;
	cout << " -kappa=" << Default_kappa << endl << endl << endl;

	cout << "Mandatory parameters description:" << endl << endl;

	cout << "-alignment_file:       Alignment input file in Fasta/PseudoFasta format (With sequence names like: >Contig_name|Species_name|Individual_name|Allele_number)" << endl;
	cout << "-ingroup:              Name of the ingroup/focal species" << endl;
	cout << endl;

	cout << "Optional parameters description:" << endl << endl;

	cout << "-outgroup:             Name of the main outgroup/external species" << endl;
	cout << "-gapN_site:            Minimum number of focal sequences without an undetermined character (or a gap) required to keep a given site [>0 - Default: " << Default_min_gapN_site <<"]"  << endl;
	cout << "-gapN_seq:             Maximum tolerated proportion of undetermined characters (or gap) to keep a given sequence [0 to 1 - Default: " << Default_max_gapN_seq << "]" << endl;
	cout << "-min_nb_codon:         Minimum number of complete codons required to perform calculations [>0 - Default: " << Default_min_nb_codon <<"]" << endl;
	cout << "-remove_frameshift:    Reject (yes) or accept (no) all frameshift-containing contigs [Default: " << remove_frameshift_as_word << "]" << endl;
	cout << "-tolerance_zone:       Number of initial/terminal positions in which Stop codons or FrameShift codons are tolerated [>0 - Default: " << Default_tolerance_zone <<"]" << endl;
	cout << "-allow_internal_bc:    Allow/Tolerate internal \"bad codons\" (i.e. Stop or FrameShift codons between the two tolerance zones) and replace them by NNN (1/true/yes) or reject corresponding dataset/contig (0/false/no) [Default: " << allow_internal_bc_as_word << "]" <<endl;
	cout << "-sample_size:          Desired sample size for calculations [>0 - Default: " << Default_sample_size << "]" << endl;
	cout << "-kappa:                Transition/Transversion ratio \"Kappa\" to use in SequenceStatistics methods [>0 - Default: " << Default_kappa <<"]" << endl;
	cout << "-debug_mode:           Activate/Desactivate debug mode [Default: " << debug_mode_as_word << "]" << endl;
	cout << endl;

	cout << "DFEa output related parameters:" << endl << endl;
	cout << "-dfea_write_start:     Matrix rows with a number lower than this parameter will not be written in the output file [>0 - Default: " << Default_DFEa_write_start <<"]" << endl;
	cout << "-dfea_min_NS_SNP:      Minimum number of Non Synonymous SNP per gene copy [>0 - Default: " << Default_DFEa_min_NS_SNP <<"]" << endl;
	cout << endl;

	std::exit(EXIT_FAILURE);
}
