// Basic C++ classes
#include <iostream>
#include <string>
#include <vector>

// Bio++ - Alphabets and genetic codes
#include <Bpp/Seq/Alphabet/StandardCodonAlphabet.h>
#include <Bpp/Seq/GeneticCode/StandardGeneticCode.h>

// Bio++ - Site manipulation
#include <Bpp/Seq/CodonSiteTools.h>

// Bio++ - Population genetics containers and tools
#include <Bpp/PopGen/PolymorphismSequenceContainer.h>
#include <Bpp/PopGen/PolymorphismSequenceContainerTools.h>
#include <Bpp/PopGen/SequenceStatistics.h>

// PopPhyl Project Developments - PopPhyl_Commons
#include <ppc/Seq/Alphabet/CustomNucleicAlphabet.h>
#include <ppc/Seq/Alphabet/FrameshiftCodonAlphabet.h>
#include <ppc/PopGen/Tools/PolymorphismSequenceContainerAddons.h>
#include <ppc/PopGen/Tools/AlignmentFileManagement.h>
#include <ppc/PopGen/AlleleFrequency.h>
#include <ppc/PopGen/SelectionIndicators.h>
#include <ppc/PopGen/SFSneutral.h>
#include <ppc/PopGen/SFSselected.h>
#include <ppc/PopGen/SFSneutralGC.h>

// Custom classes
#include "ParametersManagement.h"

/* Using of Namespaces to avoid the heavy Class::Member_function notation */
using namespace std;
using namespace bpp;
using namespace ppc;


/* Debug and versioning */
const double version = 2.5;
const string version_date = "13-06-2013";

// Time
time_t rawtime = time(NULL);

/*-------------------------------------*/
/*       Display related methods       */
/*-------------------------------------*/

void displayWelcomeMessage (ParametersManagement* SFS_params) {

	if (SFS_params->get_DEBUG()) cout << endl << "-- DEBUG MODE IS ACTIVATED --" << endl << endl;
	cout << endl << "#####################################################################################" << endl;
	cout << "#  Welcome in Site Frequency Spectra - SFS (Version " << version << " (Last update: " << version_date << "))  #" << endl;
	cout << "#####################################################################################" << endl;
	cout << "Start time: " << ctime(&rawtime) << endl;

	cout << "Selected input file: " << SFS_params->get_alignement_file() << endl << endl;
	cout << "Selected focal/ingroup species: " << SFS_params->get_ingroup_species_name() << endl;

	if (SFS_params->get_outgroup_species_name() != "none") {
		cout << "Selected external/outgroup species: " << SFS_params->get_outgroup_species_name() << endl << endl;
	} else {
		cout << endl;
	}

	if (SFS_params->get_remove_frameshift()) {
		cout << "Frameshift-containing contigs will be rejected during the initial cleaning procedure !" << endl;

		cout << "Stop codons will be tolerated in the first/last " << SFS_params->get_tolerance_zone() << " codon(s) of each sequence." << endl;

		if (SFS_params->get_allow_internal_bc()) {
			cout << "Internal Stop codons will be tolerated and replaced by NNN codons." << endl;
		} else {
			cout << "Internal Stop codons are forbidden and the corresponding contig will be rejected !" << endl;
		}
	} else {
		cout << "Frameshift-containing contigs will be conserved during the alignment file loading." << endl;

		cout << "Stop and Frameshift codons will be tolerated in the first/last " << SFS_params->get_tolerance_zone() << " codon(s) of each sequence." << endl;

		if (SFS_params->get_allow_internal_bc()) {
			cout << "Internal Stop and Frameshift codons will be tolerated and replaced by NNN codons." << endl;
		} else {
			cout << "Internal Stop and Frameshift codons are forbidden and the corresponding contig will be rejected !" << endl;
		}
	}

	cout << endl;
}

/******************************************************************************/

void displayGoodbyeMessage() {

	time_t end_rawtime = time(NULL);
	cout << "Stop time: " << ctime(&end_rawtime);
	cout << "###################################################################################" << endl;
	cout << "#                    Site Frequency Spectra - End of execution                    #" << endl;
	cout << "###################################################################################" << endl;
}


/*------------------------------------------------------*/
/*          Input file loading related methods          */
/*------------------------------------------------------*/

vector<PolymorphismSequenceContainer*> loadAlignmentFile (ParametersManagement* allParams, const StandardCodonAlphabet* codonAlphabet, const NucleicAlphabet* customNucleoAlphabet) {

	// Read the alignment file (Multi Pseudofasta) and remove frameshift-containing contigs if requested */
	cout << "Alignment input file analysis:" << endl << endl;

	vector<PolymorphismSequenceContainer*> multidata = AlignmentFileManagement::readAlignmentFile(allParams->get_alignement_file(), allParams->get_remove_frameshift(), customNucleoAlphabet, codonAlphabet, allParams->get_DEBUG());

	// Exit if the number of remaining contig/dataset is not greater than 0
	if (multidata.size() == 0 ) {
		ApplicationTools::displayError("Sorry, there is not enough dataset/contig to perform analysis ! Execution canceled..");
		displayGoodbyeMessage();
		exit(EXIT_FAILURE);
	}

	return multidata;
}

/******************************************************************************/

vector<PolymorphismSequenceContainer*> cleanAlignmentData (ParametersManagement* allParams, vector<PolymorphismSequenceContainer*>& multidata, const StandardCodonAlphabet* codonAlphabet) {

	// Clean the original multidata (Vector of OrderedSequenceContainer*)
	// Note: During the cleaning procedure, the number of site, sequence, ingroup/outgroup member are checked
	// Dataset that does not pass the checks are eliminated. Therefore their won't be anymore "Not enough data" or "less than two species" status in dNdSpiNpiS main output file

	cout << "Datasets/Contigs cleaning procedure:" << endl << endl;

	vector<PolymorphismSequenceContainer*> multidataClean = AlignmentFileManagement::cleanMultiContainer(multidata, codonAlphabet, allParams->get_ingroup_species_name(),
			allParams->get_min_nb_codon(), allParams->get_tolerance_zone(), allParams->get_allow_internal_bc(), allParams->get_min_gapN_site(), allParams->get_max_gapN_seq(),
			allParams->get_DEBUG(), allParams->get_outgroup_species_name());

	if (multidataClean.size() == 0 ) {
		ApplicationTools::displayError("Sorry, there is not enough cleaned dataset/contig to perform analysis ! Execution canceled..");
		displayGoodbyeMessage();
		exit(EXIT_FAILURE);
	}

	return multidataClean;
}

/*-------------------------------------------*/
/*         SelectionIndicators Tools         */
/*-------------------------------------------*/

double computeThreshold (vector<SelectionIndicators*> const & indicators, double nb_se_allowed) {

	// Initializations
	vector<double> allnS, allFixS, dS;

	for (unsigned int j = 0; j < indicators.size(); j++) {
		allnS.push_back(indicators[j]->get_nS());
		allFixS.push_back(indicators[j]->get_fixS());
	}

	for (unsigned int i = 0; i < allFixS.size(); i++) {
		dS.push_back((double)allFixS[i] / (double)allnS[i]);
	}

	return VectorTools::median(dS) + nb_se_allowed * VectorTools::sd<double, double>(dS, true);
}

/******************************************************************************/

vector<double> getTotalSelectionIndicators (vector<SelectionIndicators*> const & indicators, double threshold) {

	vector<double> totals(4, 0.); // In order: nS nN fixS fixN

	// Add a given nN/nS/FixN/FixS value to the correspondent total under condition
	for (unsigned int k = 0; k < indicators.size(); k++) {
		if (indicators[k]->get_fixSnS_ratio() < threshold)	{
			totals[0] += indicators[k]->get_nS();
			totals[1] += indicators[k]->get_nN();
			totals[2] += indicators[k]->get_fixS();
			totals[3] += indicators[k]->get_fixN();
		}
	}

	return totals;
}

/******************************************************************************/

double determineSynonymousSitesProportion(ParametersManagement* allParams, PolymorphismSequenceContainer* ingroupContainer, const StandardCodonAlphabet* codonAlphabet, StandardGeneticCode* stdGeneticCode) {

	// Initializations
	double synonymousSitesProportion(0.);
	double defaultSynonymousSitesProportion(1. / 3.7);

	// Compute nN and nS (on all sequences)
	double nN = SequenceStatistics::meanNonSynonymousSitesNumber(*ingroupContainer, *stdGeneticCode, allParams->get_kappa());
	double nS = SequenceStatistics::meanSynonymousSitesNumber(*ingroupContainer, *stdGeneticCode, allParams->get_kappa());

	if (nN + nS == 0) {
		synonymousSitesProportion = defaultSynonymousSitesProportion;
	} else {
		// Compute value
		synonymousSitesProportion = nS / (nN + nS);

		// Check interval
		if (synonymousSitesProportion > 1. / 2.5 || synonymousSitesProportion < 1. / 6.) {
			synonymousSitesProportion = defaultSynonymousSitesProportion;
		}
	}

	return synonymousSitesProportion;
}


/*------------------------------------------------------------*/
/*     SFS data recovery and manipulation related methods     */
/*------------------------------------------------------------*/

void replaceUnresolvedCharInCodonSite (Site& SiteToModify, const StandardCodonAlphabet* codonAlphabet, unsigned int nbSequence, unsigned int rankOfAllele2) {
	// Replace gap and N characters (unresolved char) by the letter of the second allele
	for (unsigned int i = 0; i < nbSequence; i++) {
		if (codonAlphabet->isUnresolved(SiteToModify[i])) SiteToModify[i] = SiteToModify[rankOfAllele2];
	}
}

/******************************************************************************/

bool isSynonymousSite (Site & siteToAnalyze, unsigned int nbSequence, const StandardCodonAlphabet* codonAlphabet, const StandardGeneticCode* stdGeneticCode, int allele_2_rank) {

	for (unsigned int i = 0; i < nbSequence; i++) {
		if (codonAlphabet->isUnresolved(siteToAnalyze[i])) siteToAnalyze[i] = siteToAnalyze[allele_2_rank];
	}

	if (CodonSiteTools::numberOfNonSynonymousSubstitutions(siteToAnalyze, *stdGeneticCode) != 0) return false;
	else return true;
}

/******************************************************************************/

void addConstantSites (unsigned int loopSize, double synoProportion, vector< vector<double> >& sfsNeutral, vector< vector<double> >& sfsSelected, vector<double>& monomorphic, vector<double>& numberOfSynonymousSNP, vector<double>& numberOfNonSynonymousSNP) {

	for (unsigned int i = 0; i <= loopSize; i++) {

		double monoS = (synoProportion * (monomorphic[i] + numberOfNonSynonymousSNP[i])- numberOfSynonymousSNP[i]) / (1. + synoProportion);

		if (monoS < 0.) monoS = 0.;
		if (monoS > monomorphic[i]) monoS = monomorphic[i];

		sfsNeutral[i][0] += monoS;
		sfsSelected[i][0] += (monomorphic[i] - monoS);
	}
}

/******************************************************************************/

void collectSfsData (unsigned int maxNbFocalSequence, double synoProportion, vector<unsigned int>& totalNumberOfNonSynonymousSNP,
		vector< vector<double> >& sfsNeutral, vector< vector<double> >& sfsNeutralGC, vector< vector<double> >& sfsSelected, PolymorphismSequenceContainer* currentIngroupContainer,
		const StandardCodonAlphabet* codonAlphabet, const NucleicAlphabet* nucleoAlphabet, const StandardGeneticCode* stdGeneticCode) {

	// Initializations
	vector<double> monomorphic (maxNbFocalSequence + 1, 0);
	vector<double> numberOfSynonymousSNP (maxNbFocalSequence + 1, 0);
	vector<double> numberOfNonSynonymousSNP (maxNbFocalSequence + 1, 0);
	unsigned int nbBiAllelic(0), nbPosition(0), nbSite(0);

	for (unsigned int codonSiteNumber = 0; codonSiteNumber < currentIngroupContainer->getNumberOfSites(); codonSiteNumber++) {
		nbSite++;
		for (unsigned int positionNumber = 0; positionNumber < 3; positionNumber++) { // We analyse each nucleotide of the codon
			nbPosition++;
			AlleleFrequency* SNP_data = new AlleleFrequency(currentIngroupContainer, codonSiteNumber, positionNumber, codonAlphabet, nucleoAlphabet);
			unsigned int nbtyped = SNP_data->get_nb_typed();

			if (SNP_data->get_number_of_allele() == 1) monomorphic[SNP_data->get_nb_typed()]++;

			if(SNP_data->is_bi_allelic()) {
				SNP_data->determine_GC_frequencies();
				nbBiAllelic++;

				// Get current codon site and replace unresolved characters (gap, N) to avoid errors in CodonSiteTools::numberOfSubsitutions method
				Site currentCodonSite = currentIngroupContainer->getSite(codonSiteNumber);
				replaceUnresolvedCharInCodonSite(currentCodonSite, codonAlphabet, currentIngroupContainer->getNumberOfSequences(), SNP_data->get_allele_2_rank());

				if (CodonSiteTools::numberOfSubsitutions(currentCodonSite) != 1) continue;

				bool synonymousSite = isSynonymousSite(currentCodonSite, currentIngroupContainer->getNumberOfSequences(), codonAlphabet, stdGeneticCode, SNP_data->get_allele_2_rank());

				if (synonymousSite) {
					numberOfSynonymousSNP[nbtyped]++;
					totalNumberOfNonSynonymousSNP[nbtyped]++;

					sfsNeutral[nbtyped][SNP_data->get_allele_1_occurence()]++;
					if (SNP_data->get_GC_absolute_frequency() > 0) {
						sfsNeutralGC[nbtyped][SNP_data->get_GC_absolute_frequency()]++;
					}
				} else{
					numberOfNonSynonymousSNP[nbtyped]++;

					sfsSelected[nbtyped][SNP_data->get_allele_1_occurence()]++;
				}
			}
		}
	}

	// Add datas for monomorphic sites
	addConstantSites(maxNbFocalSequence, synoProportion, sfsNeutral, sfsSelected, monomorphic, numberOfSynonymousSNP, numberOfNonSynonymousSNP);

	cout << "   --> Number of bi-allelic SNP: " << nbBiAllelic << endl << endl;
}

/*-------------------------------------------*/
/*    Output file writing related methods    */
/*-------------------------------------------*/

unsigned int determineValidData (ParametersManagement* allParameters, unsigned int maxNbTyped, vector<unsigned int>& totalNumberOfNonSynonymousSNP, vector <bool> validNbTyped) {
	// Initializations
	unsigned int nbValidNbTyped = 0;

	// Check validity
	for (unsigned int i = 0; i <= maxNbTyped; i++) {

		if (i < allParameters->get_DFEa_write_start()) {
			validNbTyped[i] = false;
		} else if (totalNumberOfNonSynonymousSNP[i] < allParameters->get_DFEa_min_NS_SNP()) {
			validNbTyped[i] = false;
		} else {
			validNbTyped[i] = true;
			nbValidNbTyped++;
		}
	}

	return nbValidNbTyped;
}

/******************************************************************************/

void writeEmptyFileForGalaxy (string fileToCreate) {

	std::ofstream outputStream (fileToCreate.c_str(), ios::out | ios::trunc);

	// Fill the file
	if (outputStream.is_open()) {
		outputStream.close();
	} else {
		cerr << "Error: Cannot create/open file: " << fileToCreate << endl;
		exit(EXIT_FAILURE);
	}
}

/******************************************************************************/

void writeDFEa (ParametersManagement* allParameters, unsigned int maxNbTyped, vector<unsigned int>& totalNumberOfNonSynonymousSNP, vector<double>& total_nSnNfixSfixN, vector < vector <double> >& neutralMatrix, vector < vector <double> >& selectedMatrix) {
	// Initializations
	string outputFilename("");
	vector <bool> validNbTyped (maxNbTyped +1, 0);

	// Generate valid filename depending on the execution mode
	if (allParameters->get_execution_mode() == "local") {
		outputFilename = FileTools::getFileName(allParameters->get_alignement_file()) + ".SFS_DFEa";

		// Hotfix for getFileName and filenames that don't have an extension
		if (outputFilename == ".SFS_DFEa") {
			StringTokenizer token(allParameters->get_alignement_file(), "/");
			outputFilename = token.getToken(token.numberOfRemainingTokens() - 1) + ".SFS_DFEa";
		}
	} else {
		outputFilename= "Site_frequency_spectra_SFS_DFEa";
	}

	// Check data validity
	unsigned int nbValidNbTyped = determineValidData (allParameters, maxNbTyped, totalNumberOfNonSynonymousSNP, validNbTyped);

	// Create/Open an output stream
	std::ofstream outputStream (outputFilename.c_str(), ios::out | ios::trunc);

	// Fill the file
	if (outputStream.is_open()) {
		outputStream << "1" << endl;
		outputStream << nbValidNbTyped << endl;

		for (unsigned int i = 0; i <= maxNbTyped; i++) {
			if (validNbTyped[i]) outputStream << i << endl;
		}

		outputStream << "1" << endl;
		outputStream << total_nSnNfixSfixN[1] << endl << total_nSnNfixSfixN[3] << endl << total_nSnNfixSfixN[0] << endl << total_nSnNfixSfixN[2] << endl; // nN, fixN, nS, fixS

		for (unsigned int i = 0; i <= maxNbTyped; i++) {
			if (!validNbTyped[i]) continue;

			for (unsigned int j = 0; j <= i; j++) {
				outputStream << neutralMatrix[i][j] << " ";
			}
			outputStream << endl;

			for (unsigned int j = 0; j <= i; j++) {
				outputStream << selectedMatrix[i][j] << " ";
			}
			outputStream << endl;
		}

		outputStream.close();

	} else {
		cerr << "Error: Cannot create/open file: " << outputFilename << endl;
		exit(EXIT_FAILURE);
	}
}

/******************************************************************************/

void writeDoFE (ParametersManagement* allParameters, SFSneutral* neutralData, SFSselected* selectedData, vector<double>& total_nSnNfixSfixN, unsigned int conservedNumberOfDataset) {
	// Initializations
	string outputFilename("");

	// Generate valid filename depending on the execution mode
	if (allParameters->get_execution_mode() == "local") {
		outputFilename = FileTools::getFileName(allParameters->get_alignement_file()) + ".SFS_DoFE";

		// Hotfix for getFileName and filenames that don't have an extension
		if (outputFilename == ".SFS_DoFE") {
			StringTokenizer token(allParameters->get_alignement_file(), "/");
			outputFilename = token.getToken(token.numberOfRemainingTokens() - 1) + ".SFS_DoFE";
		}
	} else {
		outputFilename= "Site_frequency_spectra_SFS_DoFE";
	}

	// Recover folded hypergeometric vectors
	vector <double> selectedHyperSFS = selectedData->get_Hypergeometric_draw();
	vector <double> neutralHyperSFS = neutralData->get_Hypergeometric_draw();

	// Create/Open an output stream
	std::ofstream outputStream (outputFilename.c_str(), ios::out | ios::trunc);

	// Fill the file
	if (outputStream.is_open()) {

		outputStream << allParameters->get_alignement_file() << " (" << conservedNumberOfDataset << " genes)" << endl;
		outputStream << "all_genes" << "\t" << allParameters->get_sample_size() << "\t";
		outputStream << selectedData->get_total_number_of_non_synonymous_SNP() + selectedHyperSFS[0] << "\t";

		for (unsigned int i = 1; i <= allParameters->get_sample_size() / 2; i++) {
			outputStream << selectedHyperSFS[i] << "\t";
		}
		outputStream << neutralData->get_total_number_of_synonymous_SNP() + neutralHyperSFS[0] << "\t";

		for (unsigned int i = 1; i <= allParameters->get_sample_size() / 2; i++) {
			outputStream << neutralHyperSFS[i] << "\t";
		}

		outputStream << total_nSnNfixSfixN[1] << "\t" << total_nSnNfixSfixN[3] << "\t" << total_nSnNfixSfixN[0] << "\t" << total_nSnNfixSfixN[2] << endl; // nN, fixN, nS, fixS

		outputStream.close();

	} else {
		cerr << "Error: Cannot create/open file: " << outputFilename << endl;
		exit(EXIT_FAILURE);
	}
}

/******************************************************************************/

void writeMainSFS (ParametersManagement* allParameters, SFSneutral* neutralData, SFSselected* selectedData, SFSneutralGC* gcData) {
	// Initializations
	string outputFilename("");

	// Generate valid filename depending on the execution mode
	if (allParameters->get_execution_mode() == "local") {
		outputFilename = FileTools::getFileName(allParameters->get_alignement_file()) + ".SFS";

		// Hotfix for getFileName and filenames that don't have an extension
		if (outputFilename == ".SFS") {
			StringTokenizer token(allParameters->get_alignement_file(), "/");
			outputFilename = token.getToken(token.numberOfRemainingTokens() - 1) + ".SFS";
		}
	} else {
		outputFilename= "Site_frequency_spectra_SFS";
	}

	// Recover folded hypergeometric vectors
	vector <double> selectedHyperSFS = selectedData->get_Hypergeometric_draw();
	vector <double> neutralHyperSFS = neutralData->get_Hypergeometric_draw();
	vector <double> neutralHyperGC = gcData->get_Hypergeometric_draw();

	// Create/Open an output stream
	std::ofstream outputStream (outputFilename.c_str(), ios::out | ios::trunc);

	// Fill the file
	if (outputStream.is_open()) {

		outputStream << "GC_allele_frequency\tnb_SNP_syno_unfolded\tminor_allele_frequency\tnb_SNP_syno\tnb_SNP_nonsyno\n";

		for (unsigned int k = 1; k < allParameters->get_sample_size(); k++) {
			outputStream << k << "\t" << neutralHyperGC[k];
			if (k > allParameters->get_sample_size() / 2) {
				outputStream << endl;
			} else {
				outputStream << "\t" << k << "\t" << neutralHyperSFS[k] << "\t" << selectedHyperSFS[k] << endl;
			}
		}

		outputStream.close();

	} else {
		cerr << "Error: Cannot create/open file: " << outputFilename << endl;
		exit(EXIT_FAILURE);
	}
}

/******************************************************************************/

void writeSummary (ParametersManagement* allParameters, SFSneutral* neutralData, SFSselected* selectedData, SFSneutralGC* gcData, unsigned int originalNumberOfDataset, unsigned int conservedNumberOfDataset) {
	// Initializations
	string outputFilename("");

	// Generate valid filename depending on the execution mode
	if (allParameters->get_execution_mode() == "local") {
		outputFilename = FileTools::getFileName(allParameters->get_alignement_file()) + ".SFS_sum";

		// Hotfix for getFileName and filenames that don't have an extension
		if (outputFilename == ".SFS_sum") {
			StringTokenizer token(allParameters->get_alignement_file(), "/");
			outputFilename = token.getToken(token.numberOfRemainingTokens() - 1) + ".SFS_sum";
		}
	} else {
		outputFilename= "Site_frequency_spectra_Summary";
	}

	// Create/Open an output stream
	std::ofstream outputStream (outputFilename.c_str(), ios::out | ios::trunc);

	// Fill the file
	if (outputStream.is_open()) {

		outputStream << "Analysis summary:" << endl << endl;
		outputStream << " -> Number of conserved datasets: " << conservedNumberOfDataset << " (" << ((double)conservedNumberOfDataset / (double)originalNumberOfDataset)*100 << " %)" << endl;
		outputStream << " -> Number of rejected datasets: " << originalNumberOfDataset - conservedNumberOfDataset << endl << endl;

		outputStream << "Non-synonymous SNP data:" << endl << endl;
		outputStream << " -> Number of non-synonymous SNP: " << round(selectedData->get_total_number_of_non_synonymous_SNP()) << endl;
		outputStream << " -> Average minor allele frequency of non-synonymous SNPs: " << selectedData->get_average_allele_frequency_N() << endl;
		outputStream << " -> Tajima's D (non-synonymous SNPs): " << selectedData->get_tajimasD_N() << endl << endl;

		outputStream << "Synonymous SNP data:" << endl << endl;
		outputStream << " -> Number of synonymous SNP: " << round(neutralData->get_total_number_of_synonymous_SNP()) << endl;
		outputStream << " -> Average minor allele frequency of synonymous SNPs: " << neutralData->get_average_allele_frequency_S() << endl;
		outputStream << " -> Tajima's D (synonymous SNPs): " << neutralData->get_tajimasD_S() << endl << endl;

		outputStream << "Other data:" << endl << endl;
		outputStream << " -> Number of AT<->GC synonymous SNP: " << round(gcData->get_GC_min() + gcData->get_GC_maj() + gcData->get_GC_equal()) << endl;
		outputStream << " -> Number of AT<->GC synonymous SNP with majoritary GC allele (Maj): " << round(gcData->get_GC_maj()) << endl;
		outputStream << " -> Number of AT<->GC synonymous SNP with minoritary GC allele (Min): " << round(gcData->get_GC_min()) << endl;
		outputStream << " -> Ratio of AT<->GC synonymous SNP with majoritary GC allele (Maj / (Maj + Min)): " << gcData->get_GC_maj() / (gcData->get_GC_maj() + gcData->get_GC_min()) << endl;

		outputStream.close();

	} else {
		cerr << "Error: Cannot create/open file: " << outputFilename << endl;
		exit(EXIT_FAILURE);
	}
}


/*-------------------------------------------------------*/
/*                 Encapsulation methods                 */
/*-------------------------------------------------------*/

void oneSpeciesAnalysis(ParametersManagement* allParams, vector<PolymorphismSequenceContainer*>& multidataClean, unsigned int originalNumberOfDataset, unsigned int maxNbFocalSequence,
		const StandardCodonAlphabet* codonAlphabet, const NucleicAlphabet* customNucleoAlphabet, StandardGeneticCode* stdGeneticCode) {

	// Initializations (vectors, matrix and variables)
	vector<PolymorphismSequenceContainer*> allIngroups;
	vector<double> SynonymousSitesProportions;

	vector< vector<double> > sfsNeutral (maxNbFocalSequence + 1, vector<double> (maxNbFocalSequence + 1, 0)); // If the max is 20 focal sequence then we need a matrix of 21x21
	vector< vector<double> > sfsNeutralGC (maxNbFocalSequence + 1, vector<double> (maxNbFocalSequence + 1, 0));
	vector< vector<double> > sfsSelected (maxNbFocalSequence + 1, vector<double> (maxNbFocalSequence + 1, 0));
	vector<unsigned int> totalNumberOfNonSynonymousSNP (maxNbFocalSequence + 1, 0);

	for (unsigned int i = 0; i < multidataClean.size(); i++) {
		// Extract all ingroup sequences and store them in a new PolymorphismSequenceContainer
		PolymorphismSequenceContainer* ingroupContainer = PolymorphismSequenceContainerTools::extractIngroup(*multidataClean[i]);

		allIngroups.push_back(ingroupContainer);

		// Compute and store the synonymous site proportion
		SynonymousSitesProportions.push_back(determineSynonymousSitesProportion(allParams, ingroupContainer, codonAlphabet, stdGeneticCode));
	}

	// FixN and FixS cannot be computed with one species, therefore no dS threshold can be calculated and all dataset will be conserved

	// Collect neutral and selected SFS data
	cout << "Compute neutral and selected SFS:" << endl << endl;

	for (unsigned int l = 0; l < allIngroups.size(); l++) {
		cout << " -> Clean dataset " << l + 1 << " (" << PolymorphismSequenceContainerAddons::getContigNameFromContainer(allIngroups[l], 0) << ")" << endl;

		collectSfsData(maxNbFocalSequence, SynonymousSitesProportions[l], totalNumberOfNonSynonymousSNP, sfsNeutral, sfsNeutralGC, sfsSelected, allIngroups[l], codonAlphabet, customNucleoAlphabet, stdGeneticCode);
	}

	// Compute all SFS related variables
	SFSneutral* neutralData = new SFSneutral(maxNbFocalSequence, allParams->get_sample_size(), sfsNeutral, true);
	neutralData->determine_all_attributes();

	SFSselected* selectedData = new SFSselected(maxNbFocalSequence, allParams->get_sample_size(), sfsSelected, true);
	selectedData->determine_all_attributes();

	SFSneutralGC* gcData = new SFSneutralGC(maxNbFocalSequence, allParams->get_sample_size(), sfsNeutralGC);
	gcData->collect_GC_stats();

	// Generate the main output file and the SNP Data file
	cout << "All output files will now be written.." << endl << endl;

	writeMainSFS(allParams, neutralData, selectedData, gcData);
	writeSummary(allParams, neutralData, selectedData, gcData, originalNumberOfDataset, multidataClean.size());

	// Create empty files for Galaxy from_work_dir if needed
	if (allParams->get_execution_mode() == "galaxy") {
		writeEmptyFileForGalaxy("Site_frequency_spectra_SFS_DFEa");
		writeEmptyFileForGalaxy("Site_frequency_spectra_SFS_DoFE");
	}
}

/******************************************************************************/

void twoSpeciesAnalysis(ParametersManagement* allParams, vector<PolymorphismSequenceContainer*>& multidataClean, unsigned int originalNumberOfDataset, unsigned int maxNbFocalSequence,
		const StandardCodonAlphabet* codonAlphabet, const NucleicAlphabet* customNucleoAlphabet, StandardGeneticCode* stdGeneticCode) {

	// Initializations (vectors, matrix and variables)
	vector<PolymorphismSequenceContainer*> allIngroups;
	vector<SelectionIndicators*> allIndicators;
	vector<double> SynonymousSitesProportions;

	vector< vector<double> > sfsNeutral (maxNbFocalSequence + 1, vector<double> (maxNbFocalSequence + 1, 0)); // If the max is 20 focal sequence then we need a matrix of 21x21
	vector< vector<double> > sfsNeutralGC (maxNbFocalSequence + 1, vector<double> (maxNbFocalSequence + 1, 0));
	vector< vector<double> > sfsSelected (maxNbFocalSequence + 1, vector<double> (maxNbFocalSequence + 1, 0));
	vector<unsigned int> totalNumberOfNonSynonymousSNP (maxNbFocalSequence + 1, 0);

	unsigned int numberOfConservedDataset = multidataClean.size();

	// Ex #DEFINE values
	double NB_DS_SE_ALLOWED(2);

	// Work on all cleaned dataset
	for (unsigned int i = 0; i < numberOfConservedDataset; i++) {
		// Extract all ingroup sequences and store them in a new PolymorphismSequenceContainer
		PolymorphismSequenceContainer* ingroupContainer = PolymorphismSequenceContainerTools::extractIngroup(*multidataClean[i]);
		PolymorphismSequenceContainer* outgroupContainer = PolymorphismSequenceContainerTools::extractGroup(*multidataClean[i], 1);

		// Add the newly created ingroup container to the list of ingroup containers
		allIngroups.push_back(ingroupContainer);

		// Compute and store the synonymous site proportion (computed on ingroup data only)
		SynonymousSitesProportions.push_back(determineSynonymousSitesProportion(allParams, ingroupContainer, codonAlphabet, stdGeneticCode));

		// Create an object to store all Synonymous/NonSynonymous related values
		SelectionIndicators* indicator = new SelectionIndicators (ingroupContainer, outgroupContainer, codonAlphabet, allParams->get_ingroup_species_name(), allParams->get_outgroup_species_name());
		indicator->determine_all_attributes(stdGeneticCode, allParams->get_kappa());
		allIndicators.push_back(indicator);
	}

	// Determine maximum FixS/dS allowed value from all valid datasets
	cout << "Determine dS threshold from cleaned datasets: " << endl << endl;
	double dS_threshold = computeThreshold(allIndicators, NB_DS_SE_ALLOWED);
	cout << " -> New threshold: " << dS_threshold << endl << endl;

	// Get total nS, nN, dS, dN depending on the computed threshold
	vector<double> total_nSnNfixSfixN = getTotalSelectionIndicators(allIndicators, dS_threshold);

	// Collect neutral and selected SFS data
	cout << "Compute neutral and selected SFS:" << endl << endl;

	for (unsigned int l = 0; l < allIngroups.size(); l++) {
		cout << " -> Clean dataset " << l + 1 << " (" << PolymorphismSequenceContainerAddons::getContigNameFromContainer(allIngroups[l], 0) << ")" << endl;

		// Reject extreme cases
		if (dS_threshold != 0 and (allIndicators[l]->get_fixSnS_ratio() > dS_threshold))	{
			cout << "   --> Warning: Current fixS/nS ratio is greater than the computed threshold ! Current dataset won't be conserved !" << endl << endl;
			numberOfConservedDataset--;
			continue;
		}

		collectSfsData(maxNbFocalSequence, SynonymousSitesProportions[l], totalNumberOfNonSynonymousSNP, sfsNeutral, sfsNeutralGC, sfsSelected, allIngroups[l], codonAlphabet, customNucleoAlphabet, stdGeneticCode);
	}

	// Compute all SFS related variables
	SFSneutral* neutralData = new SFSneutral(maxNbFocalSequence, allParams->get_sample_size(), sfsNeutral, true);
	neutralData->determine_all_attributes();

	SFSselected* selectedData = new SFSselected(maxNbFocalSequence, allParams->get_sample_size(), sfsSelected, true);
	selectedData->determine_all_attributes();

	SFSneutralGC* gcData = new SFSneutralGC(maxNbFocalSequence, allParams->get_sample_size(), sfsNeutralGC);
	gcData->collect_GC_stats();

	// Generate all output files
	cout << "All output files will now be written.." << endl << endl;

	writeDFEa(allParams, maxNbFocalSequence, totalNumberOfNonSynonymousSNP, total_nSnNfixSfixN, sfsNeutral, sfsSelected);
	writeDoFE(allParams, neutralData, selectedData, total_nSnNfixSfixN, numberOfConservedDataset);
	writeMainSFS(allParams, neutralData, selectedData, gcData);
	writeSummary(allParams, neutralData, selectedData, gcData, originalNumberOfDataset, multidataClean.size());
}


/*------------------------------------*/
/*                MAIN                */
/*------------------------------------*/

int main(int argc, char** argv) {

	// Initializations
	unsigned int maxNbFocalSequence(0), maxNbFocalSequenceInCleanData(0);

	// Alphabet objects creation
	NucleicAlphabet* customNucleoAlphabet = new CustomNucleicAlphabet(); // CustomNucleicAlphabet is an alphabet (derived from the Bio++ DNA alphabet) that can deal with the frameshift (!) character. PopPhyl_Commons is therefore required.
	//const StandardCodonAlphabet* codonAlphabet = new StandardCodonAlphabet(customNucleoAlphabet);
	FrameshiftCodonAlphabet* codonAlphabet = new FrameshiftCodonAlphabet(customNucleoAlphabet);

	// Genetic code object creation
	StandardGeneticCode* stdGeneticCode = new StandardGeneticCode(customNucleoAlphabet);

	// Command line parameters management
	ParametersManagement* SFS_params = new ParametersManagement(version, version_date, argc, argv);
	SFS_params->checkParameters();

	// Welcome message
	displayWelcomeMessage(SFS_params);

	// Load the alignment file (Multi Pseudofasta) into a vector of PolymorphismSequenceContainer and clean it
	vector<PolymorphismSequenceContainer*> brutDatasets = loadAlignmentFile(SFS_params, codonAlphabet, customNucleoAlphabet);

	// Get the maximum and minimum number of focal individual in the multidata
	maxNbFocalSequence = AlignmentFileManagement::getMaxNumberOfFocalIndividual(brutDatasets, SFS_params->get_ingroup_species_name());

	// Clean collected alignment data
	vector<PolymorphismSequenceContainer*> cleanDatasets = cleanAlignmentData(SFS_params, brutDatasets, codonAlphabet);

	// Adapt the value of the sample_size parameter if needed (sample_size > maximum number of focal individual in the cleaned data)
	maxNbFocalSequenceInCleanData = AlignmentFileManagement::getMaxNumberOfFocalIndividual(cleanDatasets, SFS_params->get_ingroup_species_name());

	if (SFS_params->get_sample_size() > maxNbFocalSequenceInCleanData) {
		cout << " -> Warning: the value (" << SFS_params->get_sample_size() << ") of the sample_size parameters is greater than the maximum number of focal sequences" << endl;
		SFS_params->set_sample_size(maxNbFocalSequenceInCleanData);
		cout << "   --> The sample size is now set to " << maxNbFocalSequenceInCleanData << " !" << endl << endl;
	}

	// Selection indicators and SFS data colmputation depends on the number of species to consider
	// If the user didn't use the -outgroup parameters, then outgroup sequences will not be considered
	if (SFS_params->get_outgroup_species_name() == "none") {
		// Ingroup only
		oneSpeciesAnalysis(SFS_params, cleanDatasets, brutDatasets.size(), maxNbFocalSequence, codonAlphabet, customNucleoAlphabet, stdGeneticCode);
	} else {
		// Ingroup + outgroup
		twoSpeciesAnalysis(SFS_params, cleanDatasets, brutDatasets.size(), maxNbFocalSequence, codonAlphabet, customNucleoAlphabet, stdGeneticCode);
	}

	// Goodbye message
	displayGoodbyeMessage();
}
