/*
 * ParametersManagement.cpp
 *
 *  Created on: 23 oct. 2012
 *      Author: abernard
 */

#include "ParametersManagement.h"

// Basic C++ classes
#include <stdlib.h>
#include <iostream>
#include <fstream>

// Bio++ - Basic
#include <Bpp/Io/FileTools.h>

// Bio++ - Parameters Management
#include <Bpp/App/ApplicationTools.h>
#include <Bpp/Utils/AttributesTools.h>

using namespace std;
using namespace bpp;

/*-------------------------------*/
/*          Constructor          */
/*-------------------------------*/

ParametersManagement::ParametersManagement(const double versionNumber, const string versionDate, int nbArguments, char **allArguments) {

	// Set version related attributes
	_versionNumber = versionNumber;
	_versionDate = versionDate;

	// Set default values
	Default_execution_mode = "local";
	Default_contig_usage = "all";

	Default_nb_gc_bin = 8;
	Default_min_gapN_site = 10;
	Default_max_gapN_seq = 0.5;
	Default_tolerance_zone = 20;
	Default_min_nb_codon = 10;
	Default_kappa = 1.;

	Default_bootstrap_replicates = 10000;
	Default_bootstrap_sort_precision = 1000000;
	Default_bootstrap_confidence = 0.95;

	Default_remove_frameshift = false;
	Default_compute_distances = false;
	Default_allow_internal_bc = false;
	Default_print_header = true;

	Default_DEBUG = false;

	//Check minimum required number of parameters
	if (nbArguments < 2) displayHelp();

	// Store all arguments of the command line into a map of string
	map<string, string> cmdParams = AttributesTools::getAttributesMap(AttributesTools::getVector(nbArguments, allArguments), "=");

	// File parameters
	_alignement_file = ApplicationTools::getAFilePath("-alignment_file", cmdParams, true, true, "", true);
//	_tree_file = ApplicationTools::getAFilePath("-tree_file", cmdParams, false, true, "", true);
	_population_file = ApplicationTools::getAFilePath("-population_file", cmdParams, false, true, "", true);

	_output_file_name =  ApplicationTools::getAFilePath("-out", cmdParams, false, false, "", true);

	// String parameters
	_ingroup_species_name = ApplicationTools::getStringParameter("-ingroup", cmdParams, "none", "", false, false);
	_outgroup_species_name = ApplicationTools::getStringParameter("-outgroup", cmdParams, "none", "", false, false);
	_execution_mode = ApplicationTools::getStringParameter("-exec_mode", cmdParams, Default_execution_mode, "", false, false);
	_contig_usage = ApplicationTools::getStringParameter("-contig_usage", cmdParams, Default_contig_usage, "", false, true);

	// Numeric parameters
	_nb_gc_bin = ApplicationTools::getIntParameter("-gc_bin", cmdParams, Default_nb_gc_bin, "", false, true);
	_min_gapN_site = ApplicationTools::getIntParameter("-gapN_site", cmdParams, Default_min_gapN_site, "", false, true);
	_max_gapN_seq = ApplicationTools::getDoubleParameter("-gapN_seq", cmdParams, Default_max_gapN_seq, "", false, true);
	_tolerance_zone = ApplicationTools::getIntParameter("-tolerance_zone", cmdParams, Default_tolerance_zone, "", false, true);
	_min_nb_codon = ApplicationTools::getIntParameter("-min_nb_codon", cmdParams, Default_min_nb_codon, "", false, true);
	_kappa = ApplicationTools::getIntParameter("-kappa", cmdParams, Default_kappa, "", false, true);

	// Bootstrap related parameters
	_bootstrap_replicates = ApplicationTools::getIntParameter("-bs_replicates", cmdParams, Default_bootstrap_replicates, "", false, false);
	_bootstrap_sort_precision = ApplicationTools::getIntParameter("-bs_sort", cmdParams, Default_bootstrap_sort_precision, "", false, false);
	_bootstrap_confidence = ApplicationTools::getDoubleParameter("-bs_confidence", cmdParams, Default_bootstrap_confidence, "", false, false);

	// Boolean parameters (Recall: 1 is true, 0 is false)
	_remove_frameshift = ApplicationTools::getBooleanParameter("-remove_frameshift", cmdParams, Default_remove_frameshift, "", false, true);
	_compute_distances = ApplicationTools::getBooleanParameter("-compute_distances", cmdParams, Default_compute_distances, "", false, true);
	_allow_internal_bc = ApplicationTools::getBooleanParameter("-allow_internal_bc", cmdParams, Default_allow_internal_bc, "", false, true);
	_print_header = ApplicationTools::getBooleanParameter("-print_header", cmdParams, Default_print_header, "", false, true);

	_DEBUG = ApplicationTools::getBooleanParameter("-debug_mode", cmdParams, Default_DEBUG, "", false, false);
}

/*--------------------------------*/
/*           Destructor           */
/*--------------------------------*/

ParametersManagement::~ParametersManagement() {
	// TODO Auto-generated destructor stub
}

/*--------------------------------*/
/*            Accessor            */
/*--------------------------------*/

// Get
string ParametersManagement::get_alignement_file() { return _alignement_file; }
//string ParametersManagement::get_tree_file() { return _tree_file; }
string ParametersManagement::get_population_file() { return _population_file; }

string ParametersManagement::get_ingroup_species_name() { return _ingroup_species_name; }
string ParametersManagement::get_outgroup_species_name() { return _outgroup_species_name; }

string ParametersManagement::get_execution_mode() { return _execution_mode; }
string ParametersManagement::get_output_file_name() { return _output_file_name; }
string ParametersManagement::get_contig_usage() { return _contig_usage; }

unsigned int ParametersManagement::get_nb_gc_bin() { return _nb_gc_bin; }
unsigned int ParametersManagement::get_min_gapN_site() { return _min_gapN_site; }
double ParametersManagement::get_max_gapN_seq() { return _max_gapN_seq; }
unsigned int ParametersManagement::get_tolerance_zone() { return _tolerance_zone; }
unsigned int ParametersManagement::get_min_nb_codon() { return _min_nb_codon; }
double ParametersManagement::get_kappa() { return _kappa; }

unsigned int ParametersManagement::get_bootstrap_replicates() { return _bootstrap_replicates; }
unsigned int ParametersManagement::get_bootstrap_sort_precision() { return _bootstrap_sort_precision; }
double ParametersManagement::get_bootstrap_confidence() { return _bootstrap_confidence; }

bool ParametersManagement::get_remove_frameshift() { return _remove_frameshift; }
bool ParametersManagement::get_compute_distances() { return _compute_distances; }
bool ParametersManagement::get_allow_internal_bc() { return _allow_internal_bc; }
bool ParametersManagement::get_print_header() { return _print_header; }

bool ParametersManagement::get_DEBUG() { return _DEBUG; }

// Set
void ParametersManagement::set_outgroup_species_name(string speciesName) {
	_outgroup_species_name = speciesName;
}

/*------------------------------------*/
/*           Public methods           */
/*------------------------------------*/

void ParametersManagement::checkParameters() {

	// Check if the alignment file is empty
	ifstream alignment (_alignement_file.c_str(), ios::in);

	if (alignment.is_open()) {
		alignment.seekg(0,ios::end);
		unsigned int size = alignment.tellg();

		if( size == 0) ApplicationTools::displayError("The provided alignment file is empty !");
	}

	alignment.close();

//	// Check if the alignment file is empty
//	if (_tree_file == "none") {
//		 ApplicationTools::displayWarning("No tree file specified ! All analysis that requires more than one species won't be executed !");
//	} else {
//		ifstream tree (_tree_file.c_str(), ios::in);
//
//		if (tree.is_open()) {
//			tree.seekg(0,ios::end);
//			unsigned int size = tree.tellg();
//
//			if( size == 0) ApplicationTools::displayError("The provided tree file is empty !");
//		}
//
//		tree.close();
//	}

	// Check if the specified population file is empty
	if (_population_file != "none") {
		ifstream population (_population_file.c_str(), ios::in);

		if (population.is_open()) {
			population.seekg(0,ios::end);
			unsigned int size = population.tellg();

			if( size == 0) ApplicationTools::displayError("The provided population file is empty !");
		}

		population.close();
	}

    // Check the string parameters
	if (_ingroup_species_name == "none") {
		ApplicationTools::displayError("Mandatory parameter -ingroup is missing ! No ingroup species name specified !");
		displayHelp();
	}

	if (_outgroup_species_name == "none") {
		ApplicationTools::displayWarning("Parameter -outgroup is missing ! All analysis that requires more than one species won't be executed !");
	}

	if (_execution_mode != "local" && _execution_mode != "galaxy") {
		_execution_mode = "local";
	}

	if (_contig_usage != "all" && _contig_usage != "even" && _contig_usage != "odd") {
		ApplicationTools::displayWarning("The value (" + TextTools::toString(_contig_usage) + ") of the -contig_usage parameters is not valid ! Default value (\"" + TextTools::toString(Default_contig_usage) + "\") will be used !");
		_contig_usage = Default_contig_usage;
	}

	// Check the ouput file name
	if (_output_file_name == "none") {
		if (_execution_mode == "galaxy") {
			_output_file_name = "dNdSpiNpiS_output";
		} else {
			_output_file_name = FileTools::getFileName(_alignement_file) + ".out";

			// Hotfix for getFileName and filenames that don't have an extension
			if (_output_file_name == ".out") {
				StringTokenizer token(_alignement_file, "/");
				_output_file_name = token.getToken(token.numberOfRemainingTokens() - 1) + ".out";
			}
		}
	}

	// Check the numeric parameters
	if (_nb_gc_bin < 0) {
		ApplicationTools::displayWarning("The value (" + TextTools::toString(_nb_gc_bin) + ") of the -gc_bin parameters is not valid ! Default value (\"" + TextTools::toString(Default_nb_gc_bin) + "\") will be used !");
		_nb_gc_bin = Default_nb_gc_bin;
	}

	if (_min_gapN_site < 0) {
		ApplicationTools::displayWarning("The value (" + TextTools::toString(_min_gapN_site) + ") of the -gapN_site parameters is not valid ! Default value (\"" + TextTools::toString(Default_min_gapN_site) + "\") will be used !");
		_min_gapN_site = Default_min_gapN_site;
	}

	if (_max_gapN_seq < 0 or _max_gapN_seq > 1) {
		ApplicationTools::displayWarning("The value (" + TextTools::toString(_max_gapN_seq) + ") of the -gapN_seq parameters is not a valid percentage in decimal form ! Default value (\"" + TextTools::toString(Default_max_gapN_seq) + "\") will be used !");
		_max_gapN_seq = Default_max_gapN_seq;
	}

	if (_tolerance_zone < 0) {
		ApplicationTools::displayWarning("The value (" + TextTools::toString(_tolerance_zone) + ") of the -stop_tolerance parameters is not valid ! Default value (\"" + TextTools::toString(Default_tolerance_zone) + "\") will be used !");
		_tolerance_zone = Default_tolerance_zone;
	}

	if (_min_nb_codon < 0) {
		ApplicationTools::displayWarning("The value (" + TextTools::toString(_min_nb_codon) + ") of the -min_nb_codon parameters is not valid ! Default value (\"" + TextTools::toString(Default_min_nb_codon) + "\") will be used !");
		_min_nb_codon = Default_min_nb_codon;
	}

	if (_kappa < 1) {
		ApplicationTools::displayWarning("The value (" + TextTools::toString(_kappa) + ") of the -kappa parameters is not valid ! Default value (\"" + TextTools::toString(Default_kappa) + "\") will be used !");
		_kappa = Default_kappa;
	}

	// Check bootstrap related parameters
	if (_bootstrap_replicates < 1) {
		ApplicationTools::displayWarning("The value (" + TextTools::toString(_bootstrap_replicates) + ") of the -bs_replicates parameters is not valid ! Default value (\"" + TextTools::toString(Default_bootstrap_replicates) + "\") will be used !");
		_bootstrap_replicates = Default_bootstrap_replicates;
	}

	if (_bootstrap_sort_precision < 1) {
		ApplicationTools::displayWarning("The value (" + TextTools::toString(_bootstrap_sort_precision) + ") of the -bs_sort parameters is not valid ! Default value (\"" + TextTools::toString(Default_bootstrap_sort_precision) + "\") will be used !");
		_bootstrap_sort_precision = Default_bootstrap_sort_precision;
	}

	if (_bootstrap_confidence < 0 or _bootstrap_confidence > 1) {
		ApplicationTools::displayWarning("The value (" + TextTools::toString(_bootstrap_confidence) + ") of the -bs_confidence parameters is not a valid percentage in decimal form ! Default value (\"" + TextTools::toString(Default_bootstrap_confidence) + "\") will be used !");
		_bootstrap_confidence = Default_bootstrap_confidence;
	}

	// Check the boolean parameters
	if (_remove_frameshift != 1 && _remove_frameshift != 0) {
		ApplicationTools::displayWarning("The value (" + TextTools::toString(_remove_frameshift) + ") of the -remove_frameshift parameters is not a valid boolean value (yes/no, true/false (case insensitive)) ! Default value (\"" + TextTools::toString(Default_remove_frameshift) + "\") will be used !");
		_remove_frameshift = Default_remove_frameshift;
	}

	if (_compute_distances != 1 && _compute_distances != 0) {
		ApplicationTools::displayWarning("The value (" + TextTools::toString(_compute_distances) + ") of the -compute_distances parameters is not a valid boolean value (yes/no, true/false (case insensitive)) ! Default value (\"" + TextTools::toString(Default_compute_distances) + "\") will be used !");
		_compute_distances = Default_compute_distances;
	}

	if (_allow_internal_bc != 1 && _allow_internal_bc != 0) {
		ApplicationTools::displayWarning("The value (" + TextTools::toString(_allow_internal_bc) + ") of the -allow_internal_stop parameters is not a valid boolean value (yes/no, true/false (case insensitive)) ! Default value (\"" + TextTools::toString(Default_allow_internal_bc) + "\") will be used !");
		_allow_internal_bc = Default_allow_internal_bc;
	}

	if (_print_header != 1 && _print_header != 0) {
		ApplicationTools::displayWarning("The value (" + TextTools::toString(_print_header) + ") of the -print_header parameters is not a valid boolean value (yes/no, true/false (case insensitive)) ! Default value (\"" + TextTools::toString(Default_print_header) + "\") will be used !");
		_print_header = Default_print_header;
	}

	if (_DEBUG != 1 && _DEBUG != 0) {
		_DEBUG = Default_DEBUG;
	}
}

void ParametersManagement::displayHelp() {

	// Initializations
	string remove_frameshift_as_word("no"), allow_internal_bc_as_word("no"), print_header_as_word("no"), debug_mode_as_word("no"), compute_distances_as_word("no");
	if (Default_remove_frameshift) remove_frameshift_as_word = "yes";
	if (Default_compute_distances) compute_distances_as_word = "yes";
	if (Default_allow_internal_bc) allow_internal_bc_as_word = "yes";
	if (Default_print_header) print_header_as_word = "yes";
	if (Default_DEBUG) debug_mode_as_word = "yes";

	// Display
	cout << "#################################################" << endl;
	cout << "#           dNdSpiNpiS - Help section           #" << endl;
	cout << "#################################################" << endl;
	cout << "Version: " << _versionNumber << " (" << _versionDate << ")" << endl << endl;

	cout << "Usage example (minimal):" << endl << endl;

	cout << "dNdSpiNpiS -alignment_file=seqfile.pfas -ingroup=Species_name_1" << endl << endl;

	cout << "Usage example (with optional parameters):" << endl << endl;

	cout << "dNdSpiNpiS -alignment_file=seqfile.pfas -population_file=custom_pop.txt -ingroup=Species_name_1 -outgroup=Species_name_2 -gc_bin=" << Default_nb_gc_bin;
	cout << " -gapN_site=" << Default_min_gapN_site << " -gapN_seq=" << Default_max_gapN_seq << " -min_nb_codon=" << Default_min_nb_codon << " -kappa=" << Default_kappa;
	cout << " -remove_frameshift=" << remove_frameshift_as_word << " -compute_distances=" << compute_distances_as_word << " -tolerance_zone=" << Default_tolerance_zone;
	cout << " -allow_internal_bc=" << allow_internal_bc_as_word << " -print_header=" << print_header_as_word << " -contig_usage=" << Default_contig_usage << " -out=dNdSpiNpiS_output" << endl << endl << endl;

	cout << "Mandatory parameters:" << endl << endl;

	cout << "-alignment_file:       Alignment input file in Fasta/PseudoFasta format (With sequence names like: >Contig_name|Species_name|Individual_name|Allele_number)" << endl;
	cout << "-ingroup:              Name of the ingroup/focal species" << endl;
	cout << endl;

	cout << "Optional parameters:" << endl << endl;

	cout << "-population_file:      Text file containing a list of individuals (from the focal species) on each line" << endl;
	cout << "                       If not specified, all ingroup's individuals will be used in one unique population" << endl;
//	cout << "-tree_file:            Species tree file in Newick format (with/without branch lengths)" << endl;
	cout << "-outgroup:             Name of the main outgroup/external species" << endl;
	cout << "-gc_bin:               Number of class/bin for GC allele frequency [>0 - Default: " << Default_nb_gc_bin << "]" << endl;
	cout << "-gapN_site:            Minimum number of focal sequences without an undetermined character (or a gap) required to keep a given site [>0 - Default: " << Default_min_gapN_site << "]" << endl;
	cout << "-gapN_seq:             Maximum tolerated proportion of undetermined characters (or gap) to keep a given sequence [0 to 1 - Default: " << Default_max_gapN_seq << "]" << endl;
	cout << "-min_nb_codon:         Minimum number of complete codons required to perform calculations [>0 - Default: " << Default_min_nb_codon << "]" << endl;
	cout << "-kappa:                Transition/Transversion ratio \"Kappa\" to use in SequenceStatistics methods [>0 - Default: " << Default_kappa << "]" << endl;
	cout << "-remove_frameshift:    Reject (1/true/yes) or accept (0/false/no) all frameshift-containing contigs [Default: " << remove_frameshift_as_word << "]" << endl;
	cout << "-compute_distances:    Define if the distances between individuals must be computed (1/true/yes) or not (0/false/no) [Default: " << compute_distances_as_word << "]" << endl;
	cout << "-tolerance_zone:       Number of initial/terminal positions in which Stop codons or FrameShift codons are tolerated [>0 - Default: " << Default_tolerance_zone <<"]" << endl;
	cout << "-allow_internal_bc:    Allow/Tolerate internal \"bad codons\" (i.e. Stop or FrameShift codons between the two tolerance zones) and replace them by NNN (1/true/yes) or reject corresponding dataset/contig (0/false/no) [Default: " << allow_internal_bc_as_word << "]" <<endl;
	cout << "-print_header:         Define if the header line must be output (1/true/yes), or not (0/false/no) [0 or 1 - Default: " << print_header_as_word << "]" << endl;
	cout << "-contig_usage:         Define if 'all' the contigs, 'even' contigs or 'odd' contigs must be used for dNdSpiNpiS summary [Default: " << Default_contig_usage << " - Possible values: all, even, odd]" << endl;
	cout << endl;

	cout << "-out:                  Name of dNdSpiNpiS main output file (Population genetics)" << endl;
	cout << "                       If not specified, a default name (derived from the name of the Fasta/PseudoFasta input file) will be used" << endl;
	cout << "-debug_mode:           Activate/Desactivate debug mode [Default: " << debug_mode_as_word << "]" << endl;
	cout << endl;

	cout << "Optional bootstrap related parameters:" << endl << endl;
	cout << "-bs_replicates:        Number of bootstrap resampling to execute [>1 - Default: " << Default_bootstrap_replicates << "]" << endl;
	cout << "-bs_sort:              Sort precision for the confidence interval [>1 - Default: " << Default_bootstrap_sort_precision << "]" << endl;
	cout << "-bs_confidence:        Percentage of confidence [0 to 1 - Default: " << Default_bootstrap_confidence << "]" << endl;
	cout << endl;

	std::exit(EXIT_FAILURE);
}
