/*
 * ResultsWarehouse.cpp
 *
 *  Created on: 25 oct. 2012
 *      Author: abernard
 */

#include "ResultsWarehouse.h"

// Bio++ classes
#include <Bpp/Numeric/NumTools.h>

// PopPhyl Project Developments
#include <ppc/PopGen/DistanceBetweenIndividuals.h>

using namespace std;
using namespace bpp;
using namespace ppc;

/*-------------------------------*/
/*          Constructor          */
/*-------------------------------*/

ResultsWarehouse::ResultsWarehouse(Population* popData, unsigned int numberOfBin): _name("No_name"), _status("Not_OK"), _nb_bi_allelic_SNP(0), _nb_tri_allelic_SNP(0),
		_nb_quadri_allelic_SNP(0), _nb_sequences_in_consensus(0), _nb_sites_in_consensus(0), _nN(-1.), _nS(-1.), _fixN(-1), _fixS(-1), _polymN_0(-1), _polymS_0(-1),
		_polymN_02(-1), _polymS_02(-1), _MK_nN(-1), _MK_nS(-1), _GC_allele_frequency_classes(numberOfBin, 0.), _allSpeciesData(0) {

	// Initialize Fst matrix
	vector<double> init (popData->get_number_of_population(), -999.);
	for (unsigned int a = 0; a < popData->get_number_of_population(); a ++) {
		_Fst.push_back(init);
	}

	// Initialize distance map
	vector<string> allIndividuals = popData->get_all_individuals();

	for (unsigned int i = 0; i < allIndividuals.size(); i++) {
		for (unsigned int j = 0; j < allIndividuals.size(); j++) {
			_distances[allIndividuals[i] + "|" + allIndividuals[j]] = new DistanceBetweenIndividuals(allIndividuals[i], allIndividuals[j]);
		}
	}

	// Initialize Heterozygosity map
	for (unsigned int k = 0; k < allIndividuals.size(); k++) {
		_individual_heterozygosity[allIndividuals[k]] = -999.;
	}
}


/*--------------------------------*/
/*           Destructor           */
/*--------------------------------*/

ResultsWarehouse::~ResultsWarehouse() {

	// Empty the vector of SpeciesDataWarehouse*
	for (unsigned int i = 0; i < _allSpeciesData.size(); i++) {
		delete(_allSpeciesData[i]);
	}

	_allSpeciesData.clear();
}


/*------------------------------------------------------*/
/*                     Debug method                     */
/*------------------------------------------------------*/

void ResultsWarehouse::warehouseInventory (ResultsWarehouse* warehouse, ParametersManagement* params, Population* popData, vector<double>& alleleFrequencyClass) {

	// Get all data
	SpeciesDataWarehouse* ingroupData = warehouse->get_specific_species_data(params->get_ingroup_species_name());
	map <string, double > hetero = warehouse->get_all_individual_heterozygosity();
	vector< vector<string> > populations = popData->get_population_list();
	unsigned int nbPopulations = popData->get_number_of_population();
	vector<string> populationNames =  popData->get_population_names();
	vector<double> gcAlleleFrequencies = warehouse->get_GC_allele_frequency_classes();

	// Display
	cout << endl << "#################################################################################" << endl;
	cout << "#                              Warehouse Inventory                              #" << endl;
	cout << "#################################################################################" << endl;
	cout << endl << "Content of the warehouse for dataset " << warehouse->get_name() << " (Status: " << warehouse->get_status() << "):" << endl;

	cout << endl << " -- 1 species analysis -- " << endl;

	cout << endl << " -> Focal species Data:" << endl << endl;
	cout << "  --> Name: " << ingroupData->get_species_name() << endl;
	cout << "  --> Number of sequence: " << ingroupData->get_number_of_sequences() << endl;
	cout << "  --> Number of site: " << ingroupData->get_number_of_sites() << endl;
	cout << "  --> Number of complete site: " << ingroupData->get_number_of_complete_sites() << endl;
	cout << "  --> piN: " << ingroupData->get_piN() << endl;
	cout << "  --> piS: " << ingroupData->get_piS() << endl;
	cout << "  --> Fit: " << ingroupData->get_Fit() << endl;
	cout << "  --> WC84 Fit: " << ingroupData->get_WC84_Fit() << endl;
	cout << "  --> GC3: " << ingroupData->get_GC3() << endl;

	cout << endl << " -> Individual Heterozygosity:" << endl << endl;
	for(map<string, double>::iterator it=hetero.begin(); it!=hetero.end(); ++it) {
		cout << "  --> H(" << it->first << ") = " << it->second << endl;
	}

	if (nbPopulations > 1) {
		cout << endl << " -> Fst:" << endl << endl;
		for (unsigned int popIndex1 = 0; popIndex1 < nbPopulations; popIndex1++) {
			for (unsigned int popIndex2 = popIndex1 + 1; popIndex2 < nbPopulations; popIndex2++) {
				cout << "  --> Fst(" << populationNames[popIndex1] << "," << populationNames[popIndex2] << "): " << warehouse->get_specific_Fst(popIndex1, popIndex2) << endl;
			}
		}
	}

	cout << endl << " -> GC frequencies:" << endl << endl;
	cout << "  --> Range [0 - " << alleleFrequencyClass[0] << "]: " << gcAlleleFrequencies[0] << endl;
	for (unsigned int i = 1; i < gcAlleleFrequencies.size(); i++) {
		cout << "  --> Range [" << alleleFrequencyClass[i-1] << " - " << alleleFrequencyClass[i] << "]: " << gcAlleleFrequencies[i] << endl;
	}

	cout << endl << " -> SNP data:" << endl << endl;
	cout << "  --> Bi-allelic SNP: " << warehouse->get_nb_bi_allelic_SNP() << endl;
	cout << "  --> Tri-allelic SNP: " << warehouse->get_nb_tri_allelic_SNP() << endl;
	cout << "  --> Quadri-allelic SNP: " << warehouse->get_nb_quadri_allelic_SNP() << endl;

	cout << endl << " -- 2 species analysis -- " << endl;

	cout << endl << " -> Selection indicators:" << endl << endl;
	cout << "  --> nN: " << warehouse->get_nN() << endl;
	cout << "  --> nS: " << warehouse->get_nS() << endl;
	cout << "  --> nbseq_div: " << warehouse->get_nb_sequences_in_consensus() << endl;
	cout << "  --> nbsite_div: " << warehouse->get_nb_sites_in_consensus() << endl;

	cout << endl << " -> mcDonald-Kreitman Analysis:" << endl << endl;
	cout << "  --> FixN: " << warehouse->get_fixN() << endl; // Da
	cout << "  --> FixS: " << warehouse->get_fixS() << endl; // Ds
	cout << "  --> PolymN 0.0: " << warehouse->get_polymN_0() << endl; // Pa
	cout << "  --> PolymS 0.0: " << warehouse->get_polymS_0() << endl; // Ps
	cout << "  --> PolymN 0.2: " << warehouse->get_polymN_02() << endl; // Pa 0.2
	cout << "  --> PolymS 0.2: " << warehouse->get_polymS_02() << endl; // Ps 0.2
	cout << "  --> MK_nN: " << warehouse->get_MK_nN() << endl;
	cout << "  --> MK_nS: " << warehouse->get_MK_nS() << endl;

	cout << endl << "#################################################################################" << endl;
	cout << "#                           Warehouse Inventory - END                           #" << endl;
	cout << "#################################################################################" << endl << endl;
}


/*---------------------------------------*/
/*            Accessors - Set            */
/*---------------------------------------*/

void ResultsWarehouse::set_name(string contigName) {
	_name = contigName;
}

void ResultsWarehouse::set_status(string newStatus) {
	_status = newStatus;
}


void ResultsWarehouse::set_nb_bi_allelic_SNP(unsigned int newValue) {
	_nb_bi_allelic_SNP = newValue;
}

void ResultsWarehouse::set_nb_tri_allelic_SNP(unsigned int newValue) {
	_nb_tri_allelic_SNP = newValue;
}

void ResultsWarehouse::set_nb_quadri_allelic_SNP(unsigned int newValue) {
	_nb_quadri_allelic_SNP = newValue;
}


void ResultsWarehouse::set_nb_sequences_in_consensus(unsigned int newValue) {
	_nb_sequences_in_consensus = newValue;
}

void ResultsWarehouse::set_nb_sites_in_consensus(unsigned int newValue) {
	_nb_sites_in_consensus = newValue;
}


void ResultsWarehouse::set_nN(double newValue) {
	_nN = newValue;
}

void ResultsWarehouse::set_nS(double newValue) {
	_nS = newValue;
}


void ResultsWarehouse::set_fixN(int newValue) {
	_fixN = newValue;
}

void ResultsWarehouse::set_fixS(int newValue) {
	_fixS = newValue;
}

void ResultsWarehouse::set_polymN_0(int newValue) {
	_polymN_0 = newValue;
}

void ResultsWarehouse::set_polymS_0(int newValue) {
	_polymS_0 = newValue;
}

void ResultsWarehouse::set_polymN_02(int newValue) {
	_polymN_02 = newValue;
}

void ResultsWarehouse::set_polymS_02(int newValue) {
	_polymS_02 = newValue;
}

void ResultsWarehouse::set_MK_nN(int newValue) {
	_MK_nN = newValue;
}

void ResultsWarehouse::set_MK_nS(int newValue) {
	_MK_nS = newValue;
}


void ResultsWarehouse::increase_class(int classNumber, int increaseBy) {
	_GC_allele_frequency_classes[classNumber] += increaseBy;
}

void ResultsWarehouse::set_specific_Fst(unsigned int firstPopulationId, unsigned int secondPopulationId, double computedFst) {
	_Fst[firstPopulationId][secondPopulationId] = computedFst;
}

void ResultsWarehouse::set_specific_individual_heterozygosity(string individualName, double computedValue) {
	_individual_heterozygosity[individualName] = computedValue;
}

void ResultsWarehouse::set_specific_distance_object(std::string coupleName, ppc::DistanceBetweenIndividuals* pointerToDistanceObject) {
	_distances[coupleName] = pointerToDistanceObject;
}


void ResultsWarehouse::store_species_data(SpeciesDataWarehouse* newSpeciesData) {
	_allSpeciesData.push_back(newSpeciesData);
}


/*---------------------------------------*/
/*            Accessors - Get            */
/*---------------------------------------*/

string ResultsWarehouse::get_name() {
	return _name;
}

string ResultsWarehouse::get_status() {
	return _status;
}


unsigned int ResultsWarehouse::get_nb_bi_allelic_SNP() {
	return _nb_bi_allelic_SNP;
}

unsigned int ResultsWarehouse::get_nb_tri_allelic_SNP() {
	return _nb_tri_allelic_SNP;
}

unsigned int ResultsWarehouse::get_nb_quadri_allelic_SNP() {
	return _nb_quadri_allelic_SNP;
}


unsigned int ResultsWarehouse::get_nb_sequences_in_consensus() {
	return _nb_sequences_in_consensus;
}

unsigned int ResultsWarehouse::get_nb_sites_in_consensus() {
	return _nb_sites_in_consensus;
}


double ResultsWarehouse::get_nN() {
	return _nN;
}

double ResultsWarehouse::get_nS() {
	return _nS;
}

double ResultsWarehouse::get_specific_Fst(unsigned int firstPopulationId, unsigned int secondPopulationId) {
	return _Fst[firstPopulationId][secondPopulationId];
}

double ResultsWarehouse::get_specific_individual_heterozygosity(std::string individualName) {
	return _individual_heterozygosity[individualName];
}

DistanceBetweenIndividuals* ResultsWarehouse::get_specific_distance_object(std::string coupleName) {
	return _distances[coupleName];
}


int ResultsWarehouse::get_fixN() {
	return _fixN;
}

int ResultsWarehouse::get_fixS() {
	return _fixS;
}

int ResultsWarehouse::get_polymN_0() {
	return _polymN_0;
}

int ResultsWarehouse::get_polymS_0() {
	return _polymS_0;
}

int ResultsWarehouse::get_polymN_02() {
	return _polymN_02;
}

int ResultsWarehouse::get_polymS_02() {
	return _polymS_02;
}

int ResultsWarehouse::get_MK_nN() {
	return _MK_nN;
}

int ResultsWarehouse::get_MK_nS() {
	return _MK_nS;
}


vector<double> ResultsWarehouse::get_GC_allele_frequency_classes() {
	return _GC_allele_frequency_classes;
}

vector< vector<double> > ResultsWarehouse::get_all_Fst() {
	return _Fst;
}

map <string, double> ResultsWarehouse::get_all_individual_heterozygosity() {
	return _individual_heterozygosity;
}

map <string,DistanceBetweenIndividuals*> ResultsWarehouse::get_all_distance_objects() {
	return _distances;
}


std::vector<SpeciesDataWarehouse*> ResultsWarehouse::get_all_species_data() {
	return _allSpeciesData;
}

SpeciesDataWarehouse* ResultsWarehouse::get_specific_species_data(string requestedSpeciesName) {

	SpeciesDataWarehouse* requestedData = 0;

	for (unsigned int i = 0; i < _allSpeciesData.size(); i ++) {
		if (_allSpeciesData[i]->get_species_name() == requestedSpeciesName) requestedData = _allSpeciesData[i];
	}

	return requestedData;
}
