/*
 * Summary.cpp
 *
 *  Created on: 19 avr. 2013
 *      Author: abernard
 */

#include "Summary.h"

// Bio++ classes
#include <Bpp/Numeric/Random/RandomTools.h>
#include <Bpp/Numeric/VectorTools.h>

// PopPhyl Project Developments
#include "SpeciesDataWarehouse.h"

using namespace std;
using namespace bpp;
using namespace ppc;

/*-------------------------------*/
/*          Constructor          */
/*-------------------------------*/

Summary::Summary(vector<ResultsWarehouse*> allResults, Population* popData, string ingroupSpeciesName, unsigned int nbContig, unsigned int nbResampling, unsigned int sortPrecision, double confidence):
		_total_SNP2(0), _total_SNP3(0), _total_SNP4(0), _nb_contig(nbContig), _average_piN(-1), _average_piS(-1), _average_piNpiS(-1), _average_contig_length_piNpiS(-1),
		_average_fit(-100), _average_WC84_fit(-100), _average_dN(-1), _average_dS(-1), _average_dNdS(-1), _contig_above_threshold(0), _NI(-1), _NI02(-1), _failed_piNpiS(0),
		_failed_fit(0), _failed_WC84_fit(0), _failed_dNdS(0),_failed_NI(0), _piN_standard_error(nbContig, -1.), _piS_standard_error(nbContig, -1.), _piNpiS_standard_error(nbContig, -1.),
		_fit_standard_error(nbContig, -1.), _WC84_fit_standard_error(nbContig, -1.), _dN_standard_error(nbContig, -1.), _dS_standard_error(nbContig, -1.),
		_dNdS_standard_error(nbContig, -1.), _NI_standard_error(nbContig, -1.), _NI02_standard_error(nbContig, -1.), _nb_resampling(nbResampling), _sort_precision(sortPrecision),
		_confidence(confidence), _dS_threshold(-1){

	// Initialize average individual heterozygosity map
	vector<string> allIndividuals = popData->get_all_individuals();

	for (unsigned int i = 0; i < allIndividuals.size(); i++) {
		_average_individual_heterozygosity[allIndividuals[i]] = 0;
	}

	// Initialize average Fst map
	vector<string> coupleNames =  popData->get_all_couples();

	for (unsigned int i = 0; i < coupleNames.size(); i++) {
		_average_population_fst[coupleNames[i]] = 0;
	}

	// Extract needed datas from the warehouses
	initial_data_extraction(allResults, ingroupSpeciesName, popData);
}

/*--------------------------------*/
/*           Destructor           */
/*--------------------------------*/

Summary::~Summary() {
	// TODO Auto-generated destructor stub
}


/*---------------------------------------*/
/*           Protected methods           */
/*---------------------------------------*/

void Summary::initial_data_extraction(vector<ResultsWarehouse*> allDataLines, string ingroupSpeciesName, Population* popData) {

	// Initializations
	_total_SNP2 = _total_SNP3 = _total_SNP4 = 0;

	// Browse all ResultWarehouse (each warehouse corresponds to all results for a given contig) and extract various column
	for (unsigned int i = 0; i < allDataLines.size(); i++) {

		_total_SNP2 += allDataLines[i]->get_nb_bi_allelic_SNP();
		_total_SNP3 += allDataLines[i]->get_nb_tri_allelic_SNP();
		_total_SNP4 += allDataLines[i]->get_nb_quadri_allelic_SNP();

		// Get Focal species data
		SpeciesDataWarehouse* ingroupData = allDataLines[i]->get_specific_species_data(ingroupSpeciesName);

		_piS_column.push_back(ingroupData->get_piS());
		_piN_column.push_back(ingroupData->get_piN());
		_fit_column.push_back(ingroupData->get_Fit());
		_nb_complete_site_column.push_back(ingroupData->get_number_of_complete_sites());

		_WC84_total_B_column.push_back(ingroupData->get_WC84_total_B());
		_WC84_total_C_column.push_back(ingroupData->get_WC84_total_C());

		_nN_column.push_back(allDataLines[i]->get_nN());
		_nS_column.push_back(allDataLines[i]->get_nS());
		_fixN_column.push_back(allDataLines[i]->get_fixN());
		_fixS_column.push_back(allDataLines[i]->get_fixS());

		_polymS00_column.push_back(allDataLines[i]->get_polymS_0());
		_polymN00_column.push_back(allDataLines[i]->get_polymN_0());
		_polymS02_column.push_back(allDataLines[i]->get_polymS_02());
		_polymN02_column.push_back(allDataLines[i]->get_polymN_02());

		complex_data_extraction(allDataLines[i], popData);
	}
}

/******************************************************************************/

void Summary::complex_data_extraction(ResultsWarehouse* DataLine, Population* popData) {


	// Extract all individual heterozygosity data for current contig
	vector<double> currentContigHeteroValues;
	vector<string> allIndividuals = popData->get_all_individuals();

	for (unsigned int i = 0; i < allIndividuals.size(); i++) {
		currentContigHeteroValues.push_back(DataLine->get_specific_individual_heterozygosity(allIndividuals[i]));
	}

	_individual_heterozygosity_columns.push_back(currentContigHeteroValues);


	// Extract all Fst data for current contig
	vector<double> currentContigFstValues;
	unsigned int nbPopulations = popData->get_number_of_population();

	for (unsigned int popIndex1 = 0; popIndex1 < nbPopulations; popIndex1++) {
		for (unsigned int popIndex2 = popIndex1 + 1; popIndex2 < nbPopulations; popIndex2++) {
			currentContigFstValues.push_back(DataLine->get_specific_Fst(popIndex1, popIndex2));
		}
	}

	_population_fst_columns.push_back(currentContigFstValues);
}

/******************************************************************************/

void Summary::generate_boot_dataset(std::vector<unsigned int>& bootDataset) {

	// Initializations
	unsigned int datasetSize = bootDataset.size();

	// Fill the bootDataset
	for (unsigned int i = 0; i < datasetSize; i++) {

		double ran = RandomTools::giveRandomNumberBetweenZeroAndEntry(1.);
		bootDataset[i] = (int)(ran * datasetSize);
		if (bootDataset[i] == datasetSize) bootDataset[i] = datasetSize-1;
	}
}

/******************************************************************************/

vector<double> Summary::get_standard_error(vector<double>& bootData) {

	// Initializations
	vector<int> iboot;
	vector<double> standardError(2);

	double conf2 = 1 - (1 - _confidence) / 2.;
	int nb = bootData.size();

	for (int i = 0; i < nb; i++) {
		iboot.push_back((int)(bootData[i] * _sort_precision));
	}

	std::sort(iboot.begin(), iboot.end());

	for (int k = 0; k < nb; k++) {
		bootData[k] = (double)iboot[k] / _sort_precision;
	}

	standardError[0] = bootData[(int)((1. - conf2) * nb)];
	standardError[1] = bootData[(int)(conf2 * nb)];

	return standardError;
}


/*---------------------------------------*/
/*            Accessors - Get            */
/*---------------------------------------*/

unsigned int Summary::get_total_SNP2() { return _total_SNP2; }
unsigned int Summary::get_total_SNP3() { return _total_SNP3; }
unsigned int Summary::get_total_SNP4() { return _total_SNP4; }
unsigned int Summary::get_total_number_of_SNP() { return _total_SNP2 + _total_SNP3 + _total_SNP4; }
unsigned int Summary::get_nb_contig() { return _nb_contig; }

/******************************************************************************/

double Summary::get_average_piN() { return _average_piN; }
vector<double> Summary::get_piN_standard_error() { return _piN_standard_error; }

double Summary::get_average_piS() { return _average_piS; }
vector<double> Summary::get_piS_standard_error() { return _piS_standard_error; }

double Summary::get_average_piNpiS() { return _average_piNpiS; }
vector<double> Summary::get_piNpiS_standard_error() { return _piNpiS_standard_error; }

unsigned int Summary::get_average_contig_length_piNpiS() { return _average_contig_length_piNpiS; }
unsigned int Summary::get_failed_piNpiS() { return _failed_piNpiS; };

/******************************************************************************/

double Summary::get_average_fit() { return _average_fit; }
vector<double> Summary::get_fit_standard_error() { return _fit_standard_error; }
unsigned int Summary::get_failed_fit() { return _failed_fit; }

/******************************************************************************/

double Summary::get_average_WC84_fit() { return _average_WC84_fit; }
vector<double> Summary::get_WC84_fit_standard_error() { return _WC84_fit_standard_error; }
unsigned int Summary::get_failed_WC84_fit() { return _failed_WC84_fit; }

/******************************************************************************/

double Summary::get_average_dN() { return _average_dN; }
vector<double> Summary::get_dN_standard_error() { return _dN_standard_error; }

double Summary::get_average_dS() { return _average_dS; }
vector<double> Summary::get_dS_standard_error() { return _dS_standard_error; }

double Summary::get_average_dNdS() { return _average_dNdS; }
vector<double> Summary::get_dNdS_standard_error() { return _dNdS_standard_error; }

unsigned int Summary::get_contig_above_threshold() { return _contig_above_threshold; }
unsigned int Summary::get_failed_dNdS() { return _failed_dNdS; }

/******************************************************************************/

double Summary::get_NI() { return _NI; }
vector<double> Summary::get_NI_standard_error() { return _NI_standard_error; }

double Summary::get_NI02() { return _NI02; }
vector<double> Summary::get_NI02_standard_error() { return _NI02_standard_error; }

unsigned int Summary::get_failed_NI() { return _failed_NI; }

/******************************************************************************/

double Summary::get_specific_average_individual_heterozygosity(std::string individualName) { return _average_individual_heterozygosity[individualName]; }
double Summary::get_specific_average_population_fst(std::string coupleName) { return _average_population_fst[coupleName]; }

/******************************************************************************/

double Summary::get_dS_threshold() { return _dS_threshold; }

/******************************************************************************/

double Summary::get_alpha() { return (1 - _NI); }
vector<double> Summary::get_alpha_standard_error() {

	vector<double> alpha_standard_error(2, -1.0);

	alpha_standard_error[0] = 1 - _NI_standard_error[1];
	alpha_standard_error[1] = 1 - _NI_standard_error[0];

	return alpha_standard_error;
}

double Summary::get_alpha02() { return (1 - _NI02); }
vector<double> Summary::get_alpha02_standard_error() {

	vector<double> alpha02_standard_error(2, -1.0);

	alpha02_standard_error[0] = 1 - _NI02_standard_error[1];
	alpha02_standard_error[1] = 1 - _NI02_standard_error[0];

	return alpha02_standard_error;
}


/*------------------------------------*/
/*           Public methods           */
/*------------------------------------*/

void Summary::compute_dS_threshold(double nb_se_allowed) {

	// Initializations
	vector<double> dS;

	// Compute
	for (unsigned int i = 0; i < _fixS_column.size(); i++) {
		dS.push_back((double)_fixS_column[i] / (double)_nS_column[i]);
	}

	_dS_threshold = VectorTools::median(dS) + nb_se_allowed * VectorTools::sd<double, double>(dS, true);
}

/*******************/
/*     Average     */

void Summary::compute_all_averages(Population* popData, double piNpiS_thr, double fit_thr, double WC84_fit_thr, double indH_thr, double fst_thr, double nb_se_allowed, int max_MK_value) {

	compute_average_piNpiS(piNpiS_thr);
	compute_average_fit(fit_thr);
	compute_average_WC84_fit(WC84_fit_thr);
	compute_average_dNdS(max_MK_value, nb_se_allowed);
	compute_average_NI(max_MK_value, nb_se_allowed);
	compute_average_individual_heterozygosity(popData, indH_thr);
	compute_average_population_fst(popData, fst_thr);
}

/******************************************************************************/

void Summary::compute_average_piNpiS(double threshold) {

	// Initializations
	unsigned int totalNbCompleteSite(0);
	_average_piS = _average_piN = _average_piNpiS = _average_contig_length_piNpiS = 0;

	// Compute average piN, piS, piNpiS
	for (unsigned int i = 0; i < _nb_contig; i++) {
		if (_piS_column[i] < threshold) {
			_failed_piNpiS++;
			continue;
		}

		_average_piS += _piS_column[i] * _nb_complete_site_column[i];
		_average_piN += _piN_column[i] * _nb_complete_site_column[i];
		totalNbCompleteSite += _nb_complete_site_column[i];
	}

	_average_piS /= totalNbCompleteSite;
	_average_piN /= totalNbCompleteSite;
	_average_piNpiS = _average_piN / _average_piS;
	_average_contig_length_piNpiS = totalNbCompleteSite / (_nb_contig - _failed_piNpiS); // Mean contig length
}

/******************************************************************************/

void Summary::compute_average_fit(double threshold) {

	// Initializations
	unsigned int totalNbCompleteSite(0);
	_average_fit = 0;

	// Compute average fit
	for (unsigned int  i = 0; i < _nb_contig; i++) {
		if (_fit_column[i] < threshold) {
			_failed_fit++;
			continue;
		}

		_average_fit += _fit_column[i] * _nb_complete_site_column[i];
		totalNbCompleteSite += _nb_complete_site_column[i];
	}

	_average_fit /= totalNbCompleteSite;
}

/******************************************************************************/

void Summary::compute_average_WC84_fit(double threshold) {

	// Initializations
	double WC84_total_B(0.), WC84_total_C(0.);

	// Compute average fit
	for (unsigned int  i = 0; i < _nb_contig; i++) {
		if (_WC84_total_B_column[i] < threshold || _WC84_total_C_column[i] < threshold) {
			_failed_WC84_fit++;
			continue;
		}

		WC84_total_B += _WC84_total_B_column[i];
		WC84_total_C += _WC84_total_C_column[i];
	}

	_average_WC84_fit = WC84_total_B / (WC84_total_B + WC84_total_C);
}

/******************************************************************************/

void Summary::compute_average_dNdS(int max_MK_value, double nb_se_allowed) {

	// Initializations
	int totnN(0), totnS(0), totfixN(0), totfixS(0);

	// Compute dS threshold if needed
	if (_dS_threshold == -1.0) compute_dS_threshold(nb_se_allowed);

	// Compute average dN,dS,dNdS
	for (unsigned int  i = 0; i < _nb_contig; i++) {

		if (_fixN_column[i] > max_MK_value || _fixS_column[i] > max_MK_value) {
			_failed_dNdS++;
			continue;
		}

		if ((double)_fixS_column[i] / _nS_column[i] > _dS_threshold) {
			_failed_dNdS++;
			_contig_above_threshold++;
			continue;
		}

		totfixN += _fixN_column[i];
		totfixS += _fixS_column[i];
		totnN += _nN_column[i];
		totnS += _nS_column[i];
	}

	_average_dN = (double)totfixN / (double)totnN;
	_average_dS = (double)totfixS / (double)totnS;
	_average_dNdS = _average_dN / _average_dS;

}

/******************************************************************************/

void Summary::compute_average_NI(int max_MK_value, double nb_se_allowed) {

	// Initializations
	int totfixN(0), totfixS(0), totpN00(0), totpS00(0), totpS02(0), totpN02(0);

	// Compute dS threshold if needed
	if (_dS_threshold == -1.0) compute_dS_threshold(nb_se_allowed);

	// Compute average NI and NI0.2
	for (unsigned int  i = 0; i < _nb_contig; i++) {

		if (_fixN_column[i] > max_MK_value || _fixS_column[i] > max_MK_value || _polymN00_column[i] > max_MK_value || _polymS00_column[i] > max_MK_value) {
			_failed_NI++;
			continue;
		}

		if ((double)_fixS_column[i] / _nS_column[i] > _dS_threshold) {
			_failed_NI++;
			continue;
		}

		totfixN += _fixN_column[i];
		totfixS += _fixS_column[i];
		totpN00 += _polymN00_column[i];
		totpS00 += _polymS00_column[i];
		totpN02 += _polymN02_column[i];
		totpS02 += _polymS02_column[i];
	}

	_NI = (double)(totpN00 * totfixS) / (double)(totpS00 * totfixN);
	_NI02 = (double)(totpN02 * totfixS) / (double)(totpS02 * totfixN);
}

/******************************************************************************/

void Summary::compute_average_individual_heterozygosity(Population* popData, double threshold) {

	// Initializations
	vector<string> allIndividuals = popData->get_all_individuals();
	unsigned int nbIndividuals= popData->get_number_of_individuals();
	vector<int> failedIndividualHeterozygosity (nbIndividuals, 0);
	vector<int> totalNbCompleteSite (nbIndividuals, 0);

	// Compute the average heterozygosity for each individual
	for (unsigned int  i = 0; i < _nb_contig; i++) {
		for (unsigned int  j = 0; j < nbIndividuals; j++) {
			if (_individual_heterozygosity_columns[i][j] > -100.) {
				_average_individual_heterozygosity[allIndividuals[j]] += _individual_heterozygosity_columns[i][j] * _nb_complete_site_column[i];
				totalNbCompleteSite[j] += _nb_complete_site_column[i];
			} else {
				failedIndividualHeterozygosity[j]++;
			}
		}
	}

	// Weighted average
	for (unsigned int  k = 0; k < nbIndividuals; k++) {
		_average_individual_heterozygosity[allIndividuals[k]] /= totalNbCompleteSite[k];
	}
}

/******************************************************************************/

void Summary::compute_average_population_fst(Population* popData, double threshold) {

	// Initializations
	vector<string> coupleNames =  popData->get_all_couples();
	unsigned int nbCouples = coupleNames.size();
	vector<int> failedFst (nbCouples, 0);
	vector<int> totalNbCompleteSite (nbCouples, 0);

	if (coupleNames.size() > 0) {

		// Compute the average Fst for each couple of population
		for (unsigned int  i = 0; i < _nb_contig; i++) {
			for (unsigned int j = 0; j < nbCouples; j++) {
				if (_population_fst_columns[i][j] > -100.) {
					_average_population_fst[coupleNames[j]] += _population_fst_columns[i][j] * _nb_complete_site_column[i];
					totalNbCompleteSite[j] += _nb_complete_site_column[i];
				} else {
					failedFst[j]++;
				}
			}
		}

		// Weighted average
		for (unsigned int  k = 0; k < nbCouples; k++) {
			if (totalNbCompleteSite[k] != 0) {
				_average_population_fst[coupleNames[k]] /= totalNbCompleteSite[k];
			} else {
				_average_population_fst[coupleNames[k]] = 0;
			}
		}
	}
}


/********************/
/*     Boostrap     */

void Summary::bootstrap_everything(double piNpiS_thr, double fit_thr, double WC84_fit_thr, double nb_se_allowed, int max_MK_value) {

	bootstrap_piNpiS(piNpiS_thr);
	bootstrap_fit(fit_thr);
	bootstrap_WC84_fit(WC84_fit_thr);
	bootstrap_dNdS(max_MK_value, nb_se_allowed);
	bootstrap_NI(max_MK_value, nb_se_allowed);
}

/******************************************************************************/

void Summary::bootstrap_piNpiS(double threshold) {

	// Initializations
	unsigned int totalNbCompleteSite(0), id(0);
	vector<unsigned int> bootDataset(_nb_contig, 0);
	vector<double> boot_piS(_nb_resampling, 0.0);
	vector<double> boot_piN(_nb_resampling, 0.0);
	vector<double> boot_piNpiS(_nb_resampling, 0.0);

	// Bootstrap
	for (unsigned int k = 0; k < _nb_resampling; k++) {
		// Generate a boot dataset/sample if needed
		generate_boot_dataset(bootDataset);

		// Reinitialize values
		totalNbCompleteSite = 0;

		for (unsigned int i = 0; i < _nb_contig; i++) {
			id = bootDataset[i];
			if (_piS_column[id] < threshold) continue;

			boot_piS[k] += _piS_column[id] * _nb_complete_site_column[id];
			boot_piN[k] += _piN_column[id] * _nb_complete_site_column[id];
			totalNbCompleteSite += _nb_complete_site_column[id];
		}

		boot_piS[k] /= totalNbCompleteSite;
		boot_piN[k] /= totalNbCompleteSite;
		boot_piNpiS[k] = boot_piN[k] / boot_piS[k];
	}

	// Get standard error from bootstrap data
	_piS_standard_error = get_standard_error(boot_piS);
	_piN_standard_error = get_standard_error(boot_piN);
	_piNpiS_standard_error = get_standard_error(boot_piNpiS);
}

/******************************************************************************/

void Summary::bootstrap_fit(double threshold) {

	// Initializations
	unsigned int totalNbCompleteSite(0), id(0);
	vector<unsigned int> bootDataset(_nb_contig, 0);
	vector<double> boot_fit(_nb_resampling, 0.0);

	// Bootstrap
	for (unsigned int k = 0; k < _nb_resampling; k++) {
		// Generate a boot dataset/sample if needed
		generate_boot_dataset(bootDataset);

		// Reinitialize values
		totalNbCompleteSite = 0;

		for (unsigned int i = 0; i < _nb_contig; i++) {
			id = bootDataset[i];
			if (_fit_column[id] < threshold) continue;

			boot_fit[k] += _fit_column[id] * _nb_complete_site_column[id];
			totalNbCompleteSite += _nb_complete_site_column[id];
		}

		boot_fit[k] /= totalNbCompleteSite;
	}

	// Get standard error from bootstrap data
	_fit_standard_error = get_standard_error(boot_fit);
}

/******************************************************************************/

void Summary::bootstrap_WC84_fit(double threshold) {

	// Initializations
	unsigned int id(0);
	double WC84_total_B(0.), WC84_total_C(0.);
	vector<unsigned int> bootDataset(_nb_contig, 0);
	vector<double> boot_WC84_fit(_nb_resampling, 0.0);

	// Bootstrap
	for (unsigned int k = 0; k < _nb_resampling; k++) {
		// Generate a boot dataset/sample if needed
		generate_boot_dataset(bootDataset);

		// Reinitialize values
		WC84_total_B = 0;
		WC84_total_C = 0;

		for (unsigned int i = 0; i < _nb_contig; i++) {
			id = bootDataset[i];

			if (_WC84_total_B_column[id] < threshold || _WC84_total_C_column[id] < threshold) continue;

			WC84_total_B += _WC84_total_B_column[id];
			WC84_total_C += _WC84_total_C_column[id];
		}

		boot_WC84_fit[k] = WC84_total_B / (WC84_total_B + WC84_total_C);
	}

	// Get standard error from bootstrap data
	_WC84_fit_standard_error = get_standard_error(boot_WC84_fit);
}

/******************************************************************************/

void Summary::bootstrap_dNdS(int max_MK_value, double nb_se_allowed) {

	// Initializations
	vector<unsigned int> bootDataset(_nb_contig, 0);
	vector<double> boot_dN(_nb_resampling, 0.0);
	vector<double> boot_dS(_nb_resampling, 0.0);
	vector<double> boot_dNdS(_nb_resampling, 0.0);
	unsigned int id(0);
	int boot_totnN(0), boot_totnS(0), boot_totfixN(0), boot_totfixS(0);

	// Compute dS threshold if needed
	if (_dS_threshold == -1.0) compute_dS_threshold(nb_se_allowed);

	// Bootstrap
	for (unsigned int k = 0; k < _nb_resampling; k++) {
		// Generate a boot dataset/sample if needed
		generate_boot_dataset(bootDataset);

		// Reinitialize values
		boot_totnN = boot_totnS = boot_totfixN = boot_totfixS = 0;

		for (unsigned int i = 0; i < _nb_contig; i++) {
			id = bootDataset[i];

			if (_fixN_column[id] > max_MK_value || _fixS_column[id] > max_MK_value) continue;
			if ((double)_fixS_column[id] / _nS_column[id] > _dS_threshold) continue;

			boot_totfixN += _fixN_column[id];
			boot_totfixS += _fixS_column[id];
			boot_totnN += _nN_column[id];
			boot_totnS += _nS_column[id];
		}

		boot_dN[k] = (double)boot_totfixN / (double)boot_totnN;
		boot_dS[k] = (double)boot_totfixS / (double)boot_totnS;
		boot_dNdS[k] = boot_dN[k] / boot_dS[k];
	}

	// Get standard error from bootstrap data
	_dN_standard_error = get_standard_error(boot_dN);
	_dS_standard_error = get_standard_error(boot_dS);
	_dNdS_standard_error = get_standard_error(boot_dNdS);
}

/******************************************************************************/

void Summary::bootstrap_NI(int max_MK_value, double nb_se_allowed) {

	// Initializations
	vector<unsigned int> bootDataset(_nb_contig, 0);
	vector<double> boot_NI(_nb_resampling, 0.0);
	vector<double> boot_NI02(_nb_resampling, 0.0);
	unsigned int id(0);
	int boot_totfixN(0), boot_totfixS(0), boot_totpN00(0), boot_totpS00(0), boot_totpS02(0), boot_totpN02(0);

	// Compute dS threshold if needed
	if (_dS_threshold == -1.0) compute_dS_threshold(nb_se_allowed);

	// Bootstrap
	for (unsigned int k = 0; k < _nb_resampling; k++) {
		// Generate a boot dataset/sample if needed
		generate_boot_dataset(bootDataset);

		// Reinitialize values
		boot_totfixN = boot_totfixS = boot_totpN00 = boot_totpS00 = boot_totpS02 = boot_totpN02 = 0;

		for (unsigned int i = 0; i < _nb_contig; i++) {
			id = bootDataset[i];

			if (_fixN_column[id] > max_MK_value || _fixS_column[id] > max_MK_value || _polymN00_column[i] > max_MK_value || _polymS00_column[i] > max_MK_value) continue;
			if ((double)_fixS_column[id] / _nS_column[id] > _dS_threshold) continue;

			boot_totfixN += _fixN_column[id];
			boot_totfixS += _fixS_column[id];
			boot_totpN00 += _polymN00_column[i];
			boot_totpS00 += _polymS00_column[i];
			boot_totpN02 += _polymN02_column[i];
			boot_totpS02 += _polymS02_column[i];
		}

		boot_NI[k] = (double)(boot_totpN00 * boot_totfixS) / (double)(boot_totpS00 * boot_totfixN);
		boot_NI02[k] = (double)(boot_totpN02 * boot_totfixS) / (double)(boot_totpS02 * boot_totfixN);
	}

	// Get standard error from bootstrap data
	_NI_standard_error = get_standard_error(boot_NI);
	_NI02_standard_error = get_standard_error(boot_NI02);
}
