/*
 * Summary.h
 *
 *  Created on: 19 avr. 2013
 *      Author: abernard
 */

#ifndef SUMMARY_H_
#define SUMMARY_H_

// Basic C++ classes
#include <string>
#include <vector>
#include <map>

// PopPhyl Project Developments
#include <ppc/PopGen/Population.h>
#include "ResultsWarehouse.h"

class Summary {

	protected:
		// Columns
		std::vector<double> _piN_column;
		std::vector<double> _piS_column;
		std::vector<double> _fit_column;

		std::vector<double> _WC84_total_B_column;
		std::vector<double> _WC84_total_C_column;

		std::vector<int> _nb_complete_site_column;
		std::vector<int> _nS_column;
		std::vector<int> _nN_column;
		std::vector<int> _fixS_column;
		std::vector<int> _fixN_column;
		std::vector<int> _polymS00_column;
		std::vector<int> _polymN00_column;
		std::vector<int> _polymS02_column;
		std::vector<int> _polymN02_column;

		std::vector< std::vector<double> > _individual_heterozygosity_columns;
		std::vector< std::vector<double> > _population_fst_columns;

		// Sum
		unsigned int _total_SNP2;
		unsigned int _total_SNP3;
		unsigned int _total_SNP4;
		unsigned int _nb_contig;

		// Averages and other attributes
		double _average_piN;
		double _average_piS;
		double _average_piNpiS;
		unsigned int _average_contig_length_piNpiS;

		double _average_fit;
		double _average_WC84_fit;

		double _average_dN;
		double _average_dS;
		double _average_dNdS;
		unsigned int _contig_above_threshold;

		double _NI;
		double _NI02;

		// Failed counter
		unsigned int _failed_piNpiS;
		unsigned int _failed_fit;
		unsigned int _failed_WC84_fit;
		unsigned int _failed_dNdS;
		unsigned int _failed_NI;

		// standard errors
		std::vector<double> _piN_standard_error;
		std::vector<double> _piS_standard_error;
		std::vector<double> _piNpiS_standard_error;

		std::vector<double> _fit_standard_error;
		std::vector<double> _WC84_fit_standard_error;

		std::vector<double> _dN_standard_error;
		std::vector<double> _dS_standard_error;
		std::vector<double> _dNdS_standard_error;

		std::vector<double> _NI_standard_error;
		std::vector<double> _NI02_standard_error;

		// Bootstrap
		unsigned int _nb_resampling;
		unsigned int _sort_precision;
		double _confidence;

		// Thresholds
		double _dS_threshold;

		// Individual Heterozygosity and Fst
		std::map <std::string, double> _average_individual_heterozygosity;
		std::map <std::string, double> _average_population_fst;

	public:
		// Constructors
		Summary(std::vector<ResultsWarehouse*> allResults, ppc::Population* popData, std::string ingroupSpeciesName, unsigned int nbContig, unsigned int nbResampling, unsigned int sortPrecision, double confidence);

		// Destructors
		virtual ~Summary();

		// Accessors - Get
		unsigned int get_total_SNP2();
		unsigned int get_total_SNP3();
		unsigned int get_total_SNP4();
		unsigned int get_total_number_of_SNP();
		unsigned int get_nb_contig();

		double get_average_piN();
		double get_average_piS();
		double get_average_piNpiS();
		unsigned int get_average_contig_length_piNpiS();

		double get_average_fit();
		double get_average_WC84_fit();

		double get_average_dN();
		double get_average_dS();
		double get_average_dNdS();
		unsigned int get_contig_above_threshold();

		double get_NI();
		double get_NI02();

		double get_alpha();
		double get_alpha02();

		unsigned int get_failed_piNpiS();
		unsigned int get_failed_fit();
		unsigned int get_failed_WC84_fit();
		unsigned int get_failed_dNdS();
		unsigned int get_failed_NI();

		std::vector<double> get_piN_standard_error();
		std::vector<double> get_piS_standard_error();
		std::vector<double> get_piNpiS_standard_error();

		std::vector<double> get_fit_standard_error();
		std::vector<double> get_WC84_fit_standard_error();

		std::vector<double> get_dN_standard_error();
		std::vector<double> get_dS_standard_error();
		std::vector<double> get_dNdS_standard_error();

		std::vector<double> get_NI_standard_error();
		std::vector<double> get_NI02_standard_error();

		double get_specific_average_individual_heterozygosity(std::string individualName);
		double get_specific_average_population_fst(std::string coupleName);

		double get_dS_threshold();

		// Derived get methods
		std::vector<double> get_alpha_standard_error();
		std::vector<double> get_alpha02_standard_error();

		// Computing methods
		void compute_all_averages(ppc::Population* popData, double piNpiS_thr = -0.1, double fit_thr = -100.0, double WC84_fit_thr = -900.0, double indH_thr = -100.0, double fst_thr = -100.0, double nb_se_allowed = 2, int max_MK_value = 1000);
		void bootstrap_everything(double piNpiS_thr = -0.1, double fit_thr = -100.0, double WC84_fit_thr = -900.0, double nb_se_allowed = 2, int max_MK_value = 1000);

		void compute_average_piNpiS(double threshold = -0.1);
		void bootstrap_piNpiS(double threshold = -0.1);

		void compute_average_fit(double threshold = -100.0);
		void bootstrap_fit(double threshold = -100.0);

		void compute_average_WC84_fit(double threshold = -100.0);
		void bootstrap_WC84_fit(double threshold = -100.0);

		void compute_dS_threshold(double nb_se_allowed = 2);

		void compute_average_dNdS(int max_MK_value = 1000, double nb_se_allowed = 2);
		void bootstrap_dNdS(int max_MK_value = 1000, double nb_se_allowed = 2);

		void compute_average_NI(int max_MK_value = 1000, double nb_se_allowed = 2);
		void bootstrap_NI(int max_MK_value = 1000, double nb_se_allowed = 2);

		void compute_average_individual_heterozygosity(ppc::Population* popData, double threshold = -100.0);
		void compute_average_population_fst(ppc::Population* popData, double threshold = -100.0);

	protected:
		void initial_data_extraction(std::vector<ResultsWarehouse*> results, std::string ingroupSpecies, ppc::Population* popData);
		void complex_data_extraction(ResultsWarehouse* result, ppc::Population* popData);
		void generate_boot_dataset(std::vector<unsigned int>& bootDataset);
		std::vector<double> get_standard_error(std::vector<double>& bootData);
};

#endif /* SUMMARY_H_ */
