/*
 * dNdSpiNpiS.h
 *
 *  Author: Nicolas Galtier, Aurélien Bernard
 */

/*--------------------------------------*/
/*               Includes               */
/*--------------------------------------*/

// Basic C++ classes
#include <map>

// Bio++ - Population genetics containers and tools
#include <Bpp/PopGen/PolymorphismSequenceContainerTools.h>
#include <Bpp/PopGen/SequenceStatistics.h>

// Bio++ - Math
#include <Bpp/Numeric/Random/RandomTools.h>

// PopPhyl Project Developments - Alphabets
#include <ppc/Seq/Alphabet/CustomNucleicAlphabet.h>
#include <ppc/Seq/Alphabet/FrameshiftCodonAlphabet.h>

// PopPhyl Project Developments - Alphabets
#include <ppc/PopGen/Tools/AlignmentFileManagement.h>
#include <ppc/PopGen/Tools/PolymorphismSequenceContainerAddons.h>
#include <ppc/PopGen/PopulationDiversification.h>
#include <ppc/PopGen/SelectionIndicators.h>
#include <ppc/PopGen/Polymorphism.h>
#include <ppc/PopGen/DistanceBetweenIndividuals.h>

// Custom classes
#include "dNdSpiNpiS.h"

/* Using of Namespaces to avoid the heavy Class::Member_function notation */
using namespace std;
using namespace bpp;
using namespace ppc;

/* Debug and versioning */
const double version = 3.0;
const string version_date = "12-06-2013";

// Time
time_t rawtime = time(NULL);

/*--------------------------------------------*/
/*          Main and Contig analysis          */
/*--------------------------------------------*/

int main(int argc, char** argv) {

	// Alphabet and Genetic code objects creation
	NucleicAlphabet* customNucleoAlphabet = new CustomNucleicAlphabet(); // CustomNucleicAlphabet is an alphabet (derived from the Bio++ DNA alphabet) that can deal with the frameshift (!) character. PopPhyl_Commons is therefore required.
	FrameshiftCodonAlphabet* codonAlphabet = new FrameshiftCodonAlphabet(customNucleoAlphabet);
	StandardGeneticCode* stdGeneticCode = new StandardGeneticCode(customNucleoAlphabet);

	Population* populationData = 0;

	// Command line parameters management
	ParametersManagement* dNdSpiNpiS_params = new ParametersManagement(version, version_date, argc, argv);
	dNdSpiNpiS_params->checkParameters();

	// Welcome message
	displayWelcomeMessage(dNdSpiNpiS_params);

	// Initialize and fill the vector of GC allele frequency class
	vector<double> alleleFrequencyClass = setAlleleFrequencyClass(dNdSpiNpiS_params);

	// Get populations from the input population file
	if (dNdSpiNpiS_params->get_population_file() != "none") {
		populationData = loadPopulationFile(dNdSpiNpiS_params->get_population_file());
	}

	// Load the alignment file (Multi Pseudofasta) into a vector of PolymorphismSequenceContainer and clean it
	vector<PolymorphismSequenceContainer*> brutDatasets = loadAlignmentFile(dNdSpiNpiS_params, codonAlphabet, customNucleoAlphabet);

	// Generate a basic population if none has been specified by the user
	if (dNdSpiNpiS_params->get_population_file() == "none") {
		populationData = generateBasicPopulation(dNdSpiNpiS_params, brutDatasets);
	}

	// Clean collected alignment data
	vector<PolymorphismSequenceContainer*> cleanDatasets = cleanAlignmentData(dNdSpiNpiS_params, brutDatasets, codonAlphabet);

	// Create a vector of warehouse pointer
	vector<ResultsWarehouse*> allDatasetResults;

	// Work on each clean dataset
	cout << "Dataset/Contig analysis procedure:" << endl << endl;
	cout << " -> Number of cleaned dataset: " << cleanDatasets.size() << endl << endl;

	time_t process_start = time(NULL);
	cout << "   --> Begin: " << ctime(&process_start);

	for (unsigned int i = 0; i < cleanDatasets.size(); i++) {

		ResultsWarehouse* currentDatasetResults = analyseDataset(dNdSpiNpiS_params, cleanDatasets[i], populationData, codonAlphabet, customNucleoAlphabet, stdGeneticCode, alleleFrequencyClass);
		allDatasetResults.push_back(currentDatasetResults);

		// Print all the content of the result warehouse in debug mode
		if (dNdSpiNpiS_params->get_DEBUG()) ResultsWarehouse::warehouseInventory(currentDatasetResults, dNdSpiNpiS_params, populationData, alleleFrequencyClass);
	}

	time_t process_stop = time(NULL);
	cout << "   --> Stop: " << ctime(&process_stop) << endl;

	// Write all results into the global output file
	writeMainOutputFile(dNdSpiNpiS_params, allDatasetResults, populationData, alleleFrequencyClass);

	// Write distance file if needed
	if (dNdSpiNpiS_params->get_compute_distances()) writeDistanceFile(dNdSpiNpiS_params, allDatasetResults, populationData);

	// Generate and write the Biological summary
	Summary* bioSum = biologicalSummary(allDatasetResults, dNdSpiNpiS_params, populationData);
	writeSummaryFile(dNdSpiNpiS_params, bioSum, populationData);

	// Goodbye message
	displayGoodbyeMessage();
}

/******************************************************************************/

ResultsWarehouse* analyseDataset (ParametersManagement* dNdSpiNpiS_params, PolymorphismSequenceContainer* currentDataset, Population* populationData,
		const StandardCodonAlphabet* codonAlphabet, const NucleicAlphabet* customNucleoAlphabet, StandardGeneticCode* stdGeneticCode, vector<double>& alleleFrequencyClass) {

	// Create objects to store all the results for current dataset
	ResultsWarehouse* datasetResults = new ResultsWarehouse(populationData, dNdSpiNpiS_params->get_nb_gc_bin());
	datasetResults->set_name(PolymorphismSequenceContainerAddons::getContigNameFromContainer(currentDataset, 0));

	SpeciesDataWarehouse* ingroupSpeciesData = new SpeciesDataWarehouse();

	// Get the number of sequence
	map<unsigned int, unsigned int> nbMembersByGroup = PolymorphismSequenceContainerAddons::getNumberOfMembers(currentDataset, dNdSpiNpiS_params->get_DEBUG());

	// Focal species analysis
	if (nbMembersByGroup[0] > 0) {
		// Prepare objects and containers for the focal species
		PolymorphismSequenceContainer* ingroupContainer = PolymorphismSequenceContainerTools::extractIngroup(*currentDataset);

		const Sequence* ingroupConsensusSequence = SiteContainerTools::getConsensus(*ingroupContainer, "Ingroup_consensus", true ,false);

		// Execute all the analysis for the focal species
		analyseFocalSpecies(dNdSpiNpiS_params, ingroupContainer, ingroupConsensusSequence, populationData, codonAlphabet, customNucleoAlphabet, stdGeneticCode, alleleFrequencyClass, ingroupSpeciesData, datasetResults);

		// If the user has specified an outgroup species name
		if (dNdSpiNpiS_params->get_outgroup_species_name() != "none" && nbMembersByGroup.size() > 1 && nbMembersByGroup[1] > 0) {
			// Prepare objects and containers for the main outgroup species
			PolymorphismSequenceContainer* outgroupContainer = PolymorphismSequenceContainerTools::extractGroup(*currentDataset, 1);

			const Sequence* outgroupConsensusSequence = SiteContainerTools::getConsensus(*outgroupContainer, "Outgroup_consensus", true ,false);
			PolymorphismSequenceContainer* outgroupConsensusContainer = new PolymorphismSequenceContainer(codonAlphabet);
			outgroupConsensusContainer->addSequence(*outgroupConsensusSequence);

			/* Compute selection indicators nN and nS */
			collectSelectionIndicators(ingroupConsensusSequence, outgroupConsensusSequence, codonAlphabet, stdGeneticCode, datasetResults, dNdSpiNpiS_params->get_kappa());

			/* MK analysis */
			mcDonaldKreitmanAnalysis(ingroupContainer, outgroupConsensusContainer, datasetResults, dNdSpiNpiS_params->get_kappa(), codonAlphabet, stdGeneticCode, dNdSpiNpiS_params->get_DEBUG());
		}
	}

	// If we are here then everything goes well for the current dataset
	datasetResults->set_status("OK");

	// Add dataset/contigs results to the list of results
	datasetResults->store_species_data(ingroupSpeciesData);

	return datasetResults;
}

/******************************************************************************/

void analyseFocalSpecies(ParametersManagement* dNdSpiNpiS_params, PolymorphismSequenceContainer* ingroupContainer, const Sequence* ingroupConsensusSequence, Population* populationData,
		const StandardCodonAlphabet* codonAlphabet, const NucleicAlphabet* customNucleoAlphabet, StandardGeneticCode* stdGeneticCode,
		vector<double>& alleleFrequencyClass, SpeciesDataWarehouse* ingroupSpeciesData, ResultsWarehouse* datasetResults) {

	// Build ingroup consensus container
	PolymorphismSequenceContainer* ingroupConsensusContainer = new PolymorphismSequenceContainer(codonAlphabet);
	ingroupConsensusContainer->addSequence(*ingroupConsensusSequence);

	/* POLYMORPHISM ANALYSIS */
	Polymorphism* ingroupPolymorphism = new Polymorphism(ingroupContainer);
	ingroupPolymorphism->determine_all_attributes(stdGeneticCode, dNdSpiNpiS_params->get_min_nb_codon(), dNdSpiNpiS_params->get_kappa());

	// Save some data into the species data warehouse
	ingroupSpeciesData->set_species_name(PolymorphismSequenceContainerAddons::getSpeciesNameFromContainer(ingroupContainer, 0));
	ingroupSpeciesData->set_number_of_sequences(ingroupContainer->getNumberOfSequences());
	ingroupSpeciesData->set_number_of_sites(ingroupContainer->getNumberOfSites());
	ingroupSpeciesData->set_number_of_complete_sites(PolymorphismSequenceContainerTools::getNumberOfCompleteSites(*ingroupContainer, false));
	ingroupSpeciesData->set_piN(ingroupPolymorphism->get_piN());
	ingroupSpeciesData->set_piS(ingroupPolymorphism->get_piS());

	/* GC3 */
	computeGC3(ingroupConsensusContainer, ingroupSpeciesData, codonAlphabet, dNdSpiNpiS_params->get_DEBUG());

	/* Individual Heterozygosity analysis */
	PairedSequenceContainer* ingroupPairs = new PairedSequenceContainer(ingroupContainer, codonAlphabet, "|", 3);
	vector<string> allPairNames = ingroupPairs->get_pair_names();
	computeIndividualHeterozygosity(ingroupPairs, codonAlphabet, customNucleoAlphabet, datasetResults, allPairNames);

	// Compute distance between individuals if needed
	if (dNdSpiNpiS_params->get_compute_distances()) {
		computeAllDistancesBetweenIndividuals(datasetResults, ingroupPairs, codonAlphabet, customNucleoAlphabet, dNdSpiNpiS_params->get_DEBUG());
	}

	/* Focal species Fit computation */
	computeFit(ingroupPairs, codonAlphabet, customNucleoAlphabet, ingroupSpeciesData, allPairNames, ingroupContainer->getNumberOfSites(), dNdSpiNpiS_params->get_DEBUG());
	computeWeirCockerhamFit(ingroupPairs, codonAlphabet, customNucleoAlphabet, ingroupSpeciesData, allPairNames, ingroupContainer->getNumberOfSites(), dNdSpiNpiS_params->get_DEBUG());

	 /* Population diversification analysis */
	populationDiversificationAnalysis(ingroupContainer, populationData, codonAlphabet, customNucleoAlphabet, datasetResults);

	/* Non oriented GC analysis */
	nonOrientedGCAnalysis(ingroupContainer, alleleFrequencyClass, datasetResults, codonAlphabet, customNucleoAlphabet, stdGeneticCode);
}

/*------------------------------------*/
/*       Unsorted basic methods       */
/*------------------------------------*/

vector<double> setAlleleFrequencyClass (ParametersManagement* allParams) {

	// Initializations
	unsigned int nbBin = allParams->get_nb_gc_bin();
	double classSize = 1.0 / nbBin;

	vector<double> alleleFrequencyClusters (nbBin, classSize);

	for (unsigned int i = 0; i <= nbBin; i++) alleleFrequencyClusters[i] = (i+1) * classSize;

	return alleleFrequencyClusters;
}

/******************************************************************************/

double determineWeirCockerhamTermB (double sampleSize, double observedFrequency, double observedHeterozygosity, bool debugMode) {

	// Split the calculation into subparts
	double subTerm1 = (sampleSize / (sampleSize - 1));
	double subTerm2 = observedFrequency * (1 - observedFrequency);
	double subTerm3 = ((2 * sampleSize -1) / (4 * sampleSize)) * observedHeterozygosity;

	if (debugMode) cout << "(DEBUG - determineWeirCockerhamTerm) subTerm1: " << subTerm1 << endl;
	if (debugMode) cout << "(DEBUG - determineWeirCockerhamTerm) subTerm2: " << subTerm2 << endl;
	if (debugMode) cout << "(DEBUG - determineWeirCockerhamTerm) subTerm3: " << subTerm3 << endl;

	// Compute B term
	double weirCockerhamTermB = subTerm1 * (subTerm2 - subTerm3);

	if (debugMode) cout << "(DEBUG - determineWeirCockerhamTerm) weirCockerhamTermB: " << weirCockerhamTermB << endl;

	return weirCockerhamTermB;
}

/******************************************************************************/

void replaceUnresolvedCharInCodonSite (Site& SiteToModify, const StandardCodonAlphabet* codonAlphabet, unsigned int nbSequence, unsigned int rankOfAllele2) {
	// Replace gap and N characters (unresolved char) by the letter of the second allele
	for (unsigned int i = 0; i < nbSequence; i++) {
		if (codonAlphabet->isUnresolved(SiteToModify[i])) SiteToModify[i] = SiteToModify[rankOfAllele2];
	}
}

/******************************************************************************/

bool isSynonymousSite (Site& siteToAnalyze, unsigned int nbSequence, const StandardCodonAlphabet* codonAlphabet, const StandardGeneticCode* stdGeneticCode, int allele_2_rank) {

	for (unsigned int i = 0; i < nbSequence; i++) {
		if (codonAlphabet->isUnresolved(siteToAnalyze[i])) siteToAnalyze[i] = siteToAnalyze[allele_2_rank];
	}

	if (CodonSiteTools::numberOfNonSynonymousSubstitutions(siteToAnalyze, *stdGeneticCode) != 0) return false;
	else return true;
}


/*-------------------------------------*/
/*       Display related methods       */
/*-------------------------------------*/

void displayWelcomeMessage (ParametersManagement* dNdS_params) {

	if (dNdS_params->get_DEBUG()) cout << endl << "-- DEBUG MODE IS ACTIVATED --" << endl << endl;
	cout << endl << "###################################################################################" << endl;
	cout << "#          Welcome in dNdSpiNpiS (Version " << version << " (Last update: " << version_date <<"))          #" << endl;
	cout << "###################################################################################" << endl;
	cout << "Start time: " << ctime(&rawtime) << endl;

	cout << "Selected alignment file: " << dNdS_params->get_alignement_file() << endl;
//	cout << "Selected tree file: " << dNdS_params->get_tree_file() << endl;
	cout << "Selected population file: " << dNdS_params->get_population_file() << endl << endl;

	cout << "Selected focal/ingroup species: " << dNdS_params->get_ingroup_species_name() << endl;
	if (dNdS_params->get_outgroup_species_name() != "none") {
		cout << "Selected external/outgroup species: " << dNdS_params->get_outgroup_species_name() << endl << endl;
	} else cout << endl;

	if (dNdS_params->get_remove_frameshift()) {
		cout << "Frameshift-containing contigs will be rejected during the initial cleaning procedure !" << endl;

		cout << "Stop codons will be tolerated in the first/last " << dNdS_params->get_tolerance_zone() << " codon(s) of each sequence." << endl;

		if (dNdS_params->get_allow_internal_bc()) {
			cout << "Internal Stop codons will be tolerated and replaced by NNN codons." << endl;
		} else {
			cout << "Internal Stop codons are forbidden and the corresponding contig will be rejected !" << endl;
		}
	} else {
		cout << "Frameshift-containing contigs will be conserved during the alignment file loading." << endl;

		cout << "Stop and Frameshift codons will be tolerated in the first/last " << dNdS_params->get_tolerance_zone() << " codon(s) of each sequence." << endl;

		if (dNdS_params->get_allow_internal_bc()) {
			cout << "Internal Stop and Frameshift codons will be tolerated and replaced by NNN codons." << endl;
		} else {
			cout << "Internal Stop and Frameshift codons are forbidden and the corresponding contig will be rejected !" << endl;
		}
	}

	if (dNdS_params->get_compute_distances()) {
		cout << endl << "Distance between focal individuals will be computed !" << endl << endl;
	} else cout << endl;

	cout << "The following contigs will be used to generate the Biological summary: " << dNdS_params->get_contig_usage() << " contigs" << endl << endl;
}

/******************************************************************************/

void displayGoodbyeMessage() {

	time_t end_rawtime = time(NULL);
	cout << "Stop time: " << ctime(&end_rawtime);
	cout << "###################################################################################" << endl;
	cout << "#                          dNdSpiNpiS - End of execution                          #" << endl;
	cout << "###################################################################################" << endl;
}


/*---------------------------------------------------------------*/
/*          Input file loading/cleaning related methods          */
/*---------------------------------------------------------------*/

Population* loadPopulationFile (string populationFile) {

	Population* populationData = new Population(populationFile);

	bool containsDuplicates = populationData->check_for_duplicates();

	if (!containsDuplicates) {
		populationData->display_populations();
		return populationData;

	} else {
		populationData->display_duplicates();

		ApplicationTools::displayError("Current dNdSpiNpiS version is not compatible with duplicates-containing populations !");
		displayGoodbyeMessage();
		exit(EXIT_FAILURE);
	}
}

/******************************************************************************/

Population* generateBasicPopulation (ParametersManagement* allParams, vector<PolymorphismSequenceContainer*>& allDatasets) {

	Population* populationData = new Population(allDatasets, allParams->get_ingroup_species_name());

	populationData->display_populations(true);

	return populationData;
}

/******************************************************************************/

vector<PolymorphismSequenceContainer*> loadAlignmentFile (ParametersManagement* allParams, const StandardCodonAlphabet* codonAlphabet, const NucleicAlphabet* customNucleoAlphabet) {

	cout << "Alignment input file analysis:" << endl << endl;

	// Read the alignment file (Multi Pseudofasta) and remove frameshift-containing contigs if requested */
	vector<PolymorphismSequenceContainer*> multidata = AlignmentFileManagement::readAlignmentFile(allParams->get_alignement_file(), allParams->get_remove_frameshift(), customNucleoAlphabet, codonAlphabet, allParams->get_DEBUG());

	// Exit if the number of contig/dataset is not greater than 0
	if (multidata.size() == 0 ) {
		ApplicationTools::displayError("Sorry, there is not enough dataset/contig to perform the analysis ! Execution canceled..");
		displayGoodbyeMessage();
		exit(EXIT_FAILURE);
	}

	return multidata;
}

/******************************************************************************/

vector<PolymorphismSequenceContainer*> cleanAlignmentData (ParametersManagement* allParams, vector<PolymorphismSequenceContainer*>& multidata, const StandardCodonAlphabet* codonAlphabet) {

	// Clean the original multidata (Vector of OrderedSequenceContainer*)
	// Note: During the cleaning procedure, the number of site, sequence, ingroup/outgroup member are checked
	// Dataset that does not pass the checks are eliminated. Therefore their won't be anymore "Not enough data" or "less than two species" status in dNdSpiNpiS main output file

	cout << "Datasets/Contigs cleaning procedure:" << endl << endl;

	vector<PolymorphismSequenceContainer*> multidataClean = AlignmentFileManagement::cleanMultiContainer(multidata, codonAlphabet, allParams->get_ingroup_species_name(),
			allParams->get_min_nb_codon(), allParams->get_tolerance_zone(), allParams->get_allow_internal_bc(), allParams->get_min_gapN_site(),
			allParams->get_max_gapN_seq(), allParams->get_DEBUG(), allParams->get_outgroup_species_name());

	// Exit if the number of remaining contig/dataset is not greater than 0
	if (multidataClean.size() == 0 ) {
		ApplicationTools::displayError("Sorry, there is not enough clean dataset/contig to perform the analysis ! Execution canceled..");
		displayGoodbyeMessage();
		exit(EXIT_FAILURE);
	}

	return multidataClean;
}


/*------------------------------------------------------*/
/*       Paired Sequence analysis related methods       */
/*------------------------------------------------------*/

void computeIndividualHeterozygosity (PairedSequenceContainer* currentPairedContainer, const StandardCodonAlphabet* codonAlphabet, const NucleicAlphabet* nucleoAlphabet,
		ResultsWarehouse* currentWarehouse, vector<string>& allPairNames) {

	// Browse pairs
	for (unsigned int pairNumber = 0; pairNumber < currentPairedContainer->get_nb_pair(); pairNumber++) {

		// Initializations
		unsigned int difference(0), total(0);

		// Get a pair
		PolymorphismSequenceContainer* currentPair = currentPairedContainer->get_pair(pairNumber);

		// Browse codon sites
		for (unsigned int codonSiteNumber = 0; codonSiteNumber < currentPair->getNumberOfSites(); codonSiteNumber++) {

			// Look at each position of the codon site
			for (unsigned int positionNumber = 0; positionNumber < 3; positionNumber++) {

				int intCodeAllele1 = codonAlphabet->getNPosition(currentPair->valueAt(0, codonSiteNumber), positionNumber);
				int intCodeAllele2 = codonAlphabet->getNPosition(currentPair->valueAt(1, codonSiteNumber), positionNumber);

				if (nucleoAlphabet->isUnresolved(intCodeAllele1)) continue;
				if (nucleoAlphabet->isUnresolved(intCodeAllele2)) continue;

				// Increase counters
				total++;
				if (intCodeAllele1 != intCodeAllele2) difference++;
			}
		}

		// Update warehouse if needed
		if (total != 0) {
			currentWarehouse->set_specific_individual_heterozygosity(allPairNames[pairNumber], (double)difference / (double)total);
		}
	}
}

/******************************************************************************/

void computeFit (PairedSequenceContainer* currentPairedContainer, const StandardCodonAlphabet* codonAlphabet, const NucleicAlphabet* nucleoAlphabet,
		SpeciesDataWarehouse* currentWarehouse, vector<string>& allPairNames, unsigned int globalNumberOfSite, bool debugMode) {

	// Initializations
	unsigned int nbObservedHeterozygosity(0), nbBiallelicSite(0);
	double nbExpectedHeterozygosity(0.);

	// Browse codon sites
	for (unsigned int codonSiteNumber = 0; codonSiteNumber < globalNumberOfSite; codonSiteNumber++) {

		// Look at each position of the codon site
		for (unsigned int positionNumber = 0; positionNumber < 3; positionNumber++) {

			// Initializations
			double firstAlleleFrequency(0.), secondAlleleFrequency(0.);

			AlleleFrequency* Allele_data = new AlleleFrequency(currentPairedContainer, codonSiteNumber, positionNumber, codonAlphabet, nucleoAlphabet);

			// Calculate
			if (Allele_data->is_bi_allelic()) {
				firstAlleleFrequency = (double)Allele_data->get_allele_1_occurence() / (2 * Allele_data->get_nb_typed());
				secondAlleleFrequency = (double)Allele_data->get_allele_2_occurence() / (2 * Allele_data->get_nb_typed());

				if (debugMode) {
					cout << "(DEBUG - computeFit) Resolved sequences: " << Allele_data->get_nb_typed() << endl;
					cout << "(DEBUG - computeFit) First allele (" << nucleoAlphabet->getName(Allele_data->get_allele_1_code()) << ") Occurence & frequency: " << Allele_data->get_allele_1_occurence() << " - " << firstAlleleFrequency << endl;
					cout << "(DEBUG - computeFit) Second allele (" << nucleoAlphabet->getName(Allele_data->get_allele_2_code()) << ") Occurence & frequency: " << Allele_data->get_allele_2_occurence() << " - " << secondAlleleFrequency << endl;
				}

				// Increase global values
				nbBiallelicSite++;
				nbObservedHeterozygosity += Allele_data->get_nb_observed_heterozygosity();
				nbExpectedHeterozygosity += 2 * firstAlleleFrequency * secondAlleleFrequency * Allele_data->get_nb_typed();
			}
		}
	}

	if (debugMode) {
		cout << "(DEBUG - computeFit) nbObservedHeterozygosity: " << nbObservedHeterozygosity << endl;
		cout << "(DEBUG - computeFit) nbExpectedHeterozygosity: " << nbExpectedHeterozygosity << endl;
	}

	// Compute and store Fit value
	if (nbBiallelicSite > 0 && nbExpectedHeterozygosity != 0.) {
		currentWarehouse->set_Fit(1. - nbObservedHeterozygosity / nbExpectedHeterozygosity);
	}
}

/******************************************************************************/

void computeWeirCockerhamFit (PairedSequenceContainer* currentPairedContainer, const StandardCodonAlphabet* codonAlphabet, const NucleicAlphabet* nucleoAlphabet,
		SpeciesDataWarehouse* currentWarehouse, vector<string>& allPairNames, unsigned int globalNumberOfSite, bool debugMode) {

	// Initializations
	unsigned int nbBiallelicSite(0);
	double totalWeirCockerhamTermB(0.), totalWeirCockerhamTermC(0.); // bi, ci

	// Browse codon sites
	for (unsigned int codonSiteNumber = 0; codonSiteNumber < globalNumberOfSite; codonSiteNumber++) {

		// Look at each position of the codon site
		for (unsigned int positionNumber = 0; positionNumber < 3; positionNumber++) {

			// Initializations
			double firstAlleleFrequency(0.), observedHeterozygosityInSample(0.), nbResolvedPairs(0.);

			AlleleFrequency* Allele_data = new AlleleFrequency(currentPairedContainer, codonSiteNumber, positionNumber, codonAlphabet, nucleoAlphabet);
			nbResolvedPairs = Allele_data->get_nb_typed();

			// Calculate
			if (Allele_data->is_bi_allelic() && nbResolvedPairs > 1) {
				firstAlleleFrequency = (double)Allele_data->get_allele_1_occurence() / (2 * nbResolvedPairs); // p
				observedHeterozygosityInSample = (double)Allele_data->get_nb_observed_heterozygosity() / nbResolvedPairs; // h

				if (debugMode) {
					cout << "(DEBUG - computeWeirCockerhamFit) Resolved Pairs: " << nbResolvedPairs << endl;
					cout << "(DEBUG - computeWeirCockerhamFit) First allele (" << nucleoAlphabet->getName(Allele_data->get_allele_1_code()) << ") Occurence & frequency: " << Allele_data->get_allele_1_occurence() << " - " << firstAlleleFrequency << endl;
				}

				// Increase global values
				nbBiallelicSite++;
				totalWeirCockerhamTermB += determineWeirCockerhamTermB(nbResolvedPairs, firstAlleleFrequency, observedHeterozygosityInSample, debugMode); // bi += determineWeirCockerhamTerm(n, p, h)
				totalWeirCockerhamTermC += observedHeterozygosityInSample / 2; // ci += h / 2
			}
		}
	}

	if (debugMode) {
		cout << "(DEBUG - computeWeirCockerhamFit) totalWeirCockerhamTermB: " << totalWeirCockerhamTermB << endl; // bi
		cout << "(DEBUG - computeWeirCockerhamFit) totalWeirCockerhamTermC: " << totalWeirCockerhamTermC << endl; // ci
	}

	// Compute and store Fit value
	if (nbBiallelicSite > 0 && (totalWeirCockerhamTermB + totalWeirCockerhamTermC) != 0.) {
		currentWarehouse->set_WC84_total_B(totalWeirCockerhamTermB); // bi
		currentWarehouse->set_WC84_total_C(totalWeirCockerhamTermC); // ci

		currentWarehouse->set_WC84_Fit(totalWeirCockerhamTermB / (totalWeirCockerhamTermB + totalWeirCockerhamTermC)); // bi / (bi + ci)
		if (debugMode) cout << "(DEBUG - computeWeirCockerhamFit) WC84_Fit: " << currentWarehouse->get_WC84_Fit() << endl;
	}
}


/*-------------------------------------------------------*/
/*          Population analysis related methods          */
/*-------------------------------------------------------*/

void populationDiversificationAnalysis (PolymorphismSequenceContainer* selectedContainer, Population* populationData, const StandardCodonAlphabet* codonAlphabet,
		const NucleicAlphabet* nucleoAlphabet, 	ResultsWarehouse* currentWarehouse) {

	// Create a PopulationDiversification and fill the Fst matrix
	PopulationDiversification FstData(selectedContainer, populationData, codonAlphabet, nucleoAlphabet);
	FstData.determine_Fst();

	// Update the warehouse
	vector<string> populationNames =  populationData->get_population_names();

	for (unsigned int popIndex1 = 0; popIndex1 < populationData->get_number_of_population(); popIndex1++) {
		for (unsigned int popIndex2 = popIndex1 + 1; popIndex2 < populationData->get_number_of_population(); popIndex2++) {
			double computedFst = FstData.get_specific_Fst(popIndex1, popIndex2);

			currentWarehouse->set_specific_Fst(popIndex1, popIndex2, computedFst);
			currentWarehouse->set_specific_Fst(popIndex2, popIndex1, computedFst);
		}
	}
}


/*------------------------------------------------------*/
/*       Non oriented GC analysis related methods       */
/*------------------------------------------------------*/

void nonOrientedGCAnalysis (PolymorphismSequenceContainer* currentIngroupContainer,	vector<double> alleleFrequencyClass, ResultsWarehouse* currentWarehouse,
		const StandardCodonAlphabet* codonAlphabet, const NucleicAlphabet* nucleoAlphabet, const StandardGeneticCode* stdGeneticCode) {

	// Initializations
	unsigned int nbBiAllelic(0), nbTriAllelic(0), nbQuadriAllelic(0), nbPosition(0), nbSite(0);

	for (unsigned int codonSiteNumber = 0; codonSiteNumber < currentIngroupContainer->getNumberOfSites(); codonSiteNumber++) {
			nbSite++;
			for (unsigned int positionNumber = 0; positionNumber < 3; positionNumber++) { // We analyse each nucleotide of the codon
				nbPosition++;

				AlleleFrequency* SNP_data = new AlleleFrequency(currentIngroupContainer, codonSiteNumber, positionNumber, codonAlphabet, nucleoAlphabet);
				unsigned int nbOfAllele = SNP_data->get_number_of_allele();

				// Treat bi-allelic SNP or increase other counters
				if(SNP_data->is_bi_allelic()) {
					nbBiAllelic++;

					// Get current codon site and replace unresolved characters (gap, N) to avoid errors in CodonSiteTools::numberOfSubsitutions method
					Site currentCodonSite = currentIngroupContainer->getSite(codonSiteNumber);
					replaceUnresolvedCharInCodonSite(currentCodonSite, codonAlphabet, currentIngroupContainer->getNumberOfSequences(), SNP_data->get_allele_2_rank());

					if (CodonSiteTools::numberOfSubsitutions(currentCodonSite) != 1) continue;

					bool synonymousSite = isSynonymousSite(currentCodonSite, currentIngroupContainer->getNumberOfSequences(), codonAlphabet, stdGeneticCode, SNP_data->get_allele_2_rank());

					// Collect data for synonymous site
					if (synonymousSite) {
						SNP_data->determine_GC_frequencies();

						// Update spectrum
						for (unsigned int k = 0; k < alleleFrequencyClass.size(); k++) {
							if (SNP_data->get_GC_relative_frequency() <= alleleFrequencyClass[k]) {
								currentWarehouse->increase_class(k, 1);
								break;
							}
						}
					}

				} else {
					if (nbOfAllele == 3) {
						nbTriAllelic++;
					} else if (nbOfAllele == 4) {
						nbQuadriAllelic++;
					}
				}

				delete(SNP_data);
			}
	}

	// Update result warehouse
	currentWarehouse->set_nb_bi_allelic_SNP(nbBiAllelic);
	currentWarehouse->set_nb_tri_allelic_SNP(nbTriAllelic);
	currentWarehouse->set_nb_quadri_allelic_SNP(nbQuadriAllelic);
}


/*---------------------------------------------------------*/
/*           Divergence Analysis related methods           */
/*---------------------------------------------------------*/

void collectSelectionIndicators (const Sequence* ingroupConsensusSequence, const Sequence* outgroupConsensusSequence, const StandardCodonAlphabet* codonAlphabet,
		StandardGeneticCode* stdGeneticCode, ResultsWarehouse* currentWarehouse, double kappa) {

	SelectionIndicators allSelectionIndicators (ingroupConsensusSequence, outgroupConsensusSequence, codonAlphabet);

	// Determine nN and nS and store them in the warehouse
	allSelectionIndicators.determine_nN(stdGeneticCode, kappa);
	allSelectionIndicators.determine_nS(stdGeneticCode, kappa);
	allSelectionIndicators.countSequencesAndSites();

	currentWarehouse->set_nN(allSelectionIndicators.get_nN());
	currentWarehouse->set_nS(allSelectionIndicators.get_nS());
	currentWarehouse->set_nb_sequences_in_consensus(allSelectionIndicators.get_number_of_sequences());
	currentWarehouse->set_nb_sites_in_consensus(allSelectionIndicators.get_number_of_complete_sites());
}

/******************************************************************************/

void computeGC3 (PolymorphismSequenceContainer* currentConsensusContainer, SpeciesDataWarehouse* currentWarehouse, const StandardCodonAlphabet* codonAlphabet, bool debugMode) {

	// Initializations
	unsigned int countAT(0), countGC(0), countOther(0);
	double computedGC3(0.);

	// Count bases
	for (unsigned int k = 0; k < currentConsensusContainer->getNumberOfSites(); k++) {
		int thirdPosition = codonAlphabet->getNPosition(currentConsensusContainer->valueAt(0,k),2);

		if (thirdPosition == 0 || thirdPosition == 3) {
			countAT++;
		} else if (thirdPosition == 1 || thirdPosition == 2) {
			countGC++;
		} else {
			countOther++;
		}
	}

	if (debugMode) cout << "(DEBUG - computeGC3) Counters: GC = " << countGC << " - AT = " << countAT << " - Other = " << countOther << endl;

	// Calculate GC3
	computedGC3 = countGC / (double)(countGC + countAT);

	// Store the result in the SDW
	currentWarehouse->set_GC3(computedGC3);
}

/******************************************************************************/

void mcDonaldKreitmanAnalysis (PolymorphismSequenceContainer* currentIngroupContainer, PolymorphismSequenceContainer* outgroupConsensusContainer, ResultsWarehouse* currentWarehouse,
		unsigned int kappa, const StandardCodonAlphabet* codonAlphabet, StandardGeneticCode* stdGeneticCode, bool debugMode) {

	// Create a new PolymorphsimSequenceContainer with all ingroup sequences and the outgroup consensus
	PolymorphismSequenceContainer customContainer(*currentIngroupContainer);

	for (unsigned int i = 0; i < outgroupConsensusContainer->getNumberOfSequences(); i++) {
		customContainer.addSequence(outgroupConsensusContainer->getSequence(i));
		customContainer.setAsOutgroupMember(outgroupConsensusContainer->getName(i));
	}

	// Compute MK_nN and MK_nS from ingroup data
	int MK_nN = SequenceStatistics::meanNonSynonymousSitesNumber(customContainer, *stdGeneticCode, kappa);
	int MK_nS = SequenceStatistics::meanSynonymousSitesNumber(customContainer, *stdGeneticCode, kappa);

	// Build MKtable from both ingroup and outgroup data
	vector<unsigned int> MKTable0 = SequenceStatistics::MKtable(*currentIngroupContainer, *outgroupConsensusContainer, *stdGeneticCode, 0.);
	vector<unsigned int> MKTable02 = SequenceStatistics::MKtable(*currentIngroupContainer, *outgroupConsensusContainer, *stdGeneticCode, 0.2);

	if (debugMode) {
		cout << "(DEBUG - mcDonaldKreitmanAnalysis) MK table (all SNPs): " << "pn=" << MKTable0[0] << " ps=" << MKTable0[1] << " dn=" << MKTable0[2] << " ds=" << MKTable0[3] << endl; // tmp
		cout << "(DEBUG - mcDonaldKreitmanAnalysis) MK table (freq<0.2): " << "pn=" << MKTable02[0] << " ps=" << MKTable02[1] << " dn=" << MKTable02[2] << " ds=" << MKTable02[3] << endl; // tmp
	}

	// Save computed data into the warehouse
	currentWarehouse->set_fixN(MKTable0[2]); // Da
	currentWarehouse->set_fixS(MKTable0[3]); // Ds
	currentWarehouse->set_polymN_0(MKTable0[0]); // Pa
	currentWarehouse->set_polymS_0(MKTable0[1]); // Ps
	currentWarehouse->set_polymN_02(MKTable02[0]); // Pa 0.2
	currentWarehouse->set_polymS_02(MKTable02[1]); // Ps 0.2

	currentWarehouse->set_MK_nN(MK_nN);
	currentWarehouse->set_MK_nS(MK_nS);
}


/*------------------------------------------------------------------*/
/*     Distance between individuals computation related methods     */
/*------------------------------------------------------------------*/

void computeAllDistancesBetweenIndividuals (ResultsWarehouse* currentWarehouse, PairedSequenceContainer* allPairs, const StandardCodonAlphabet* codonAlphabet,
		const NucleicAlphabet* customNucleoAlphabet, bool debugMode) {

	// Collect Pair names
	vector<string> allNames = allPairs->get_pair_names();

	for (unsigned int i = 0; i < allPairs->get_nb_pair(); i++) {
		for (unsigned int j = 0; j < allPairs->get_nb_pair(); j++) {
			DistanceBetweenIndividuals* newDistanceObject = new DistanceBetweenIndividuals(allNames[i], allNames[j]);
			newDistanceObject->computeDistanceWithPair(allPairs, i, j, codonAlphabet, customNucleoAlphabet, debugMode);

			// Save computed values in the warehouse
			currentWarehouse->set_specific_distance_object(newDistanceObject->get_pair_name(), newDistanceObject);
		}
	}
}

vector<double> summarizeDistances (vector<ResultsWarehouse*>& allDatasetResults, string coupleName) {

	// Initializations
	vector<double> bothDistances(2, -1.);
	double sumDistanceBySiteW(0.), sumNbResolvedSitesW(0.), sumDistanceBySiteUw(0.), sumNbResolvedSitesUw(0.);

	// Distance data recovery and summing
	for (unsigned int i = 0; i < allDatasetResults.size(); i++) {
		// Recover distance data from the warehouse
		DistanceBetweenIndividuals* distanceObject = allDatasetResults[i]->get_specific_distance_object(coupleName);

		double currentDistance = distanceObject->get_distance();
		int currentNbResolvedSite = distanceObject->get_nb_resolved_positions();

		if (currentDistance >= 0) {
			// Weighted
			sumDistanceBySiteW += currentDistance * currentNbResolvedSite;
			sumNbResolvedSitesW += currentNbResolvedSite;

			// Unweighted
			sumDistanceBySiteUw += currentDistance;
			sumNbResolvedSitesUw++;
		}
	}

	// Calculate weighted distances..
	if (sumNbResolvedSitesW != 0) bothDistances[0] = sumDistanceBySiteW / sumNbResolvedSitesW;

	// .. and unweighted distances for current dataset
	if (sumNbResolvedSitesUw != 0) bothDistances[1] = sumDistanceBySiteUw / sumNbResolvedSitesUw;

	return bothDistances;
}


/*---------------------------------------------------------*/
/*      Biological summary generation related methods      */
/*---------------------------------------------------------*/

Summary* biologicalSummary (vector<ResultsWarehouse*>& allResults, ParametersManagement* allParams, Population* popData) {

	// Initializations
	vector<ResultsWarehouse*> requestedResults(0);

	// Get the results associated with requested contigs
	if (allParams->get_contig_usage() == "even") {
		// Note: if the user request all EVEN contigs (of the Population Genetics file) then we have to keep all ODD vector elements (because vectors starts at 0) !
		for (unsigned int i = 1; i < allResults.size(); i = i + 2) requestedResults.push_back(allResults[i]); // We keep element with index 1,3,5,7..

	} else if (allParams->get_contig_usage() == "odd") {
		// Note: if the user request all ODD contigs (of the Population Genetics file) then we have to keep all EVEN vector elements (because vectors starts at 0) !
		for (unsigned int i = 0; i < allResults.size(); i = i + 2) requestedResults.push_back(allResults[i]); // We keep element with index 0,2,4,6..

	} else {
		requestedResults.insert(requestedResults.end(), allResults.begin(), allResults.end());
	}

	Summary* bioSummary = new Summary(requestedResults, popData, allParams->get_ingroup_species_name(), requestedResults.size(), allParams->get_bootstrap_replicates(), allParams->get_bootstrap_sort_precision(), allParams->get_bootstrap_confidence());

	bioSummary->compute_all_averages(popData);
	bioSummary->bootstrap_everything();

	return bioSummary;
}


/*--------------------------------------------------*/
/*       Output files writing related methods       */
/*--------------------------------------------------*/

string buildHeaderLine(Population* popData, vector<double>& alleleFrequencyClass, string ingroupSpeciesName, bool debugMode) {

	string fullHeaderLine("");

	// Fixed global columns
	fullHeaderLine.append("Contig_name\tStatus");
	fullHeaderLine.append("\tBiAllelic_SNP\tTriAllelic_SNP\tQuadriAllelic_SNP\tnN\tnS");
	fullHeaderLine.append("\tfixN\tfixS\tpolymN0.0\tpolymS0.0\tpolymN0.2\tpolymS0.2\tMK_nN\tMK_nS\t");


	// GC allele frequency classes (dynamic)
	for (unsigned int i = 0; i < alleleFrequencyClass.size(); i++) {
		fullHeaderLine.append("GCalfreq<=" + TextTools::toString(alleleFrequencyClass[i]) + "\t");
	}

	// Focal species data (fixed)
	fullHeaderLine.append(ingroupSpeciesName + "_nb_sequence" + "\t" + ingroupSpeciesName + "_nb_complete_site" + "\t" + ingroupSpeciesName + "_GC3" + "\t");
	fullHeaderLine.append(ingroupSpeciesName + "_piN" + "\t" + ingroupSpeciesName + "_piS" + "\t" + ingroupSpeciesName + "_Fit" + "\t");
	fullHeaderLine.append(ingroupSpeciesName + "_WeirCockerham84_Fit" + "\t");

	// Population data (dynamic)
	vector< vector<string> > populations = popData->get_population_list();
	unsigned int nbPopulations = popData->get_number_of_population();
	vector<string> populationNames =  popData->get_population_names();

	for (unsigned int ii = 0; ii < nbPopulations; ii++) {
		for(unsigned int jj = 0; jj < populations[ii].size(); jj++) {
			fullHeaderLine.append("H(" + populations[ii][jj] + ")" + "\t");
		}
	}

	for (unsigned int popIndex1 = 0; popIndex1 < nbPopulations; popIndex1++) {
		for (unsigned int popIndex2 = popIndex1 + 1; popIndex2 < nbPopulations; popIndex2++) {
			fullHeaderLine.append("Fst(" + populationNames[popIndex1] + "," + populationNames[popIndex2] + ")" + "\t");
		}
	}

	if (debugMode) cout << "(buildHeaderLine) Builded Header line:" << endl << endl << fullHeaderLine << endl << endl;

	return fullHeaderLine;
}

/******************************************************************************/

void writeUnitaryResultLine(ofstream& outputStream, ResultsWarehouse* unitaryResults, Population* popData, string ingroupSpeciesName) {

	/* Fixed global columns */
	outputStream << unitaryResults->get_name() << "\t";
	outputStream << unitaryResults->get_status() << "\t";

	outputStream << unitaryResults->get_nb_bi_allelic_SNP() << "\t";
	outputStream << unitaryResults->get_nb_tri_allelic_SNP() << "\t";
	outputStream << unitaryResults->get_nb_quadri_allelic_SNP() << "\t";

	outputStream << unitaryResults->get_nN() << "\t";
	outputStream << unitaryResults->get_nS() << "\t";

	outputStream << unitaryResults->get_fixN() << "\t";
	outputStream << unitaryResults->get_fixS() << "\t";

	outputStream << unitaryResults->get_polymN_0() << "\t";
	outputStream << unitaryResults->get_polymS_0() << "\t";
	outputStream << unitaryResults->get_polymN_02() << "\t";
	outputStream << unitaryResults->get_polymS_02() << "\t";

	outputStream << unitaryResults->get_MK_nN() << "\t";
	outputStream << unitaryResults->get_MK_nS() << "\t";

//	outputStream << "separator" << "\t";

	// GC allele frequency classes (dynamic)
	vector<double> gcAlleleFrequencies = unitaryResults->get_GC_allele_frequency_classes();

	for (unsigned int i = 0; i < gcAlleleFrequencies.size(); i++) {
		outputStream << gcAlleleFrequencies[i] << "\t";
	}

	// Focal species data (fixed)
	SpeciesDataWarehouse* ingroupSpeciesData = unitaryResults->get_specific_species_data(ingroupSpeciesName);

	outputStream << ingroupSpeciesData->get_number_of_sequences() << "\t";
	outputStream << ingroupSpeciesData->get_number_of_complete_sites() << "\t";
	outputStream << ingroupSpeciesData->get_GC3() << "\t";
	outputStream << ingroupSpeciesData->get_piN() << "\t";
	outputStream << ingroupSpeciesData->get_piS() << "\t";
	outputStream << ingroupSpeciesData->get_Fit() << "\t";
	outputStream << ingroupSpeciesData->get_WC84_Fit() << "\t";

	// Population data (dynamic)
	vector< vector<string> > populations = popData->get_population_list();
	unsigned int nbPopulations = popData->get_number_of_population();

	for (unsigned int ii = 0; ii < nbPopulations; ii++) {
		for(unsigned int jj = 0; jj < populations[ii].size(); jj++) {
			outputStream << unitaryResults->get_specific_individual_heterozygosity(populations[ii][jj]) << "\t";
		}
	}

	for (unsigned int popIndex1 = 0; popIndex1 < nbPopulations; popIndex1++) {
		for (unsigned int popIndex2 = popIndex1 + 1; popIndex2 < nbPopulations; popIndex2++) {
			outputStream << unitaryResults->get_specific_Fst(popIndex1, popIndex2) << "\t";
		}
	}

	outputStream << endl;
}

/******************************************************************************/

void writeMainOutputFile (ParametersManagement* allParameters, vector<ResultsWarehouse*>& allResults, Population* popData, vector<double>& alleleFrequencyClass) {

	// Initializations
	string outputFilename = allParameters->get_output_file_name();

	cout << "All results are now written into the following file: " << outputFilename << endl << endl;

	// Create/Open an output stream
	ofstream outputStream (outputFilename.c_str(), ios::out | ios::trunc);

	// Fill the file
	if (outputStream.is_open()) {

		// Deal with the header line
		if (allParameters->get_print_header()) {
			outputStream << buildHeaderLine(popData, alleleFrequencyClass, allParameters->get_ingroup_species_name(), allParameters->get_DEBUG()) << endl;
		}

		// Write one result line per dataset/contig
		for (unsigned int i = 0; i < allResults.size(); i++) {
			writeUnitaryResultLine(outputStream, allResults[i], popData, allParameters->get_ingroup_species_name());
		}

		outputStream.close();

	} else {
		cerr << "Error: Cannot create/open file: " << outputFilename << endl;
		exit(EXIT_FAILURE);
	}
}

/******************************************************************************/

void writeDistanceFile (ParametersManagement* allParameters, vector<ResultsWarehouse*>& allResults, Population* popData) {

	// Initializations
	string outputFilename("");

	// Generate valid filename depending on the execution mode
	if (allParameters->get_execution_mode() == "local") {
		outputFilename = FileTools::getFileName(allParameters->get_output_file_name()) + ".dist";

		// Hotfix for getFileName and filenames that don't have an extension
		if (outputFilename == ".dist") {
			StringTokenizer token(allParameters->get_output_file_name(), "/");
			outputFilename = token.getToken(token.numberOfRemainingTokens() - 1) + ".dist";
		}
	} else {
		outputFilename= "dNdSpiNpiS_distances";
	}

	cout << "Distances between each focal individuals are now written into the following file : " << outputFilename << endl << endl;

	// Create/Open an output stream
	ofstream outputStream (outputFilename.c_str(), ios::out | ios::trunc);

	// Fill the file
	if (outputStream.is_open()) {

		if (allParameters->get_print_header()) {
			outputStream << "Individual_1\tIndividual_2\tWeighted_distance\tUnweighted_distance" << endl;
		}

		vector<string> allIndividuals = popData->get_all_individuals();

		// Compute weighted distance for each couples
		for (unsigned int i = 0; i < allIndividuals.size(); i++) {
			for (unsigned int j = 0; j < allIndividuals.size(); j++) {
				vector<double> distances = summarizeDistances(allResults, allIndividuals[i] + "|" + allIndividuals[j]);

				outputStream << allIndividuals[i] << "\t" << allIndividuals[j] << "\t" << distances[0] << "\t" << distances[1] << endl;
			}
		}

		outputStream.close();

	} else {
		cerr << "Error: Cannot create/open file: " << outputFilename << endl;
		exit(EXIT_FAILURE);
	}
}

/******************************************************************************/

void writeSummaryFile (ParametersManagement* allParameters, Summary* bioSum, Population* popData) {

	// Initializations
	string outputFilename("");

	// Generate valid filename depending on the execution mode
	if (allParameters->get_execution_mode() == "local") {
		outputFilename = FileTools::getFileName(allParameters->get_output_file_name()) + ".sum";

		// Hotfix for getFileName and filenames that don't have an extension
		if (outputFilename == ".sum") {
			StringTokenizer token(allParameters->get_output_file_name(), "/");
			outputFilename = token.getToken(token.numberOfRemainingTokens() - 1) + ".sum";
		}
	} else {
		outputFilename= "dNdSpiNpiS_summary";
	}

	cout << "The Biological Summary is now written into the following file : " << outputFilename << endl << endl;

	// Create/Open an output stream
	ofstream outputStream (outputFilename.c_str(), ios::out | ios::trunc);

	// Fill the file
	if (outputStream.is_open()) {

		outputStream << "################################################################################" << endl;
		outputStream << "#                              Biological Summary                              #" << endl;
		outputStream << "################################################################################" << endl << endl;

		outputStream << "Selected ingroup species: " << allParameters->get_ingroup_species_name() << endl << endl;

		outputStream << "Number of analyzed individual: " << popData->get_number_of_individuals() << " (from " << popData->get_number_of_population() << " population(s))" << endl;
		outputStream << endl;

		outputStream << "Total number of contig used for sequence analysis: " << bioSum->get_nb_contig() << endl;
		outputStream << endl;

		outputStream << "Total number of SNPs: " << bioSum->get_total_number_of_SNP() << endl << endl;
		outputStream << "  - Biallelic: " << bioSum->get_total_SNP2() << endl;
		outputStream << "  - Triallelic: " << bioSum->get_total_SNP3() << endl;
		outputStream << "  - Quadriallelic: " << bioSum->get_total_SNP4() << endl;
		outputStream << endl;

		vector<double> fit_standard_error = bioSum->get_fit_standard_error();
		outputStream << "Fit:" << endl << endl;
		outputStream << "Average Fit: " << bioSum->get_average_fit() << " [" << fit_standard_error[0] << "; " << fit_standard_error[1] << "]" << endl;
		outputStream << "(Fit calculated in " << bioSum->get_nb_contig() - bioSum->get_failed_fit() << " contigs)" << endl;
		outputStream << endl;

		vector<double> WC84_fit_standard_error = bioSum->get_WC84_fit_standard_error();
		outputStream << "Weir & Cockerham Fit (Evolution 1984):" << endl << endl;
		outputStream << "Average Weir & Cockerham Fit: " << bioSum->get_average_WC84_fit() << " [" << WC84_fit_standard_error[0] << "; " << WC84_fit_standard_error[1] << "]" << endl;
		outputStream << "(Fit calculated in " << bioSum->get_nb_contig() - bioSum->get_failed_WC84_fit() << " contigs)" << endl;
		outputStream << endl;

		vector<double> piS_standard_error = bioSum->get_piS_standard_error();
		vector<double> piN_standard_error = bioSum->get_piN_standard_error();
		vector<double> piNpiS_standard_error = bioSum->get_piNpiS_standard_error();
		outputStream << "piN/piS ratio:" << endl << endl;
		outputStream << "Average piS in focal species: " << bioSum->get_average_piS() << " [" << piS_standard_error[0] << "; " << piS_standard_error[1] << "]"  << endl;
		outputStream << "Average piN in focal species: " << bioSum->get_average_piN() << " [" << piN_standard_error[0] << "; " << piN_standard_error[1] << "]"  << endl;
		outputStream << "Average piN / average piS: " << bioSum->get_average_piNpiS() << " [" << piNpiS_standard_error[0] << "; " << piNpiS_standard_error[1] << "]" << endl;
		outputStream << "(piS and piN calculated in " << bioSum->get_nb_contig() - bioSum->get_failed_piNpiS() << " contigs of average length " << bioSum->get_average_contig_length_piNpiS() << ")" << endl;
		outputStream << endl;

		if (allParameters->get_outgroup_species_name() != "none") {
			vector<double> dS_standard_error = bioSum->get_dS_standard_error();
			vector<double> dN_standard_error = bioSum->get_dN_standard_error();
			vector<double> dNdS_standard_error = bioSum->get_dNdS_standard_error();
			outputStream << "dN/dS ratio:" << endl << endl;
			outputStream << "Average dS: " << bioSum->get_average_dS() << " [" << dS_standard_error[0] << "; " << dS_standard_error[1] << "]" << endl;
			outputStream << "Average dN: " << bioSum->get_average_dN() << " [" << dN_standard_error[0] << "; " << dN_standard_error[1] << "]" << endl;
			outputStream << "Average dN/dS: " << bioSum->get_average_dNdS() << " [" << dNdS_standard_error[0] << "; " << dNdS_standard_error[1] << "]" << endl;
			outputStream << "(" << bioSum->get_contig_above_threshold() << " contigs with dS > " << bioSum->get_dS_threshold() << " has been discarded)" << endl;
			outputStream << endl;

			vector<double> alpha_standard_error = bioSum->get_alpha_standard_error();
			vector<double> alpha02_standard_error = bioSum->get_alpha02_standard_error();
			outputStream << "Alpha:" << endl << endl;
			outputStream << "NI: " << bioSum->get_NI() << " => alpha: " << bioSum->get_alpha() << " [" << alpha_standard_error[0] << ";" << alpha_standard_error[1] << "]" << " (all SNPs)" << endl;
			outputStream << "NI0.2: " << bioSum->get_NI02() << " => alpha0.2: " << bioSum->get_alpha02() << " [" << alpha02_standard_error[0] << ";" << alpha02_standard_error[1] << "]" << " (allele freq > 0.2)" << endl;
			outputStream << "(MK-analysis performed in " << bioSum->get_nb_contig() - bioSum->get_failed_NI() << " contigs)" << endl;
			outputStream << endl;
		}

		outputStream << "Individual heterozygosity:" << endl << endl;
		vector<string> allIndividuals = popData->get_all_individuals();
		//sort (allIndividuals.begin(), allIndividuals.end());

		for (unsigned int i = 0; i < allIndividuals.size(); i++) {
			outputStream << "H(" << allIndividuals[i] << "): " << bioSum->get_specific_average_individual_heterozygosity(allIndividuals[i]) << endl;
		}
		outputStream << endl;

		vector<string> allCoupleNames =  popData->get_all_couples();
		if (allCoupleNames.size() > 0) {
			outputStream << "Population differentiation:" << endl << endl;

			if (allCoupleNames.size() > 0) {
				for (unsigned int i = 0; i < allCoupleNames.size(); i++) {
					outputStream << "Fst(" << allCoupleNames[i] << "): " << bioSum->get_specific_average_population_fst(allCoupleNames[i]) << endl;
				}
			}
		}
		outputStream << endl;

		outputStream.close();

	} else {
		cerr << "Error: Cannot create/open file: " << outputFilename << endl;
		exit(EXIT_FAILURE);
	}
}
