/*
 * dNdSpiNpiS.h
 *
 *  Created on: 15 mai 2013
 *      Author: Aurélien Bernard
 */

#ifndef DNDSPINPIS_H_
#define DNDSPINPIS_H_

/*--------------------------------------*/
/*               Includes               */
/*--------------------------------------*/

// Basic C++ classes
#include <iostream>
#include <string>
#include <vector>

// Bio++ - Alphabets and genetic codes
#include <Bpp/Seq/Alphabet/StandardCodonAlphabet.h>
#include <Bpp/Seq/GeneticCode/StandardGeneticCode.h>

// Bio++ - Site manipulation
#include <Bpp/Seq/CodonSiteTools.h>

// Bio++ - Population genetics containers and tools
#include <Bpp/PopGen/PolymorphismSequenceContainer.h>
#include <Bpp/PopGen/SequenceStatistics.h>

// PopPhyl Project Developments
#include <ppc/PopGen/Container/PairedSequenceContainer.h>
#include <ppc/PopGen/Population.h>
#include <ppc/PopGen/AlleleFrequency.h>

// Custom classes
#include "ParametersManagement.h"
#include "SpeciesDataWarehouse.h"
#include "ResultsWarehouse.h"
#include "Summary.h"

/*--------------------------------------*/
/*              Prototypes              */
/*--------------------------------------*/

/*  Unsorted basic methods  */
std::vector<double> setAlleleFrequencyClass (ParametersManagement* allParams);
double determineWeirCockerhamTermB (double sampleSize, double observedFrequency, double observedHeterozygosity, bool debugMode);
void replaceUnresolvedCharInCodonSite (bpp::Site& SiteToModify, const bpp::StandardCodonAlphabet* codonAlphabet, unsigned int nbSequence, unsigned int rankOfAllele2);
bool isSynonymousSite (bpp::Site& siteToAnalyze, unsigned int nbSequence, const bpp::StandardCodonAlphabet* codonAlphabet, const bpp::StandardGeneticCode* stdGeneticCode, int allele_2_rank);


/*  Display related methods  */
void displayWelcomeMessage (ParametersManagement* dNdS_params);
void displayGoodbyeMessage();


/*  Input file loading/cleaning related methods  */
ppc::Population* loadPopulationFile (std::string PopulationFile);
ppc::Population* generateBasicPopulation (ParametersManagement* allParams, std::vector<bpp::PolymorphismSequenceContainer*>& allDatasets);
std::vector<bpp::PolymorphismSequenceContainer*> loadAlignmentFile (ParametersManagement* allParams, const bpp::StandardCodonAlphabet* codonAlphabet, const bpp::NucleicAlphabet* customNucleoAlphabet);
std::vector<bpp::PolymorphismSequenceContainer*> cleanAlignmentData (ParametersManagement* allParams, std::vector<bpp::PolymorphismSequenceContainer*>& multidata, const bpp::StandardCodonAlphabet* codonAlphabet);


/*  Paired Sequence analysis related methods  */
void computeIndividualHeterozygosity (ppc::PairedSequenceContainer* currentPairedContainer, const bpp::StandardCodonAlphabet* codonAlphabet, const bpp::NucleicAlphabet* nucleoAlphabet, ResultsWarehouse* currentWarehouse, std::vector<std::string>& allPairNames);
void computeFit (ppc::PairedSequenceContainer* currentPairedContainer, const bpp::StandardCodonAlphabet* codonAlphabet, const bpp::NucleicAlphabet* nucleoAlphabet, SpeciesDataWarehouse* currentWarehouse, std::vector<std::string>& allPairNames, unsigned int globalNumberOfSite, bool debugMode = false);
void computeWeirCockerhamFit (ppc::PairedSequenceContainer* currentPairedContainer, const bpp::StandardCodonAlphabet* codonAlphabet, const bpp::NucleicAlphabet* nucleoAlphabet, SpeciesDataWarehouse* currentWarehouse, std::vector<std::string>& allPairNames, unsigned int globalNumberOfSite, bool debugMode);


/*  Population analysis related methods  */
void populationDiversificationAnalysis (bpp::PolymorphismSequenceContainer* selectedContainer, ppc::Population* PopulationData, const bpp::StandardCodonAlphabet* codonAlphabet, const bpp::NucleicAlphabet* nucleoAlphabet, 	ResultsWarehouse* currentWarehouse);


/*  Non oriented GC analysis related methods  */
void nonOrientedGCAnalysis (bpp::PolymorphismSequenceContainer* currentIngroupContainer,	std::vector<double> alleleFrequencyClass, ResultsWarehouse* currentWarehouse, const bpp::StandardCodonAlphabet* codonAlphabet, const bpp::NucleicAlphabet* nucleoAlphabet, const bpp::StandardGeneticCode* stdGeneticCode);


/*  Divergence Analysis related methods  */
void collectSelectionIndicators (const bpp::Sequence* ingroupConsensusSequence, const bpp::Sequence* outgroupConsensusSequence, const bpp::StandardCodonAlphabet* codonAlphabet, bpp::StandardGeneticCode* stdGeneticCode, ResultsWarehouse* currentWarehouse, double kappa);
void computeGC3 (bpp::PolymorphismSequenceContainer* currentConsensusContainer, SpeciesDataWarehouse* currentWarehouse, const bpp::StandardCodonAlphabet* codonAlphabet, bool debugMode = false);
void mcDonaldKreitmanAnalysis (bpp::PolymorphismSequenceContainer* currentIngroupContainer, bpp::PolymorphismSequenceContainer* outgroupConsensusContainer, ResultsWarehouse* currentWarehouse, unsigned int kappa, const bpp::StandardCodonAlphabet* codonAlphabet, bpp::StandardGeneticCode* stdGeneticCode, bool debugMode = false);


/*  Distance between individuals computation related methods  */
void computeAllDistancesBetweenIndividuals (ResultsWarehouse* currentWarehouse, ppc::PairedSequenceContainer* allPairs, const bpp::StandardCodonAlphabet* codonAlphabet, const bpp::NucleicAlphabet* customNucleoAlphabet, bool debugMode = false);
std::vector<double> summarizeDistances (std::vector<ResultsWarehouse*>& allDatasetResults, std::string coupleName);


/*  Biological summary generation related methods  */
Summary* biologicalSummary (std::vector<ResultsWarehouse*>& allResults, ParametersManagement* allParams, ppc::Population* popData);


/*  Output files writing related methods  */
std::string buildHeaderLine(ppc::Population* popData, std::vector<double>& alleleFrequencyClass, std::string ingroupSpeciesName, bool debugMode = false);
void writeUnitaryResultLine(std::ofstream& outputStream, ResultsWarehouse* unitaryResults, ppc::Population* popData, std::string ingroupSpeciesName);
void writeMainOutputFile (ParametersManagement* allParameters, std::vector<ResultsWarehouse*>& allResults, ppc::Population* popData, std::vector<double>& alleleFrequencyClass);
void writeDistanceFile (ParametersManagement* allParameters, std::vector<ResultsWarehouse*>& allResults, ppc::Population* popData);
void writeSummaryFile (ParametersManagement* allParameters, Summary* bioSum, ppc::Population* popData);


/*  Dataset analysis related methods  */
void analyseFocalSpecies(ParametersManagement* dNdSpiNpiS_params, bpp::PolymorphismSequenceContainer* ingroupContainer, const bpp::Sequence* ingroupConsensusSequence, ppc::Population* PopulationData, const bpp::StandardCodonAlphabet* codonAlphabet, const bpp::NucleicAlphabet* customNucleoAlphabet, bpp::StandardGeneticCode* stdGeneticCode, std::vector<double>& alleleFrequencyClass, SpeciesDataWarehouse* ingroupSpeciesData, ResultsWarehouse* datasetResults);
ResultsWarehouse* analyseDataset (ParametersManagement* dNdSpiNpiS_params, bpp::PolymorphismSequenceContainer* currentDataset, ppc::Population* PopulationData, const bpp::StandardCodonAlphabet* codonAlphabet, const bpp::NucleicAlphabet* customNucleoAlphabet, bpp::StandardGeneticCode* stdGeneticCode, std::vector<double>& alleleFrequencyClass);


#endif /* DNDSPINPIS_H_ */
