//============================================================================
// Name        : Alr.hpp
// Author      : Vincent Cahais
// Created on  : 23 avril 2012
// Copyright   : http://www.cecill.info/licences/Licence_CeCILL-C_V1-fr.html
//============================================================================


#ifndef ALR_H_
#define ALR_H_

#include <vector>
#include <iostream>
#include <valarray>
#include <Bpp/Text/TextTools.h>
#include <Bpp/Text/StringTokenizer.h>
#include <map>
#include "Utils.hpp"

using namespace std;

namespace bpp
{

/*********************************************************************************************/
/* Alrheaders*/
/*********************************************************************************************/


	class Alrheaders
	{
	public:

		vector<string> indiv;
		vector<string> species;
		map<string, unsigned int>list;

		Alrheaders(){}
		~Alrheaders(){}

		void readline(string linebuffer)
		{
			StringTokenizer st (linebuffer, "\t|", false, false);
			st.nextToken(); st.nextToken();
			for(unsigned int i=0; st.hasMoreToken(); i++)
			{
				species.push_back(st.nextToken());
				indiv.push_back(st.nextToken());
				list[species[i]]++;
			}
		}

		void writeline(ostream & out)
		{
			out << "Maj\tM/P";
			for (unsigned int i=0; i<indiv.size(); i++)
				out << "\t" << species[i] << "|" << indiv[i];
			out << endl;
		}

		string getindiv(unsigned int i){return indiv[i];}
		unsigned int numberofindiv(){return indiv.size();}
		unsigned int numberofindiv(string spe){return list[spe];}
		unsigned int numberofspecies(){return list.size();}
		string getspecies(unsigned int i){return species[i];}
		string getspeciesname(unsigned int i)
		{
			map<string, unsigned int>::iterator it=list.begin();
			for (unsigned int x=0; x<i; x++)it++;
			return it->first;
		}

	};


/*********************************************************************************************/
/* Alrline*/
/*********************************************************************************************/

	class Alrline
	{
	public:
		string nuc;
		bool ispolymorph;
		valarray< unsigned int> values;

		Alrline():ispolymorph(false){}
		~Alrline(){}

		void readline(string & linebuffer)
		{
			unsigned int size=10;
			StringTokenizer st1 (linebuffer, "\t[/]", false, false);

			nuc = st1.nextToken();
			if (st1.nextToken() == "P")	{ ispolymorph = true; size = 0.8*st1.numberOfRemainingTokens();}
			else size = 4*st1.numberOfRemainingTokens();

			values.resize(size,0);
			unsigned int n=utils::toint(nuc);
			unsigned int j=0;
			for(unsigned int i=0; st1.hasMoreToken(); i++)
			{
				if (ispolymorph)
				{
					if ( (i%5==0) ) st1.nextToken();
					else
					{
						values[j] = atoi(st1.nextToken().c_str());
						j++;
					}
				}
				else
				{
					values[(i*4)+n] = atoi(st1.nextToken().c_str());
				}
			}
		}

		void writeline(ostream & out)
		{
			out << nuc << "\t";
			if (!ispolymorph)
			{
				unsigned int n=utils::toint(nuc);
				out << "M";
				for (unsigned int i=n; i<values.size(); i+=4)
					out << "\t" << values[i];
			}
			else
			{
				out << "P";
				for (unsigned int i=0; i<values.size(); i+=4)
				{
					out << "\t" << values[i]+values[i+1]+values[i+2]+values[i+3] << "[";
					out << values[i] << "/" << values[i+1] << "/" << values[i+2] << "/" << values[i+3];
					out << "]";
				}
			}
			out << endl;
		}

		unsigned int size(){return values.size();}
		unsigned int getValue(unsigned int i){return values[i];}
		unsigned int numberOfreads(unsigned int i){return values[(i*4)]+values[(i*4)+1]+values[(i*4)+2]+values[(i*4)+3];}

		void setValue(unsigned int pos, unsigned int v){values[pos]=v;}
	};


/*********************************************************************************************/
/* AlrContig*/
/*********************************************************************************************/

	class AlrContig
	{

	private:
		Alrheaders headers_;
		vector<Alrline*> seq_;

	public:
		string name;

		AlrContig(){}
		virtual ~AlrContig(){}

		Alrline* getline(unsigned int i){return seq_[i];}
		Alrheaders getheaders_(){return headers_;}
		unsigned int size(){return seq_.size();}
		void readheaders(string line){headers_.readline(line);}
		void readline(string line)
		{
			Alrline* newline = new Alrline;
			newline->readline(line);
			seq_.push_back(newline);
		}

		void writeContig(ostream & out)
		{
			out << ">" << name << endl;
			headers_.writeline(out);
			for (unsigned int i=0; i< seq_.size(); i++)
			{
				seq_[i]->writeline(out);
			}
		}

		unsigned int numberofindiv(){return headers_.numberofindiv();}
		unsigned int numberofindiv(string spe){return headers_.list[spe];}
		unsigned int numberofspecies(){return headers_.numberofspecies();}
		string getspecies(unsigned int i){return headers_.getspecies(i);}
		string getspeciesname(unsigned int i){return headers_.getspeciesname(i);}
		string getindiv(unsigned int i){return headers_.getindiv(i);}
		bool ispolymorph(unsigned int i){return seq_[i]->ispolymorph;}

		valarray<unsigned int> getvaluesbyspecies(unsigned int line, string species)
		{
			unsigned int size = numberofindiv(species)*4;
			valarray<unsigned int> values(size);
			unsigned int pos=0;
			for(unsigned int i=0; i<headers_.numberofindiv(); i++)
			{
				//cerr << "i=" << i << "/" << headers_.numberofindiv() << endl;
				//cerr << headers_.species[i] << "=" << species << endl;
				if (headers_.species[i]==species)
				{
					for (unsigned int x=0; x<4; x++)
					{
						//cerr << "pos=" << pos << "/" << numberofindiv(species)*4 << " x=" << x << "/4" << endl;
						values[pos+x]=seq_[line]->values[(i*4)+x];
					}
					pos+=4;
				}
			}
			return values;
		}

	};

/*********************************************************************************************/
/* Alrparser*/
/*********************************************************************************************/

	class Alrparser
	{

	public:

		Alrparser(){}
		virtual ~Alrparser(){}

		/*******************************************************************************/
		/*nextsequence : read one contig in alr file*/
		/*
		 * @return : true if a sequence has been read*/
		/*******************************************************************************/
		bool nextsequence(istream & input, AlrContig & contig) const throw (Exception)
		{
		  if (!input)
			throw IOException("Alr::nextseq_uence: can't read from istream input");
		  if (input.eof()) return false;

		  string linebuffer = "";
		  getline(input, linebuffer); //read first line
		  if (linebuffer[0] != '>')
			throw Exception("Alr::nextseq_uence: wrong format");

		  contig.name = string(linebuffer.begin() + 1, linebuffer.end()); // Get the seq_uence name line
		  getline(input, linebuffer); //read second line
		  contig.readheaders(linebuffer);

		  while (!input.eof())
		  {
			getline(input, linebuffer);
			contig.readline(linebuffer);
			if (input.peek() == '>') return true; // Stop if find a new sequence
		  }
		  return true;
		}

	};


}
#endif /* ALR_H_ */

