//============================================================================
// Name        : LnL.cpp
// Author      : Vincent Cahais, Nicolas Galtier
// Created on  : 17 septembre 2012
// Copyright   : http://www.cecill.info/licences/Licence_CeCILL-C_V1-fr.html
//============================================================================

#include "LnL.h"

#define USE_ERR_ALL_FREQ false

namespace bpp
{


LnL::LnL(Stat * s, AlrContig * c, Model * m, unsigned int n, double f) :
		stat(s), contig(c), model(m), min_reads_per_ind(n), fis_(f),
		ln_cond_posterior_genotypes(MAX_NB_CLASSES,
				vector <vector <vector < vector<double> > > > (MAX_NB_CLASSES,
						vector <vector < vector<double> > > (4,
								vector < vector<double> > (4,
										 vector<double>(c->numberofindiv()) ) ) ) )
{

	for (unsigned int i=0; i<c->numberofspecies(); i++)
		data.species_data.push_back(new Speciesdata(c->getspecies(i)));
	for(unsigned int position=0; position<c->size(); position++)
		data.addPos(new Posdata(c->numberofindiv()));
}

LnL::~LnL(){}


/* lnL */
/* Needs global variables NB_CLASSES_ERR, NB_CLASSES_AEB, ERR_EQUALS_ZERO, AEB_EQUALS_ZERO and DO_BEST */
/* to be set correctly. */
/* Also needs global variable totdata to be allocated and filled */

double LnL::lnL(double er_alpha, double er_beta, double er_tstv, double aeb_alpha, double aeb_beta)
{
	//GammaDiscreteDistribution *prior_er, *prior_aeb;
	double er, lnL;
	double ret;

	/* build prior distributions of error rate */
	if (er_alpha <= 0. || model->ERR_EQUALS_ZERO)
	{
		prior_er_cat[0] = 0.;
		prior_er_prob[0] = 1.;
		model->NB_CLASSES_ERR = 1;
	}
	else if (model->NB_CLASSES_ERR == 1)
	{
		prior_er_cat[0] = er_alpha/(er_alpha + er_beta);
		prior_er_prob[0] = 1.;
	}
	else
	{
		stat->discretize_beta(er_alpha, er_beta, model->NB_CLASSES_ERR, prior_er_cat);
		for (unsigned int i=0; i < model->NB_CLASSES_ERR; i++)
			prior_er_prob[i] = 1./model->NB_CLASSES_ERR;
	}
	/* build prior distributions of allelic expression bias */
	if (aeb_alpha <= 0. || model->AEB_EQUALS_ZERO)
	{
		prior_aeb_cat[0] = 0.;
		prior_aeb_prob[0] = 1.;
		model->NB_CLASSES_AEB = 1;
	}
	else if (model->NB_CLASSES_AEB == 1)
	{
		prior_aeb_cat[0] = aeb_alpha/(aeb_alpha + aeb_beta);
		prior_aeb_prob[0] = 1.;
	}
	else
	{
		stat->discretize_beta(aeb_alpha, aeb_beta, model->NB_CLASSES_AEB, prior_aeb_cat);
		for (unsigned int i=0; i < model->NB_CLASSES_AEB; i++)
			prior_aeb_prob[i] = 1./model->NB_CLASSES_AEB;
	}

	/* build error matrices */
	for (unsigned int r=0; r < model->NB_CLASSES_ERR; r++)
	{
		er = prior_er_cat[r];
		build_error_matrix(er, er_tstv, mat_er[r]);
	}
	/* calculate likelihood */
	lnL = 0.;
	ret = calculate_lnL_contig();

	if (ret > 0)
	{
		/*cout << "l: imp\n"; getchar();*/
		return 102.;
	}
	lnL += ret;
	return lnL;
}


/* calculate_lnL_contig */

double LnL::calculate_lnL_contig()
{
	double aeb, p, lnL, lnLpos; //Lall, Lerr
	double lnLa; //partialL
	vector<double> v;
	for (unsigned int i=0; i < model->NB_CLASSES_AEB; i++)
	{
		aeb = prior_aeb_cat[i];
		p = prior_aeb_prob[i];
		lnLa = 0.;
		for (unsigned int j=0; j < contig->size(); j++)
		{
			Posdata * pos = data.getPos(j);
			Alrline * alrline = contig->getline(j);
			lnLpos = calculate_lnL_pos(pos, alrline, aeb, i, USE_ERR_ALL_FREQ);
			if (lnLpos > 0) { return 101.; }
			lnLa += lnLpos;
		}

		data.posterior_aeb_p[i] = lnLa + log(p);
		v.push_back(data.posterior_aeb_p[i]);
	}
	lnL = stat->log_sum_exp(v);
	if (CALCULATE_POSTERIOR)
	{
		//posterior aeb distribution for contig
		for (unsigned int i=0; i < model->NB_CLASSES_AEB; i++) {
			data.posterior_aeb_p[i] = exp(data.posterior_aeb_p[i] - lnL);
		}
		//mean posterior aeb for contig
		data.posterior_aeb = 0.;
		for (unsigned int i=0; i < model->NB_CLASSES_AEB; i++) {
			data.posterior_aeb += data.posterior_aeb_p[i]*prior_aeb_cat[i];
		}
		//posterior err distributions + mean for all pos in contig
		for (unsigned int j=0; j < contig->size(); j++) {
			for (unsigned int k=0; k < model->NB_CLASSES_ERR; k++) {
				data.getPos(j)->posterior_er_p[k] = 0.;
				for (unsigned int i=0; i < model->NB_CLASSES_AEB; i++)
					data.getPos(j)->posterior_er_p[k] += data.posterior_aeb_p[i]*data.getPos(j)->cond_posterior_er_p[i][k];
			}
			data.getPos(j)->posterior_er = 0.;
			for (unsigned int k=0; k < model->NB_CLASSES_ERR; k++)
				data.pos_d[j]->posterior_er += data.getPos(j)->posterior_er_p[k]*prior_er_cat[k];
		}
		//posterior genoypes (some recalculation needed here; too large to be kept in memory) */
		calculate_posterior_genotypes();
	}

	return lnL;
}


/* calculate_lnL_pos */
double LnL::calculate_lnL_pos(Posdata * pos, Alrline * line, double aeb, int aeb_class, bool use_all_freq_err)
{
	double lnL, lnLr, lnLindab, lnLind, alfreq[4]; // partialL
	vector<double> vlnLindab, vlncondpost;
	int ind;

	for (unsigned int r=0; r < model->NB_CLASSES_ERR; r++)
	{
		lnLr = 0;
		for (unsigned int k=0; k < LNL_IND.size(); k++)
		{
			ind = LNL_IND[k];
			string species = contig->getspecies(ind);

			if (use_all_freq_err)
			{
				for (unsigned int j=0; j < 4; j++) {
					alfreq[j] = pos->allele_freq_err[species][j];
				}
			} else {
				for (unsigned int j=0; j < 4; j++)
					alfreq[j] = pos->allele_freq_all[species][j];
			}
			vlnLindab.clear();

			for (unsigned int a=0; a < 4; a++)
			{
				if (alfreq[a] == 0.) continue;
				for (unsigned int b=a; b < 4; b++)
				{
					if (alfreq[b] == 0.) continue;
					valarray<unsigned int> values = line->values[slice(ind*4,4,1)];
					lnLindab = calculate_lnL_pos_ind(values, a, b, mat_er[r], aeb, pos->indiv[ind]->log_multinomial_term);

					if (lnLindab <= 0.)
					{
						//vlnLindab.push_back(lnLindab + log(alfreq[a]) + log(alfreq[b]) + cond_log2(a,b));
						if(a!=b) vlnLindab.push_back(lnLindab + log(alfreq[a]) + log(alfreq[b]) + log(2) + log((1-fis_)));
						else vlnLindab.push_back(lnLindab + log( alfreq[a] * alfreq[b] + fis_*alfreq[a]*(1-alfreq[a]) ));
					}


				}
			}
			//cerr << "vlnLindab=" << vlnLindab.size() << endl;
			if (vlnLindab.size() == 0) return 105.; // to improve
			lnLind = stat->log_sum_exp(vlnLindab);
			lnLr += lnLind;
		}
		/*
		partialL = prior_er_prob[r]*exp(lnLr);
		L += partialL;
		if (CALCULATE_POSTERIOR)
		pos_d->cond_posterior_er_p[aeb_class][r] = partialL;
		*/
		vlncondpost.push_back(lnLr + log(prior_er_prob[r]));
	}
	lnL = stat->log_sum_exp(vlncondpost);

	if (CALCULATE_POSTERIOR) {
		for (unsigned int r=0; r < model->NB_CLASSES_ERR; r++) {
			pos->cond_posterior_er_p[aeb_class][r] = exp(vlncondpost[r] - lnL);
		}
	}
	//return log(L);
	return lnL;
}


/* calculate_lnL_pos_ind */
double LnL::calculate_lnL_pos_ind(valarray<unsigned int> & values, int a, int b, double r_mat_er[4][4], double aeb, double log_multi_term)
{
	/*cerr << "calculate_lnL_pos_ind" << endl;
	utils::printVector(values);
	cerr << "a=" << a << " b=" << b << " aeb=" << aeb <<  " log_multi_term=" << log_multi_term << endl;
	for (unsigned int i=0; i<4; i++)
		for (unsigned int j=0; j<4; j++)
			cerr << r_mat_er[i][j] << " ";
	cerr << endl;*/


	double lnL, lnt1, lnt2, lnt, upaeb, umaeb; //diff

	for (int x=0; x < 4; x++)
		if (r_mat_er[a][x] == 0. && r_mat_er[b][x] == 0. && values[x] > 0) return 99.;

	lnL = log_multi_term;

	upaeb = 1 + aeb;
	umaeb = 1 - aeb;
	lnt1 = lnt2 = 0;
	for (int x=0; x < 4; x++)
	{
		if (values[x] == 0 || (r_mat_er[a][x] == 0. && r_mat_er[b][x] == 0.)) continue;
		lnt1 += values[x]*log((r_mat_er[a][x]*upaeb + r_mat_er[b][x]*umaeb)/2);
		lnt2 += values[x]*log((r_mat_er[a][x]*umaeb + r_mat_er[b][x]*upaeb)/2);
	}


	//lnt = log(exp(lnt1) + exp(lnt2));

	/*
	vector<double> v;
	v.push_back(lnt1);
	v.push_back(lnt2);
	lnt = log_sum_exp(&v);
	*/

	lnt = log(1 + exp(lnt2 - lnt1)) + lnt1;
	lnL += lnt;

	/*
	if (lnL > 100 || lnL < -100) {
		cout << lnL << " " << exp(lnL);
		getchar();
	}
	*/
	//cerr << "lnL=" << lnL << "log_multi_term=" << log_multi_term << "lnt1=" << lnt1 << " lnt2=" << lnt2 << endl;
	return lnL;
}

/* calculate_posterior_genotypes */
void LnL::calculate_posterior_genotypes()
{
	double  max, logtot; //partial, sum,
	unsigned int aa, bb;
	vector<double> v;
	bool allzero;

	for (unsigned int j=0; j < contig->size(); j++)
	{
		Posdata * pos = data.getPos(j);
		Alrline * line = contig->getline(j);
		for (unsigned int ae=0; ae < model->NB_CLASSES_AEB; ae++)
			calculate_posterior_genotypes_pos(pos, line, prior_aeb_cat[ae], ae, false);

		for (unsigned int  k=0; k < LNL_IND.size(); k++)
		{
			unsigned int ind = LNL_IND[k];
			Posdata::ind * indiv = pos->indiv[ind];
			allzero = false;

			/* integrate over r and aeb classes (posterior) */
			for (unsigned int a=0; a < 4; a++) {
				for (unsigned int b=a; b < 4; b++) {
					v.clear();
					for (unsigned int r=0; r < model->NB_CLASSES_ERR; r++)
						for (unsigned int ae=0; ae < model->NB_CLASSES_AEB; ae++)
							if (ln_cond_posterior_genotypes[ae][r][a][b][ind] <= 0.) v.push_back(ln_cond_posterior_genotypes[ae][r][a][b][ind] + log(data.posterior_aeb_p[ae]) + log(pos->posterior_er_p[r]));

					if (v.size() == 0) indiv->posterior_genotypes_p[a][b] = 104;
					else indiv->posterior_genotypes_p[a][b] = stat->log_sum_exp(v); // at this stage posterior_genotypes_p is a log
					//cerr << indiv->posterior_genotypes_p[a][b] << " ";
				}
			}
			//cerr << endl;

			v.clear();
			for (int a=0; a < 4; a++)
				for (int b=a; b < 4; b++)
					if (indiv->posterior_genotypes_p[a][b] <= 0.) v.push_back(indiv->posterior_genotypes_p[a][b]);

			if (v.size() == 0)
				allzero = true;
			else
			{
				logtot = stat->log_sum_exp(v);
				for (int a=0; a < 4; a++) {
					for (int b=a; b < 4; b++) {
						if (indiv->posterior_genotypes_p[a][b] > 0)
							indiv->posterior_genotypes_p[a][b] = 0.;
						else
							indiv->posterior_genotypes_p[a][b] = exp(indiv->posterior_genotypes_p[a][b] - logtot);
					}
				}
				if (DO_BEST)  /* find most probable genotype */
				{
					aa = 0; bb = 0; max = indiv->posterior_genotypes_p[0][0];

					for (int a=0; a < 4; a++) {
						for (int b=a; b < 4; b++) {
							if (indiv->posterior_genotypes_p[a][b] > max) {
								max = indiv->posterior_genotypes_p[a][b];
								aa = a;
								bb = b;
							}
						}
					}
					indiv->best_genotype.allele[0] = aa;
					indiv->best_genotype.allele[1] = bb;
					indiv->best_genotype.prob = max;

				} else {  /* randomly draw genotypes */
					double cum = 0.;
					double ra = RandomTools::giveRandomNumberBetweenZeroAndEntry(1.);
					bool br;

					for (int a=0; a < 4; a++) {
						br = false;
						for (int b=a; b < 4; b++) {
							cum += indiv->posterior_genotypes_p[a][b];
							if (cum >= ra) {
								indiv->best_genotype.allele[0] = a;
								indiv->best_genotype.allele[1] = b;
								indiv->best_genotype.prob = indiv->posterior_genotypes_p[a][b];
								br = true;
								break;
							}
						}
						if (br) break;
					}
				}
			}

			if (allzero || line->numberOfreads(ind) < min_reads_per_ind)
			{
				indiv->best_genotype.allele[0] = 100;
				indiv->best_genotype.allele[1] = 100;
				indiv->best_genotype.prob = 0.;
			}
		}
	}
}

/* calculate_posterior_genotypes_pos */
void LnL::calculate_posterior_genotypes_pos(Posdata * pos, Alrline * line, double aeb, int aeb_class, bool use_all_freq_err)
{
	double  alfreq[4];// max, Lindab
	unsigned int ind;
	string sp;

	for (unsigned int r=0; r < model->NB_CLASSES_ERR; r++)
	{
		for (unsigned int k=0; k < LNL_IND.size(); k++)
		{
			Posdata::ind * indiv = pos->indiv[k];
			ind = LNL_IND[k];
			sp = contig->getspecies(ind);
			if (use_all_freq_err) {
				for (unsigned int j=0; j < 4; j++) {
					alfreq[j] = pos->allele_freq_err[sp][j];
				}
			} else {
				for (unsigned int j=0; j < 4; j++) {
					alfreq[j] = pos->allele_freq_all[sp][j];
				}
			}

			for (unsigned int a=0; a < 4; a++)
			{
				if (alfreq[a] == 0.)
				{
					for (unsigned int b=0; b < 4; b++)
						ln_cond_posterior_genotypes[aeb_class][r][a][b][ind] = 103.;
					continue;
				}
				for (unsigned int b=a; b < 4; b++)
				{
					if (alfreq[b] == 0.)
					{
						ln_cond_posterior_genotypes[aeb_class][r][a][b][ind] = 103.;
						continue;
					}

					valarray<unsigned int> values = line->values[slice(ind*4,4,1)];
					ln_cond_posterior_genotypes[aeb_class][r][a][b][ind] = calculate_lnL_pos_ind(values, a, b, mat_er[r], aeb, indiv->log_multinomial_term);

					//ln_cond_posterior_genotypes[aeb_class][r][a][b][ind] += log(alfreq[a])+log(alfreq[b]);
					//if (a != b) ln_cond_posterior_genotypes[aeb_class][r][a][b][ind] += log(2);
					if(a!=b) ln_cond_posterior_genotypes[aeb_class][r][a][b][ind] += log(2)+log(alfreq[a])+log(alfreq[b])+log((1-fis_));
					else     ln_cond_posterior_genotypes[aeb_class][r][a][b][ind] += log( alfreq[a] * alfreq[b] + (fis_*alfreq[a]*(1-alfreq[a])));
				}
			}
		}
	}
}


/* 	build_error_matrix */
void LnL::build_error_matrix(double ertot, double tstv, double mat_er[4][4])
{

	double er = ertot/(2. + tstv);

	mat_er[0][1] = mat_er[1][0] = er;
	mat_er[0][3] = mat_er[3][0] = er;
	mat_er[1][2] = mat_er[2][1] = er;
	mat_er[2][3] = mat_er[3][2] = er;
	mat_er[0][2] = mat_er[2][0] = tstv*er;
	mat_er[1][3] = mat_er[3][1] = tstv*er;

	mat_er[0][0] = 1 - (tstv + 2)*er;
	mat_er[1][1] = 1 - (tstv + 2)*er;
	mat_er[2][2] = 1 - (tstv + 2)*er;
	mat_er[3][3] = 1 - (tstv + 2)*er;
}

/* cond_log2 */
double LnL::cond_log2(int a, int b)
{
	if (a != b) return log(2);
	return 0.;
}



}
