//============================================================================
// Name        : Reads2snp.h
// Author      : Vincent Cahais, Nicolas Galtier
// Created on  : 17 septembre 2012
// Copyright   : http://www.cecill.info/licences/Licence_CeCILL-C_V1-fr.html
//============================================================================

#include "Reads2snp.h"

using namespace std;

namespace bpp
{
	/**************************************************************************************/
	/*Reads2snp : constructor*/
	/**************************************************************************************/
	Reads2snp::Reads2snp(Stat * s): stat(s), min_reads_per_ind(1), do_sh(0), do_sep(0), do_best(0),
			use_paraclean(0), MULTI_ALLELES("tolerated"), fis_(0.0), threshold_(0.001), prec_(0.001),
			dirichlet_ (0), skipParalogous_(0), keepunresolved_(0), optimizer_("newton"), error_(0)
	{
		//models initialization
		//               		  name  aligned   description                   do NB_CLASSES_ERR	NB_CLASSES_AEB	ERR_EQUALS_ZERO	AEB_EQUALS_ZERO	er_beta	er_tstv	aeb_beta
		models.push_back(new Model("M0" ,"M0" ,"[no error, no bias]            ",0  ,1				,1				,true			,true			,false	,false	,false));
		models.push_back(new Model("M1", "M1" ,"[constant error, no bias]      ",0  ,1				,1				,false			,true			,true	,true	,false));
		models.push_back(new Model("M2" ,"M2" ,"[constant error, constant bias]",0  ,1				,1				,false			,false			,true	,true	,true));

		//models.push_back(new Model("M1b","M1b","[no error, constant bias]      ",0  ,1				,1				,true			,false			,false	,false	,true));
		//models.push_back(new Model("M2a","M2a","[variable error, no bias]      ",0  ,MAX_NB_CLASSES	,1				,false			,true			,true	,true	,false));
		//models.push_back(new Model("M2b","M2b","[no error, variable bias]      ",0  ,1				,MAX_NB_CLASSES	,true			,false			,false	,false	,true));
		//models.push_back(new Model("M3a","M3a","[variable error, constant bias]",0  ,MAX_NB_CLASSES	,1				,false			,false			,true	,true	,true));
		//models.push_back(new Model("M3b","M3b","[constant error, variable bias]",0  ,1				,MAX_NB_CLASSES	,false			,false			,true	,true	,true));
		//models.push_back(new Model("M4" ,"M4" ,"[variable error, variable bias]",0  ,MAX_NB_CLASSES	,MAX_NB_CLASSES	,false			,false			,true	,true	,true));

		lnLM_sh.resize(models.size(),0.0);
		lnLM_sep.resize(models.size(),0.0);
	}

	Reads2snp::~Reads2snp()
	{
		if (calcul) delete calcul;
		for (unsigned int i=0; i<models.size(); i++)
			delete models[i];
	}

	/**************************************************************************************/
	/*Reads2snp::compute
	 * @INPUT : AlrContig Object
	 * @DESC  : Compute reads2snp on given contig
	 */
	/**************************************************************************************/
	void Reads2snp::compute(AlrContig & contig)
	{
		//create LnL object
		//cout << "create LnL object step" << endl;
		calcul = new LnL(stat, &contig, models[0], min_reads_per_ind, fis_);
		calcul->DO_BEST = do_best;

		//Check_paralogy
		//cout << "Check_paralogy step" << endl;
		check_paralogy(contig);
		cout << "paralogy=" << calcul->data.species_data[0]->paralogy_counter << endl;

		//Read to snp
		//cout << "Read to snp step" << endl;
		for(unsigned int position=0; position<contig.size(); position++)
		{
			Posdata * pos = calcul->data.getPos(position);
			//calculate allele frequencies per pos
			for (unsigned int sp=0; sp < contig.numberofspecies(); sp++)
			{
				string current_species = contig.getspeciesname(sp);
				valarray<unsigned int> values = contig.getvaluesbyspecies(position, current_species);
				pos->calculate_allele_frequencies(values,current_species, *stat);
			}
			//calculate multinomial terms
			calculate_multinomial_terms(*calcul->data.getPos(position) ,contig.getline(position)->values);
			//
			NO_ERROR_POSSIBLE = pos->check_no_error_possible(contig.getline(position)->values);
		}
		//calculate max likelihoods
		//cout << "Maximizing likelihood..." << endl;
		vector<Optimizer*> opt_sh(models.size(),0);
		vector< vector<Optimizer*> > opt_sep(models.size(), vector<Optimizer*>(contig.numberofspecies(),0));
		if (models[0]->do_model) model0(contig);
		try{
			if (do_sh) dosh(contig, opt_sh);
			if (do_sep && contig.numberofspecies() > 1) dosep(contig, opt_sep);
		}
		catch (Exception & e) {
			cerr << "ERROR : Optimization failed" << endl;
			calcul->data.in_error = true;
			return;
		}
		//calculate posterior probabilities
		//cout << "Calculating posterior probabilities..." << endl;
		calcul->CALCULATE_POSTERIOR = true;
		//cout << "CALCULATE_POSTERIOR" << endl;
		if (do_sep) chosen_sh_sep(contig, opt_sep);
		else chosen_sh_dist(contig, opt_sh);
	}

	/**************************************************************************************/
	/*Reads2snp::calculate_multinomial_terms
	* @INPUT : Posdata object, position values
    * @DESC  : calculate multinomial terms
	*/
	/**************************************************************************************/
	void Reads2snp::calculate_multinomial_terms(Posdata & pos, valarray<unsigned int> & values)
	{
		double lnL;

		for (unsigned int i=0; i < pos.numberofindiv(); i++)
		{
			lnL = 0.;
			unsigned int total_number_of_reads = 0;
			for (unsigned int j=0; j < 4; j++)
			{
				lnL -= stat->logfact(values[(i*4)+j]);
				total_number_of_reads+=values[(i*4)+j];
			}

			lnL += stat->logfact(total_number_of_reads);
			lnL -= log(2);
			pos.indiv[i]->log_multinomial_term = lnL;
		}
	}

	/**************************************************************************************/
	/*Reads2snp::is_more_than_2_alleles
	* @INPUT : position values
    * @DESC  : calculate if there is more than 2 alleles for selected position and species
	*/
	/**************************************************************************************/
	bool Reads2snp::is_more_than_2_alleles(valarray<unsigned int> values)
	{
		unsigned int nb_sig_all;
		unsigned int nb_ind = values.size()/4;
		double prop_thresh = SIG_ALLELE_MIN_PERCENT/100.;
		bool morethan2=false;

		for (unsigned int i=0; i < nb_ind; i++)
		{
			nb_sig_all=0;
			double sum = values[(i*4)]+values[(i*4)+1]+values[(i*4)+2]+values[(i*4)+3];
			for (unsigned int j=0; j < 4; j++)
			{
				if (values[(i*4)+j] >= SIG_ALLELE_MIN_NBREADS && (double)values[(i*4)+j]/sum >= prop_thresh)
					nb_sig_all++;
			}
			if (nb_sig_all > 2) morethan2 = true;
		}
		return morethan2;
	}

	/**************************************************************************************/
	/*Reads2snp::check_paralogy
	* @INPUT : AlrContig Object
	* @DESC  : check paralogy
	*/
	/**************************************************************************************/
	void Reads2snp::check_paralogy(AlrContig & contig)
	{
		if (MULTI_ALLELES != "accepted")
		{
			for (unsigned int line=0; line<contig.size(); line++)
			{
				if (!contig.ispolymorph(line)) continue;
				for (unsigned int species=0; species < contig.numberofspecies(); species++)
				{
					string curr_species = contig.getspecies(species);
					bool morethan2 = is_more_than_2_alleles(contig.getvaluesbyspecies(line,curr_species));
					if (morethan2)
					{
						calcul->data.species_data[species]->paralogy_counter++;
						if (MULTI_ALLELES == "tolerated")
						{
							for (unsigned int indiv=0; indiv < contig.numberofindiv(curr_species); indiv++)
							{
								if (contig.getspecies(indiv) == curr_species)
								{
									for (unsigned int k=0; k < 4; k++)
										contig.getline(line)->setValue((indiv*4+k), 0);
								}
							}
						}
						else
						{
							cout << "More than 2 alleles detected at position " << line << endl;
							exit(0);
						}
					}
				}
			}
		}
	}

	/**************************************************************************************/
	/*Reads2snp::model0
	* @INPUT : AlrContig Object
	* @DESC  : calcul likelihood for model M0
	*/
	/**************************************************************************************/
	void Reads2snp::model0(AlrContig & contig)
	{
		calcul->LNL_IND.clear();
		for (unsigned int i=0; i < contig.numberofindiv(); i++) calcul->LNL_IND.push_back(i);

		calcul->setModel(models[0]);

		calcul->CALCULATE_POSTERIOR = false;

		if (NO_ERROR_POSSIBLE) lnLM_sh[0] = calcul->lnL(0, 0, 1, 0, 0);
		else lnLM_sh[0] = log(0);

		if (lnLM_sh[0] <= 0.)
			cout << models[0]->aligned << " " << models[0]->description << "\tlnL=" << lnLM_sh[0] << "\t(epsilon=0, alpha=0)" << endl;
		else
			cout << models[0]->aligned << " " << models[0]->description << "\tlnL=-inf" << "\t(epsilon=0, alpha=0)" << endl;
	}

	/**************************************************************************************/
	/*Reads2snp::dosh
	* @INPUT : AlrContig Object, vector of Optimizer*
	* @DESC  : calcul likelihood for all models (except M0) with parameters shared across species
	*/
	/**************************************************************************************/
	void Reads2snp::dosh(AlrContig & contig, vector<Optimizer*> & opt_sh)
	{
		if (contig.numberofspecies() > 1) cout << "Shared (across species) parameters:" << endl;

		calcul->LNL_IND.clear();
		for (unsigned int i=0; i < contig.numberofindiv(); i++) calcul->LNL_IND.push_back(i);

		calcul->CALCULATE_POSTERIOR = false;

		for (unsigned int i=1; i < models.size(); i++)
		{
			if (!models[i]->do_model) continue;
			double err, aeb, a, b;

			if (models[i]->ERR_EQUALS_ZERO && !NO_ERROR_POSSIBLE)
			{
				lnLM_sh[i] = log(0.);
				cout << models[i]->aligned << " " << models[i]->description << "\tlnL=-inf" << "\t(epsilon=0, alpha=0)" << endl;
				continue;
			}
			calcul->setModel(models[i]);

			try{opt_sh[i] = my_optimize();} catch(Exception & e){throw e;}

			lnLM_sh[i] =-opt_sh[i]->getFunctionValue();

			if (!models[i]->ERR_EQUALS_ZERO)
			{
				a = opt_sh[i]->getFunction()->getParameterValue("er_alpha");
				b = opt_sh[i]->getFunction()->getParameterValue("er_beta");
				err = a/(a + b);
			}
			else err = 0.;

			if (!models[i]->AEB_EQUALS_ZERO)
			{
				a = opt_sh[i]->getFunction()->getParameterValue("aeb_alpha");
				b = opt_sh[i]->getFunction()->getParameterValue("aeb_beta");
				aeb = a/(a + b);
			}
			else aeb = 0.;

			if (lnLM_sh[i] <= 0.)
				cout << models[i]->aligned << " " << models[i]->description << "\tlnL=" << lnLM_sh[i] << "\t(epsilon=" << err << ", alpha=" << aeb << ")" << endl;
			else
				cout << models[i]->aligned[i] << " " << models[i]->description << "\tlnL=-inf" << "\t(epsilon=" << err << ", alpha=" << aeb << ")" << endl;
		}
	}

	/**************************************************************************************/
	/*Reads2snp::dosep
	* @INPUT : AlrContig Object, vector of Optimizer*
	* @DESC  : calcul likelihood for all models (except M0) with parameters separated across species
	*/
	/**************************************************************************************/
	void Reads2snp::dosep(AlrContig & contig, vector< vector<Optimizer*> > & opt_sep)
	{
		cout << "Separate (between species) parameters:" << endl;
		for (unsigned int i=1; i < models.size(); i++)
		{
			if (!models[i]->do_model) continue;
			if (models[i]->ERR_EQUALS_ZERO && !NO_ERROR_POSSIBLE)
			{
				lnLM_sep[i] = log(0.);
				cout << "  --> " << models[i]->name << " " << models[i]->description << "\tlnL=-inf"  << endl;
				continue;
			}
			calcul->setModel(models[i]);

			calcul->CALCULATE_POSTERIOR = false;

			lnLM_sep[i] = 0.;
			map<string, unsigned int>::iterator it;

			for (unsigned int j=0; j<contig.numberofspecies(); j++)
			{
				calcul->LNL_IND.clear();
				for (unsigned int x=0; x< contig.numberofindiv(); x++)
					if (contig.getspecies(x) == contig.getspecies(j)) calcul->LNL_IND.push_back(x);

				try{opt_sep[i][j]=my_optimize();} catch(Exception & e){throw e;}
				lnLM_sep[i] -= opt_sep[i][j]->getFunctionValue();

			}

			if (lnLM_sep[i] <= 0.)
				cout << "  --> " << models[i]->aligned << " " << models[i]->description[i] << "\tlnL=" << lnLM_sep[i] << endl;
			else
				cout << "  --> " << models[i]->aligned << " " << models[i]->description[i] << "\tlnL=-inf"  << endl;
		}
		cout << endl;
	}

	/**************************************************************************************/
	/*Reads2snp::chosen_sh_sep
	* @INPUT : AlrContig Object, vector of Optimizer*
	* @DESC  : calcul likelihood and posterior probabilities for chosen models (except M0) with shared separated across species
	*/
	/**************************************************************************************/
	void Reads2snp::chosen_sh_sep(AlrContig & contig, vector< vector<Optimizer*> > & opt_sep)
	{
		double opt_er_beta=0.0, opt_tstv=0.0, opt_aeb_beta=0.0, lnl=0.0;

		for (unsigned int sp=0; sp < contig.numberofspecies(); sp++)
		{
			if (chosen_model > 0)
			{
				opt_er_beta = opt_sep[chosen_model][sp]->getFunction()->getParameterValue("er_beta");
				opt_tstv = opt_sep[chosen_model][sp]->getFunction()->getParameterValue("er_tstv");
				opt_aeb_beta = opt_sep[chosen_model][sp]->getFunction()->getParameterValue("aeb_beta");
			}

			calcul->LNL_IND.clear();
			for (unsigned int i=0; i< contig.numberofindiv(); i++)
				if (contig.getspecies(i) == contig.getspecies(sp)) calcul->LNL_IND.push_back(i);

			calcul->setModel(models[chosen_model]);
			if (chosen_model > 0) lnl = calcul->lnL(1, opt_er_beta, opt_tstv, 1, opt_aeb_beta);
			else lnl = calcul->lnL(0, 0, 1, 0, 0);

			// prepare output
			// aeb
			Contigdata::aeb_data * aeb = new Contigdata::aeb_data;
			calcul_aeb(aeb, opt_aeb_beta);
			calcul->data.aeb.push_back(aeb);

			//err
			Contigdata::err_data * err = new Contigdata::err_data;
			calcul_err(err, opt_er_beta, opt_tstv, contig.size());
			calcul->data.err.push_back(err);
		}
	}

	/**************************************************************************************/
	/*Reads2snp::chosen_sh_dist
	* @INPUT : AlrContig Object, vector of Optimizer*
	* @DESC  : calcul likelihood and posterior probabilities for chosen models (except M0) with parameters separated across species
	*/
	/**************************************************************************************/
	void Reads2snp::chosen_sh_dist(AlrContig & contig, vector<Optimizer*> & opt_sh)
	{
		double opt_er_beta=0.0, opt_tstv=0.0, opt_aeb_beta=0.0, lnl=0.0;

		if (chosen_model > 0)
		{
			opt_er_beta = opt_sh[chosen_model]->getFunction()->getParameterValue("er_beta");
			opt_tstv = opt_sh[chosen_model]->getFunction()->getParameterValue("er_tstv");
			opt_aeb_beta = opt_sh[chosen_model]->getFunction()->getParameterValue("aeb_beta");
		}

		calcul->LNL_IND.clear();
		for (unsigned int i=0; i < contig.numberofindiv(); i++) calcul->LNL_IND.push_back(i);
		calcul->setModel(models[chosen_model]);
		if (chosen_model > 0) lnl = calcul->lnL(1, opt_er_beta, opt_tstv, 1, opt_aeb_beta);
		else lnl = calcul->lnL(0, 0, 1, 0, 0);

		//prepare output
		//aeb
		Contigdata::aeb_data * aeb = new Contigdata::aeb_data;
		calcul_aeb(aeb, opt_aeb_beta);
		calcul->data.aeb.push_back(aeb);

		//err
		Contigdata::err_data * err = new Contigdata::err_data;
		calcul_err(err, opt_er_beta, opt_tstv, contig.size());
		calcul->data.err.push_back(err);
	}

	/**************************************************************************************/
	/*Reads2snp::calcul_aeb
	* @INPUT : calculated data for contig, ?
	* @DESC  : calcul aeb
	*/
	/**************************************************************************************/
	void Reads2snp::calcul_aeb(Contigdata::aeb_data * aeb, double opt_aeb_beta)
	{
		if (models[chosen_model]->AEB_EQUALS_ZERO) aeb->out_av_aeb = 0.;
		else aeb->out_av_aeb = 1/(1 + opt_aeb_beta);

		if (models[chosen_model]->NB_CLASSES_AEB > 1)
		{
			aeb->out_contig_posterior_aeb = calcul->data.posterior_aeb;
			for (unsigned int k=0; k < MAX_NB_CLASSES; k++)
			{
				aeb->out_prior_aeb_cat.push_back(calcul->prior_aeb_cat[k]);
				aeb->out_contig_posterior_aeb_p.push_back(calcul->data.posterior_aeb_p[k]);
			}
		}
	}

	/**************************************************************************************/
	/*Reads2snp::calcul_err
	* @INPUT : calculated data for contig, ? , ? ,contig size
	* @DESC  : calcul err
	*/
	/**************************************************************************************/
	void Reads2snp::calcul_err(Contigdata::err_data * err, double opt_er_beta, double opt_tstv, unsigned int size)
	{
		if (models[chosen_model]->ERR_EQUALS_ZERO)
		{
			err->out_av_err = 0.;
			err->out_tstv = 0.;
		}
		else
		{
			err->out_av_err = 1/(1 + opt_er_beta);
			err->out_tstv = opt_tstv;
		}
		if (models[chosen_model]->NB_CLASSES_ERR > 1)
		{
			for (unsigned int k=0; k < MAX_NB_CLASSES; k++)
				err->out_prior_er_cat.push_back(calcul->prior_er_cat[k]);

			for (unsigned int i=0; i < size; i++)
			{
				err->out_pos_posterior_er.push_back(calcul->data.getPos(i)->posterior_er);
				vector<double> tmp;
				for (unsigned int k=0; k < MAX_NB_CLASSES; k++)
					tmp.push_back(calcul->data.getPos(i)->posterior_er_p[k]);
				err->out_pos_posterior_er_p.push_back(tmp);
			}
		}
	}


	/**************************************************************************************/
	/*Reads2snp::my_optimize
	* @DESC  : optimization function
	*/
	/**************************************************************************************/
	Optimizer* Reads2snp::my_optimize()
	{
		LnL_fct * mlf = new LnL_fct(*calcul);
		ParameterList pl = mlf->getParameters();

		//ReparametrizationFunctionWrapper rpf1(mlf);
		//rpf1.setParameters(mlf->getParameters());
		//ParameterList pl = rpf1.getParameters();

		//ThreePointsNumericalDerivative tpnd1(&rpf1);
		//tpnd1.setParametersToDerivate(pl1.getParameterNames());

		Optimizer * opt = new DownhillSimplexMethod(mlf);
		//Optimizer * opt = new BfgsMultiDimensions(&tpnd1);
		opt->setVerbose(0);
		opt->setProfiler(0);
		opt->setMessageHandler(0);
		opt->setConstraintPolicy(AutoParameter::CONSTRAINTS_AUTO);

		vector<string> plist;
		if (calcul->model->er_beta) plist.push_back("er_beta");
		if (calcul->model->er_tstv) plist.push_back("er_tstv");
		if (calcul->model->aeb_beta) plist.push_back("aeb_beta");
		ParameterList pl2 = pl.subList(plist);
		opt->init(pl2);
		try{opt->optimize();} catch(Exception & e){throw e;}
		return opt;
	}

	/**************************************************************************************/
	/*Reads2snp::write_speciesdata
	* @INPUT : output stream
	* @DESC  : write info about each species
	*/
	/**************************************************************************************/
	void Reads2snp::write_speciesdata(ofstream & out)
	{
		// Initializations
		string no_value = "NA";

		if (calcul->data.in_error) return;
		// Write data into the additional output file
		for (unsigned int i=0; i<calcul->data.species_data.size(); i++)
		{
			calcul->data.species_data[i]->allelic_expression_bias = calcul->data.aeb[i]->out_av_aeb;
			calcul->data.species_data[i]->transversion_error_rate = calcul->data.err[i]->out_av_err;
			calcul->data.species_data[i]->transition_error_rate = calcul->data.err[i]->out_tstv * calcul->data.err[i]->out_av_err;
			calcul->data.species_data[i]->tstv_ratio = calcul->data.err[i]->out_tstv;
			// Species name
			if (calcul->data.species_data[i]->species_name != "") out << "Species_name=" << calcul->data.species_data[i]->species_name << ";";
			else out << "Species_name=" << no_value << ";";
			// Estimated (average) allelic expression bias
			if (calcul->data.species_data[i]->allelic_expression_bias != -1.) out << "Allelic_expression_bias=" <<calcul->data.species_data[i]->allelic_expression_bias << ";";
			else out << "Allelic_expression_bias=" << no_value << ";";
			// Estimated (average) error rates
			if (calcul->data.species_data[i]->transversion_error_rate != -1.) out << "Transversion_error_rate=" << calcul->data.species_data[i]->transversion_error_rate << ";";
			else out << "Transversion_error_rate=" << no_value << ";";
			if (calcul->data.species_data[i]->transition_error_rate != -1.) out << "Transition_error_rate=" << calcul->data.species_data[i]->transition_error_rate << ";";
			else out << "Transition_error_rate=" << no_value << ";";
			if (calcul->data.species_data[i]->tstv_ratio != -1.) out << "TsTv_ratio=" << calcul->data.species_data[i]->tstv_ratio << ";";
			else out << "TsTv_ratio=" << no_value << ";";
			// Cases of supposed paralogy
			out << "Multi_alleles_positions=" << calcul->data.species_data[i]->paralogy_counter << ";";
			out << endl;
		}
	}


	/**************************************************************************************/
	/*Reads2snp::write_aeb
	* @INPUT : AlrContig object, output stream
	* @DESC  : write aeb info
	*/
	/**************************************************************************************/
	void Reads2snp::write_aeb(AlrContig & contig, std::ofstream & outfile_aeb)
	{
		string model = models[chosen_model]->name;
		cout << endl << "Estimated (average) allelic expression bias:" << endl;

		unsigned int nbout=1;
		if (do_sep == true) nbout = contig.numberofspecies();

		for (unsigned int i=0; i < nbout; i++)
		{
			cout << "  --> " << contig.getspecies(i) << " : " << calcul->data.aeb[i]->out_av_aeb << endl;
			calcul->data.species_data[i]->allelic_expression_bias = calcul->data.aeb[i]->out_av_aeb;
		}

		if (model == "M2b" || model == "M3b" || model == "M4")
		{
			outfile_aeb << "contig";
			for (unsigned int k=0; k < nbout; k++)
			{
				outfile_aeb << "\tav_aeb[" << contig.getspecies(k) << "]";
				for (unsigned int j=0; j < calcul->model->NB_CLASSES_AEB; j++)
					outfile_aeb << "\tprob_class" << j + 1 << "(" << calcul->data.aeb[k]->out_prior_aeb_cat[j] << ")";
			}

			outfile_aeb << endl;
			outfile_aeb << contig.name;
			for (unsigned int k=0; k < nbout; k++)
			{
				outfile_aeb << "\t" << calcul->data.aeb[k]->out_contig_posterior_aeb;
				for (unsigned int j=0; j < calcul->model->NB_CLASSES_AEB; j++)
					outfile_aeb << "\t" << calcul->data.aeb[k]->out_contig_posterior_aeb_p[j];
			}
			outfile_aeb << endl;
		}
	}

	/**************************************************************************************/
	/*Reads2snp::write_err
	* @INPUT : AlrContig object, output stream
	* @DESC  : write err info
	*/
	/**************************************************************************************/
	void Reads2snp::write_err(AlrContig & contig, ofstream & outfile_err)
	{
		//ERR
		string model = models[chosen_model]->name;
		cout << endl << "Estimated (average) error rates [transversions, transitions, Ts/Tv ratio]:" << endl;

		unsigned int nbout=1;
		if (do_sep == true)
		{
			nbout = contig.numberofspecies();

			for (unsigned int i=0; i < nbout; i++)
			{
				calcul->data.species_data[i]->species_name = contig.getspecies(i);
				calcul->data.species_data[i]->transversion_error_rate = calcul->data.err[i]->out_av_err;
				calcul->data.species_data[i]->transition_error_rate = calcul->data.err[i]->out_tstv*calcul->data.err[i]->out_av_err;
				calcul->data.species_data[i]->tstv_ratio = calcul->data.err[i]->out_tstv;

				cout << "  --> " << contig.getspecies(i) << " : " << calcul->data.err[i]->out_av_err << ", " << calcul->data.species_data[i]->transition_error_rate << ", " << calcul->data.err[i]->out_tstv << endl;
			}
		}
		if (model == "M2a" || model == "M3a" || model == "M4")
		{
			outfile_err << "contig\tposition";

			for (unsigned int k=0; k < nbout; k++)
			{
				outfile_err << "\tav_err[" << contig.getspecies(k) << "]";
				for (unsigned int j=0; j < calcul->model->NB_CLASSES_ERR; j++)
					outfile_err << "\tprob_class" << j + 1 << "(" << calcul->data.err[k]->out_prior_er_cat[j] << ")";
			}
			outfile_err << endl;

			for (unsigned int j=0; j < contig.size(); j++)
			{
				outfile_err << contig.name << "\t" << j + 1;
				for (unsigned int kk=0; kk < nbout; kk++)
				{
					outfile_err << "\t" << calcul->data.err[kk]->out_pos_posterior_er[j];
					for (unsigned int k=0; k < calcul->model->NB_CLASSES_ERR; k++)
						outfile_err << "\t" << calcul->data.err[kk]->out_pos_posterior_er_p[j][k];
				}
				outfile_err << endl;
			}
			outfile_err.close();
		}
	}

	/**************************************************************************************/
	/*Reads2snp::write_genotypes
	* @INPUT : AlrContig object, output stream
	* @DESC  : write main output (alr.gen file)
	*/
	/**************************************************************************************/
	string Reads2snp::write_genotypes(AlrContig & contig, ofstream & outfile_geno)
	{
		reset_outlog();

		ostringstream contigStream;
		contigStream << ">" << contig.name << endl;
		contigStream << "position";
		string zero = "";
		outlog.nb_positions = contig.size();

		//headers
		for (unsigned int j=0; j < contig.numberofindiv(); j++)
		{
			contigStream << "\t" << contig.getspecies(j) << "|" << contig.getindiv(j);
			zero += "\tNN|0";
		}
		contigStream << endl;

		//If optimization returned an error
		if (calcul->data.in_error)
		{
			for (unsigned int j=0; j < contig.size(); j++)
				outfile_geno << j + 1 << zero << endl;
			return "";
		}

		for (unsigned int j=0; j < contig.size(); j++)
		{
			contigStream << j + 1;

			ostringstream line;
			bool isheterozygous = false;
			for (unsigned int k=0; k < contig.numberofindiv(); k++)
			{
				line << "\t";
				line << utils::tont(calcul->data.getPos(j)->indiv[k]->best_genotype.allele[0]);
				line << utils::tont(calcul->data.getPos(j)->indiv[k]->best_genotype.allele[1]);
				line << "|" << calcul->data.getPos(j)->indiv[k]->best_genotype.prob;
				if (calcul->data.getPos(j)->indiv[k]->best_genotype.allele[0] != calcul->data.getPos(j)->indiv[k]->best_genotype.allele[1])
					isheterozygous = true;
			}
			if (!isheterozygous) outlog.nb_homozygous++;

			//paraclean
			if (use_paraclean)
			{
				if (paraclean(*contig.getline(j), isheterozygous, j))
				{
					if (skipParalogous_) return "";
					outlog.nb_hidden_positions++;
					line.str("");
					line << zero;
				}
			}
			contigStream << line.str() << endl;
		}
		outfile_geno << contigStream.str();
		return contigStream.str();
	}

	bool Reads2snp::paraclean(Alrline & pos, bool isheterozygous, unsigned int line)
	{
		Paraclean para;
		para.set_fis(fis_);
		para.set_threshold(threshold_);
		para.set_dirichlet(dirichlet_);
		para.set_prec(prec_);
		para.set_optimizer(optimizer_);

		//Warning: if error value > 0.333333 paraclean may crach.
		if (error_ > 0 && error_ < 0.333333) para.set_error(error_);
		else
		{
			double error = calcul->data.err[0]->out_tstv*calcul->data.err[0]->out_av_err;
			if (error > 0.333333) error = 0.333333;
			para.set_error(error);
		}

		if (pos.ispolymorph && isheterozygous)
		{
			outlog.nb_checked_positions++;
			double value = para.assignation(pos.values, *stat);
			cout << "paraclean on line " << line+1 << ", pvalue= " << value << endl;
			if (value < 0) outlog.nb_position_in_error++;
			if (value < threshold_ && !(value <0 && keepunresolved_)) return true;
		}
		return false;
	}

	void Reads2snp::reset_outlog()
	{
		//compteurs
		outlog.nb_homozygous = 0;
		outlog.nb_checked_positions = 0;
		outlog.nb_hidden_positions  = 0;
		outlog.nb_position_in_error = 0;
	}




/********************************************************/
}
