//============================================================================
// Name        : Reads2snp.cpp
// Author      : Vincent Cahais, Nicolas Galtier
// Created on  : 17 septembre 2012
// Copyright   : http://www.cecill.info/licences/Licence_CeCILL-C_V1-fr.html
//============================================================================

#include "../Stat/Stat.h"
#include "../Alr.hpp"
#include "Contigdata.h"
#include "LnL.h"
#include "LnL_fct.hpp"
#include "Model.hpp"
#include "../Paraclean/Paraclean.h"

#include <valarray>

#ifndef READ2SNP_H_
#define READ2SNP_H_

#define CANDIDATE_ERR_THR 0.01
#define SIG_ALLELE_MIN_NBREADS 3
#define SIG_ALLELE_MIN_PERCENT 5

namespace bpp
{

class Reads2snp
{

private:

	bool NO_ERROR_POSSIBLE;

	valarray<double> log_multinomial_term;
	vector<double> lnLM_sh;  // likelihoods results (shared option)
	vector<double> lnLM_sep; // likelihoods results (separated option)

	double cond_lnL_aeb[MAX_NB_CLASSES];

	Stat * stat;             // calcul methods
	LnL * calcul;            // likelihoods calcul methods

	vector<Model*> models;


public:

	//parameters
	unsigned int chosen_model;       // chosen model used to calculate posterior probabilities
	unsigned int min_reads_per_ind;
	bool do_sh, do_sep, do_best;
	bool use_paraclean;
	string MULTI_ALLELES;
	double fis_;
	double threshold_;
	double prec_;
	bool dirichlet_;
	bool skipParalogous_;
	bool keepunresolved_;
	string optimizer_;
	double error_;

	struct Log
	{
		unsigned int nb_positions;
		unsigned int nb_homozygous;
		unsigned int nb_checked_positions;
		unsigned int nb_hidden_positions;
		unsigned int nb_position_in_error;
	} outlog;

public:

	/**************************************************************************************/
	/*Reads2snp : constructor*/
	/**************************************************************************************/
	Reads2snp(Stat * s);
	virtual ~Reads2snp();

	/**************************************************************************************/
	/*Reads2snp::models*/
	/**************************************************************************************/
	void domodel(unsigned int i){models[i]->do_model=true;}
	unsigned int getnumberofmodels(){return models.size();}
	unsigned int getnumberofactivemodels()
	{
		unsigned int mcount=0;
		for (unsigned int i=0; i<getnumberofmodels(); i++)
				mcount+=models[i]->do_model;
		return mcount;
	}
	void doallmodel()
	{
		for (unsigned int i=0; i<getnumberofmodels(); i++)
			models[i]->do_model = true;
	}

	/**************************************************************************************/
	/*Reads2snp::compute
	 * @INPUT : AlrContig Object
	 * @DESC  : Compute reads2snp on given contig
	 */
	/**************************************************************************************/
	void compute(AlrContig & contig);

	/**************************************************************************************/
	/*Reads2snp::write_aeb
	* @INPUT : AlrContig object, output stream
	* @DESC  : write aeb info
	*/
	/**************************************************************************************/
	void write_aeb(AlrContig & contig, ofstream & outfile_aeb);

	/**************************************************************************************/
	/*Reads2snp::write_err
	* @INPUT : AlrContig object, output stream
	* @DESC  : write err info
	*/
	/**************************************************************************************/
	void write_err(AlrContig & contig, ofstream & outfile_err);

	/**************************************************************************************/
	/*Reads2snp::write_genotypes
	* @INPUT : AlrContig object, output stream
	* @DESC  : write main output (alr.gen file)
	*/
	/**************************************************************************************/
	string write_genotypes(AlrContig & contig, ofstream & outfile_geno);

	/**************************************************************************************/
	/*Reads2snp::write_speciesdata
	* @INPUT : output stream
	* @DESC  : write info about each species
	*/
	/**************************************************************************************/
	void write_speciesdata (ofstream & out);

private:

	/**************************************************************************************/
	/*Reads2snp::check_paralogy
	* @INPUT : AlrContig Object
	* @DESC  : check paralogy
	*/
	/**************************************************************************************/
	void check_paralogy(AlrContig & contig);

	/**************************************************************************************/
	/*Reads2snp::calculate_multinomial_terms
	* @INPUT : Posdata object, position values
	   * @DESC  : calculate multinomial terms
	*/
	/**************************************************************************************/
	void calculate_multinomial_terms(Posdata & pos, valarray<unsigned int> & values);

	/**************************************************************************************/
	/*Reads2snp::is_more_than_2_alleles
	* @INPUT : position values
	* @DESC  : calculate if there is more than 2 alleles for selected position and species
	*/
	/**************************************************************************************/
	bool is_more_than_2_alleles(valarray<unsigned int> values);

	/**************************************************************************************/
	/*Reads2snp::calcul_aeb
	* @INPUT : calculated data for contig, ?
	* @DESC  : calcul aeb
	*/
	/**************************************************************************************/
	void calcul_aeb(Contigdata::aeb_data * aeb, double opt_aeb_beta);

	/**************************************************************************************/
	/*Reads2snp::calcul_err
	* @INPUT : calculated data for contig, ? , ? ,contig size
	* @DESC  : calcul err
	*/
	/**************************************************************************************/
	void calcul_err(Contigdata::err_data * err, double opt_er_beta, double opt_tstv, unsigned int size);

	/**************************************************************************************/
	/*Reads2snp::model0
	* @INPUT : AlrContig Object
	* @DESC  : calcul likelihood for model M0
	*/
	/**************************************************************************************/
	void model0(AlrContig & contig);

	/**************************************************************************************/
	/*Reads2snp::dosh
	* @INPUT : AlrContig Object, vector of Optimizer*
	* @DESC  : calcul likelihood for all models (except M0) with parameters shared across species
	*/
	/**************************************************************************************/
	void dosh(AlrContig & contig,vector<Optimizer*> & opt_sh);

	/**************************************************************************************/
	/*Reads2snp::dosep
	* @INPUT : AlrContig Object, vector of Optimizer*
	* @DESC  : calcul likelihood for all models (except M0) with parameters separated across species
	*/
	/**************************************************************************************/
	void dosep(AlrContig & contig,vector< vector<Optimizer*> > & opt_sep);

	/**************************************************************************************/
	/*Reads2snp::chosen_sh_sep
	* @INPUT : AlrContig Object, vector of Optimizer*
	* @DESC  : calcul likelihood and posterior probabilities for chosen models (except M0) with shared separated across species
	*/
	/**************************************************************************************/
	void chosen_sh_sep(AlrContig & contig,vector< vector<Optimizer*> > & opt_sep);

	/**************************************************************************************/
	/*Reads2snp::chosen_sh_dist
	* @INPUT : AlrContig Object, vector of Optimizer*
	* @DESC  : calcul likelihood and posterior probabilities for chosen models (except M0) with parameters separated across species
	*/
	/**************************************************************************************/
	void chosen_sh_dist(AlrContig & contig,vector< Optimizer* > & opt_sh);

	/**************************************************************************************/
	/*Reads2snp::my_optimize
	* @DESC  : optimization function
	*/
	/**************************************************************************************/
	Optimizer* my_optimize();

	bool paraclean(Alrline & pos, bool isheterozygous, unsigned int line);
	void reset_outlog();



};

} /* namespace bpp */
#endif /* READ2SNP_H_ */
