//============================================================================
// Name        : Logvrais.cpp
// Author      : Vincent Cahais, Sylvain Glemin
// Created on  : 13 septembre 2012
// Copyright   : http://www.cecill.info/licences/Licence_CeCILL-C_V1-fr.html
//============================================================================


#include "Stat.h"

namespace bpp
{

/*****************************************************************/
/*functions*/
/*****************************************************************/

	//Probability of a given genotype
	double Stat::pgeno(valarray<unsigned int> & values, valarray<double> proba, double freq)
	{
		return freq*dmultinom(values, proba);
	}

	//Use of the dirichlet-multinomial distribution to take overdispersion into account ###
	double Stat::pgeno_dir(valarray<unsigned int> & values, valarray<double> proba, double freq, double alpha)
	{
		unsigned int n = values.sum();
		//to handle large numbers, lfactorial and lgamma are used then the exponent is taken
		valarray<double> lfv(4);
		for (unsigned int i=0; i<4; i++)
		{
			lfv[i] = logfact(values[i]);
			proba[i]= LnGamma((double)values[i]+(proba[i]*alpha)) - LnGamma(proba[i]*alpha);
		}
		double A = logfact(n) + LnGamma(alpha) - lfv.sum() - LnGamma(n+alpha);
		double B = proba.sum();
 		return freq*exp(A+B);
	}

	//Log likelyhood for an individual: Model0: heterozygots
	double Stat::lnL0 (valarray<unsigned int> & values, double p1)
	{
		double A = pgeno(values, phom1, fhom(p1,fis));
		A += pgeno(values, phet, fhet(p1,fis));
		A += pgeno(values, phom2, fhom(1-p1,fis));
		return log(A);
	}

	//Log-likelihood for one individual Model0: heterozygote
	double Stat::lnL0_dir(valarray<unsigned int> & values, double p1, double alpha)
	{
		double A = pgeno_dir(values, phom1, fhom(p1,fis), alpha);
		A += pgeno_dir(values, phet,  fhet(p1,fis),  alpha);
		A += pgeno_dir(values, phom2, fhom(1-p1,fis), alpha);
		return log(A);
	}

	double Stat::logvrais0(double x)
	{
		double y=0.0;
		for(unsigned int i=0; i<nbind; i+=4) //nbind vaut 4* nombre d'individus ici
		{
			valarray<unsigned int> foo = values_[slice(i,4,1)];
			y = y + lnL0(foo,x);
		}
		return -y;
	}

	double Stat::logvrais0_dir(double x, double alpha)
	{
		double y=0.0;
		for(unsigned int i=0; i<nbind; i+=4) //nbind vaut 4* nombre d'individus ici
		{
			valarray<unsigned int> foo = values_[slice(i,4,1)];
			y = y + lnL0_dir(foo,x, alpha);
		}
		return -y;
	}

	//Log likelyhood for an individual: Model1: paralogs
	double Stat::lnL1(valarray<unsigned int> values, double p1, double p2, double xi)
	{
		double v[]={xi,xi,xi,xi};
		double w[]={1-xi,1-xi,1-xi,1-xi};
		valarray<double> x(v,4);
		valarray<double> y(w,4);
		valarray<double> phom1x = phom1*x;
		valarray<double> phom2x =  phom2*x;
		valarray<double> phetx =  phet*x;
		valarray<double> phom1mx = phom1*y;
		valarray<double> phom2mx =  phom2*y;
		valarray<double> phetmx =  phet*y;

		/*cerr << "A0=" << pgeno(values, phom1      , fhom(p1,fis)   * fhom(p2,fis)) << endl;
		cerr << "A1=" << pgeno(values, phom1x + phetmx	, fhom(p1,fis) 	 * fhet(p2,fis)	 ) << endl;
		cerr << "A2=" << pgeno(values, phom1x + phom2mx	, fhom(p1,fis)	 * fhom(1-p2,fis)) << endl;
		cerr << "A3=" << pgeno(values, phetx  + phom1mx	, fhet(p1,fis)	 * fhom(p2,fis)	) << endl;
		cerr << "A4=" << pgeno(values, phet             , fhet(p1,fis)	 * fhet(p2,fis)) << endl;
		cerr << "A5=" << pgeno(values, phetx  + phom2mx	, fhet(p1,fis)   * fhom(1-p2,fis)) << endl;
		cerr << "A6=" << pgeno(values, phom2x + phom1mx	, fhom(1-p1,fis) * fhom(p2,fis)	) << endl;
		cerr << "A7=" << pgeno(values, phom2x + phetmx	, fhom(1-p1,fis) * fhet(p2,fis)	 ) << endl;
		cerr << "A8=" << pgeno(values, phom2            , fhom(1-p1,fis) * fhom(1-p2,fis)) << endl;*/

		double A = pgeno(values, phom1		, fhom(p1,fis) 	 * fhom(p2,fis));
		A += pgeno(values, phom1x + phetmx	, fhom(p1,fis) 	 * fhet(p2,fis));
		A += pgeno(values, phom1x + phom2mx	, fhom(p1,fis) 	 * fhom(1-p2,fis));
		A += pgeno(values, phetx  + phom1mx	, fhet(p1,fis) 	 * fhom(p2,fis));
		A += pgeno(values, phet				, fhet(p1,fis) 	 * fhet(p2,fis));
		A += pgeno(values, phetx  + phom2mx	, fhet(p1,fis) 	 * fhom(1-p2,fis));
		A += pgeno(values, phom2x + phom1mx	, fhom(1-p1,fis) * fhom(p2,fis));
		A += pgeno(values, phom2x + phetmx	, fhom(1-p1,fis) * fhet(p2,fis));
		A += pgeno(values, phom2			, fhom(1-p1,fis) * fhom(1-p2,fis));
		return log(A);
	}

	//Log-likelihood for one individual: Model1: paralogue
	double Stat::lnL1_dir(valarray<unsigned int> & values, double p1, double p2, double xi, double alpha)
	{
		double v[]={xi,xi,xi,xi};
		double w[]={1-xi,1-xi,1-xi,1-xi};
		valarray<double> x(v,4);
		valarray<double> y(w,4);
		valarray<double> phom1x = phom1*x;
		valarray<double> phom2x =  phom2*x;
		valarray<double> phetx =  phet*x;
		valarray<double> phom1mx = phom1*y;
		valarray<double> phom2mx =  phom2*y;
		valarray<double> phetmx =  phet*y;

		double A = pgeno_dir(values, phom1      , fhom(p1,fis)   * fhom(p2,fis)	 , alpha);
		A += pgeno_dir(values, phom1x + phetmx	, fhom(p1,fis) 	 * fhet(p2,fis)	 , alpha);
		A += pgeno_dir(values, phom1x + phom2mx	, fhom(p1,fis)	 * fhom(1-p2,fis), alpha);
		A += pgeno_dir(values, phetx  + phom1mx	, fhet(p1,fis)	 * fhom(p2,fis)	 , alpha);
		A += pgeno_dir(values, phet             , fhet(p1,fis)	 * fhet(p2,fis)	 , alpha);
		A += pgeno_dir(values, phetx  + phom2mx	, fhet(p1,fis)   * fhom(1-p2,fis), alpha);
		A += pgeno_dir(values, phom2x + phom1mx	, fhom(1-p1,fis) * fhom(p2,fis)	 , alpha);
		A += pgeno_dir(values, phom2x + phetmx	, fhom(1-p1,fis) * fhet(p2,fis)	 , alpha);
		A += pgeno_dir(values, phom2            , fhom(1-p1,fis) * fhom(1-p2,fis), alpha);
		return log(A);
	}

	double Stat::logvrais1(double p1, double p2, double x)
	{
		double y=0.0;
		for(unsigned int i=0; i<nbind; i+=4) //nbind vaut 4* nombre d'individus ici
		{
			valarray<unsigned int> foo = values_[slice(i,4,1)];
			y = y + lnL1(foo,p1,p2,x);
		}
		return -y;
	}

	double Stat::logvrais1_dir(double p1, double p2, double x, double alpha)
	{
		double y=0.0;
		for(unsigned int i=0; i<nbind; i+=4)  //nbind vaut 4* nombre d'individus ici
		{
			valarray<unsigned int> foo = values_[slice(i,4,1)];
			y = y + lnL1_dir(foo,p1,p2,x,alpha);
		}
		return -y;
	}

}


