//============================================================================
// Name        : QuantileBeta.cpp
// Author      : Vincent Cahais
// Created on  : 13 juillet 2012
// Copyright   : http://www.cecill.info/licences/Licence_CeCILL-C_V1-fr.html
//============================================================================

#include "Stat.h"


#define max2(a,b) ((a)>(b)?(a):(b))

namespace bpp
{

	double Stat::CDFBeta (double x, double pin, double qin, double lnbeta)
	{
	/* Returns distribution function of the standard form of the beta distribution,
		that is, the incomplete beta ratio I_x(p,q).

		This is also known as the incomplete beta function ratio I_x(p, q)

		lnbeta is log of the complete beta function; provide it if known,
		and otherwise use 0.

		This is called from QuantileBeta() in a root-finding loop.

		This routine is a translation into C of a Fortran subroutine
		by W. Fullerton of Los Alamos Scientific Laboratory.
		Bosten and Battiste (1974).
		Remark on Algorithm 179, CACM 17, p153, (1974).
	*/

		double ans, c, finsum, p, ps, p1, q, term, xb, xi, y, small = 1e-15;
		int n, i, ib;
		static double eps = 0, alneps = 0, sml = 0, alnsml = 0;

		if (x < small) return 0;
		else if (x > 1 - small) return 1;

		if (pin <= 0 || qin <= 0)  {
			cout << "p=" << pin << " q=" << qin << ": parameter outside range in CDFBeta" << endl;
			return (-1);
		}

		/* initialize machine constants ONCE */
		if (eps == 0) {
			eps = pow((double)FLT_RADIX, -(double)DBL_MANT_DIG);
			alneps = log(eps);
			sml = DBL_MIN;
			alnsml = log(sml);
		}

		y = x;  p = pin;  q = qin;

		/* swap tails if x is greater than the mean */
		if (p / (p + q) < x) {
			y = 1 - y;
			p = qin;
			q = pin;
		}

		if (lnbeta == 0) lnbeta = LnGamma(p) + LnGamma(q) - LnGamma(p + q);

		/* tail approximation */
		if ((p + q) * y / (p + 1) < eps) {
			ans = 0;
			xb = p * log(max2(y, sml)) - log(p) - lnbeta;
			if (xb > alnsml && y != 0) ans = exp(xb);
			if (y != x || p != pin) ans = 1 - ans;

		} else {
			/* evaluate the infinite sum first.  term will equal */
			/* y^p / beta(ps, p) * (1 - ps)-sub-i * y^i / fac(i) */
			ps = q - floor(q);
			if (ps == 0) ps = 1;

			xb = LnGamma(ps) + LnGamma(p) - LnGamma(ps + p);
			xb = p * log(y) - xb - log(p);

			ans = 0;
			if (xb >= alnsml) {
				ans = exp(xb);
				term = ans * p;
				if (ps != 1) {
					n = (int)max2(alneps/log(y), 4.0);
					for (i=1 ; i <=  n ; i++) {
						xi = i;
						term = term * (xi - ps) * y / xi;
						ans = ans + term / (p + xi);
					}
				}
			}

			/* evaluate the finite sum. */
			if (q > 1) {
				xb = p * log(y) + q * log(1 - y) - lnbeta - log(q);
				ib = (int) (xb/alnsml);  if (ib < 0) ib=0;
				term = exp(xb - ib * alnsml);
				c = 1 / (1 - y);
				p1 = q * c / (p + q - 1);

				finsum = 0;
				n = (int) q;
				if (q == (double)n) n = n - 1;

				for (i=1 ; i <= n ; i++) {
					if (p1 <= 1 && term / eps <= finsum) break;

					xi = i;
					term = (q - xi + 1) * c * term / (p + q - xi);

					if (term > 1) {
						ib = ib - 1;
						term = term * sml;
					}

					if (ib == 0) finsum = finsum + term;
				}
				ans = ans + finsum;
			}

			if (y != x || p != pin) ans = 1 - ans;
			if (ans > 1) ans=1;
			if (ans < 0) ans=0;
		}

		return ans;
	}




	double Stat::QuantileBeta(double prob, double p, double q, double lnbeta)
	{
	/* This calculates the Quantile of the beta distribution
		Cran, G. W., K. J. Martin and G. E. Thomas (1977).
		Remark AS R19 and Algorithm AS 109, Applied Statistics, 26(1), 111-114.
		Remark AS R83 (v.39, 309-310) and correction (v.40(1) p.236).

		My own implementation of the algorithm did not bracket the variable well.
		This version is Adpated from the pbeta and qbeta routines from
		"R : A Computer Language for Statistical Data Analysis".  It fails for
		extreme values of p and q as well, although it seems better than my
		previous version.
		Ziheng Yang, May 2001
	*/
		double fpu = 3e-308, acu_min = 1e-300, lower = fpu, upper = 1-2.22e-16;
		/* acu_min >=  fpu: Minimal value for accuracy 'acu' which will depend on (a,p); */
		int swap_tail, i_pb, i_inn, niterations = 2000;
		double a, adj, g, h, pp, prev = 0, qq, r, s, t, tx = 0, w, y, yprev;
		double acu, xinbta;

		/* define accuracy and initialize */
		xinbta = prob;

		/* test for admissibility of parameters */
		if (p < 0 || q < 0 || prob < 0 || prob > 1) {
			cerr << "beta parameter error" << endl;
			exit(0);
		}
		if (prob == 0 || prob == 1) return prob;

		if (lnbeta == 0) lnbeta = LnGamma(p) + LnGamma(q) - LnGamma(p + q);

		/* change tail if necessary;  afterwards   0 < a <= 1/2    */
		if (prob <= 0.5) {
			a = prob;
			pp = p;
			qq = q;
			swap_tail = 0;

		} else {
			a = 1. - prob;
			pp = q;
			qq = p;
			swap_tail = 1;
		}

		/* calculate the initial approximation */
		r = sqrt(-log(a * a));
		y = r - (2.30753+0.27061*r)/(1.+ (0.99229+0.04481*r) * r);

		if (pp > 1. && qq > 1.) {
			r = (y * y - 3.) / 6.;
			s = 1. / (pp*2. - 1.);
			t = 1. / (qq*2. - 1.);
			h = 2. / (s + t);
			w = y * sqrt(h + r) / h - (t - s) * (r + 5./6. - 2./(3.*h));
			xinbta = pp / (pp + qq * exp(w + w));

		} else {
			r = qq*2.;
			t = 1. / (9. * qq);
			t = r * pow(1. - t + y * sqrt(t), 3.);
			if (t <= 0.) {
				xinbta = 1. - exp((log((1. - a) * qq) + lnbeta) / qq);

			} else {
				t = (4.*pp + r - 2.) / t;
				if (t <= 1.) {
					xinbta = exp((log(a * pp) + lnbeta) / pp);
				} else {
					xinbta = 1. - 2./(t + 1.);
				}
			}
		}

		/* solve for x by a modified newton-raphson method, using CDFBeta */
		r = 1. - pp;
		t = 1. - qq;
		yprev = 0.;
		adj = 1.;

		/* Changes made by Ziheng to fix a bug in qbeta()
		qbeta(0.25, 0.143891, 0.05) = 3e-308   wrong (correct value is 0.457227)
		*/
		if (xinbta <= lower || xinbta >= upper) xinbta = (a + .5)/2;

		/* Desired accuracy should depend on (a,p)
		* This is from Remark .. on AS 109, adapted.
		* However, it's not clear if this is "optimal" for IEEE double prec.
		* acu = fmax2(acu_min, pow(10., -25. - 5./(pp * pp) - 1./(a * a)));
		* NEW: 'acu' accuracy NOT for squared adjustment, but simple;
		* ---- i.e.,  "new acu" = sqrt(old acu)
		*/
		acu = pow(10., -13. - 2.5/(pp * pp) - 0.5/(a * a));
		acu = max2(acu, acu_min);

		for (i_pb=0; i_pb < niterations; i_pb++) {
			y = CDFBeta(xinbta, pp, qq, lnbeta);
			y = (y - a) *
			exp(lnbeta + r * log(xinbta) + t * log(1. - xinbta));

			if (y * yprev <= 0) prev = max2(fabs(adj),fpu);

			for (i_inn=0,g=1; i_inn < niterations; i_inn++) {
				adj = g * y;
				if (fabs(adj) < prev) {
					tx = xinbta - adj; /* trial new x */
					if (tx >= 0. && tx <= 1.) {
						if (prev <= acu || fabs(y) <= acu) goto L_converged;
						if (tx != 0. && tx != 1.) break;
					}
				}
				g /= 3.;
			}
			if (fabs(tx - xinbta) < fpu) goto L_converged;
			xinbta = tx;
			yprev = y;
		}

		L_converged:
		return (swap_tail ? 1. - xinbta : xinbta);
	}




	/* discretize_beta */
	void Stat::discretize_beta(double a, double b, int nbcat, double* discr_value)
	{

		double p;
		double lnbeta;

		lnbeta = LnGamma(a) + LnGamma(b) - LnGamma(a + b);

		for (int j=0; j < nbcat; j++) {
			p = (j*2.+1)/(2.*nbcat);
			discr_value[j] = QuantileBeta(p, a, b, lnbeta);
		}
	}


}
