//============================================================================
// Name        : Stat.h
// Author      : Vincent Cahais, Sylvain Glemin
// Created on  : 13 juillet 2012
// Copyright   : http://www.cecill.info/licences/Licence_CeCILL-C_V1-fr.html
//============================================================================

#ifndef STAT_H_
#define STAT_H_

#include <vector>
#include <math.h>
#include <cmath>
#include <stdexcept>
#include <float.h>
#include <valarray>
#include <iostream>
#include <sstream>

#include "../Utils.hpp"


using namespace std;

namespace bpp
{

class Stat
{
public:

	valarray<unsigned int> values_;
	unsigned int nbind;

	double error_;
	double fis;
	double prec_;

	valarray <double> phom1;
	valarray <double> phom2;
	valarray <double> phet;

	vector<double> logfact_;

	Stat(): error_(0.001), fis(0), prec_(0.001), phom1(4), phom2(4), phet(4){init_logfact(10); set_error(0.001);}

	virtual ~Stat(){}

	/*****************************************************************/
	/*initializers*/
	/*****************************************************************/
	void set_fis(double f){fis = f;}
	void set_precision(double p){prec_ = p;}

	void set_error(double e)
	{
		error_ = e;
		phom1[1] = phom1[2] = phom1[3] = phom2[0] = phom2[2] = phom2[3] = phet[2] = phet[3] = e;
		phom1[0] = phom2[1] = 1-3*e;
		phet[0] = phet[1] = (1-2*e)/2;
	}

	void set_values(valarray<unsigned int> & v){values_ = v, nbind = v.size();}

	void init_logfact(unsigned int MAX_PRECALCULATED_LOGFACT);
	void init_loggamma(unsigned int MAX_PRECALCULATED_LOGGAMMA);


	/*****************************************************************/
	/*getters*/
	/*****************************************************************/

	double logfact(unsigned int n);
	double LnGamma (unsigned int x);
	double LnGamma(double x);
	valarray<double> logfact (valarray<unsigned int > val);
	valarray<double> LnGamma (valarray<double> val);

	//Genotype frequency
	double fhom(double p, double f){ return p*p*(1-f)+f*p; }
	double fhet(double p, double f){ return 2*p*(1-p)*(1-f); }

	/*****************************************************************/
	/*Dmultinomiale.cpp*/
	/*****************************************************************/
	double dmultinom(valarray<unsigned int > & values, valarray<double> & proba);

	/*****************************************************************/
	/*Logvrais.cpp*/
	/*****************************************************************/
	//Probability of a given genotype
	double pgeno(valarray<unsigned int> & values, valarray<double> proba, double freq);
	double pgeno_dir(valarray<unsigned int> & values, valarray<double> proba, double freq, double alpha);

	//Log likelyhood for an individual: Model0: heterozygots
	double lnL0 (valarray<unsigned int> & values, double p1);
	double lnL0_dir(valarray<unsigned int> & values, double p1, double alpha);
	double logvrais0(double x);
	double logvrais0_dir(double x, double alpha);

	//Log likelyhood for an individual: Model1: paralogs
	double lnL1(valarray<unsigned int> values, double p1, double p2, double xi);
	double lnL1_dir(valarray<unsigned int> & values, double p1, double p2, double xi, double alpha);
	double logvrais1(double p1, double p2, double x);
	double logvrais1_dir(double p1, double p2, double x, double alpha);

	/*****************************************************************/
	/*QuantileBeta.cpp*/
	/*****************************************************************/
	double CDFBeta (double x, double pin, double qin, double lnbeta);
	double QuantileBeta(double prob, double p, double q, double lnbeta);
	void discretize_beta(double a, double b, int nbcat, double* discr_value);


	/*****************************************************************/
	/*bino*/
	/*****************************************************************/
	double log_bino_proba(int k, int n, double p)
	{
		return logfact(n) - logfact(k) - logfact(n - k) + k*log(p) + (n - k)*log(1 - p);
	}

	double cumul_bino_proba(int k, int n, double p)
	{
		double cb = 0.;
		for (int i=0; i <= k; i++)
			cb += exp(log_bino_proba(i, n, p));
		return cb;
	}

	/* log_sum_exp */
	/* Calculate the log of the sum of the exponentials of entries in v, */
	/* accounting for high negative values in v such that exp(v[i]) is underflow zero */
	/* //Only negative entries of v are considered.  */

	double log_sum_exp(vector<double> & v)
	{
		int n = v.size();
		if (n == 1) return v[0];

		if (n == 0) {
			cerr << "Empty vector passed to log_sum_exp\n";
			exit(0);
		}

		int max = v[0];

		for (int i=1; i < n; i++) {
			if (v[i] > max) max = v[i];
		}

		double sum = 0.;

		for (int i=0; i < n; i++)
			/*if (v[i] < 0.)*/ sum += exp(v[i] - max);

		return log(sum) + max;
	}


};

} /* namespace paraclean */
#endif /* STAT_H_ */
